/*
FILE:     CifFile.C
*/
/*
VERSION:  3.004
*/
/*
DATE:     9/11/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


/*!
** \file CifFile.C
**
** \brief Implementation file for CifFile class.
*/


/* 
  PURPOSE:    Base class for read/write cif files
*/

// VLAD - Carefully examine all methods that accept with, for cases where
// with is 0. This may happen if the value is empty !!

#include "GenString.h"
#include "CifString.h"
#include "regex.h"
#include "CifFile.h"

#ifdef VLAD_DEBUG
bool VladDebug = false;
#endif


CifFile::CifFile(const eFileMode fileMode, const string& objFileName,
  const bool verbose, const StringCompare::eCompareType caseSense,
  const unsigned int maxLineLength, const string& nullValue) :
  TableFile(fileMode, objFileName, caseSense)
{

    Init();

    _maxCifLineLength = maxLineLength;
    _nullValue = nullValue;
    _verbose = verbose;

}


CifFile::CifFile(const bool verbose,
  const StringCompare::eCompareType caseSense,
  const unsigned int maxLineLength, const string& nullValue) :
  TableFile(caseSense)
{

    Init();

    _maxCifLineLength = maxLineLength;
    _nullValue = nullValue;
    _verbose = verbose;

}


CifFile::~CifFile()
{

}


void CifFile::Init()
{

    _maxCifLineLength = STD_CIF_LINE_LENGTH;
    _nullValue = CifString::NullValue;
    _verbose = false;
    _smartPrint = false;
    _quotes = "\'";

}


void CifFile::SetQuoting(eQuoting quoting)
{

    if (quoting == eSINGLE)
    {
        _quotes = "\'";
    }
    else if (quoting == eDOUBLE)
    {
        _quotes = "\"";
    } 
    // VLAD - ERROR HANDLING
}


unsigned int CifFile::GetQuoting()
{

    if (_quotes == "\'")
        return(eSINGLE);
    else if (_quotes == "\"")
        return(eDOUBLE);
    else
        // VLAD - ERROR HANDLING
        return(eDOUBLE);
       
}


void CifFile::Write(const string& cifFileName, const bool sortTables,
  const bool writeEmptyTables)
{

    ofstream cifo(cifFileName.c_str(), ios::out | ios::trunc);

    Write(cifo, sortTables, writeEmptyTables);

    cifo.close();

}


void CifFile::Write(ostream& cifo, const bool sortTables,
  const bool writeEmptyTables)
{

    vector<unsigned int> tablesIndices;

    if (sortTables)
    {
        GetSortedTablesIndices(tablesIndices);
    }
    else
    {
        GetTablesIndices(tablesIndices);
    }

    Write(cifo, tablesIndices, writeEmptyTables);

}


void CifFile::Write(const string& cifFileName, const vector<string>& catOrder,
  const bool writeEmptyTables)
{

    ofstream cifo(cifFileName.c_str(), ios::out | ios::trunc);

    Write(cifo, catOrder, writeEmptyTables);

    cifo.close();

}


void CifFile::Write(ostream& cifo, const vector<string>& catOrder,
  const bool writeEmptyTables)
{

    vector<bool> processedTables(GetTotalNumTables());
    vector<unsigned int> tables;
    unsigned int tableIndex = 0;

    // VLAD : Can this be improved by first searching the category and
    // then doing something about it? Well maybe this is not doable.
    for (unsigned int blockI = 0, numTables = 0; blockI < _blocks.size();
      blockI++)
    {
        for (unsigned int l = 0; l < catOrder.size(); l++)
        {
            tableIndex = _blocks[blockI]._tables.\
              find(catOrder[l]);

            if (tableIndex != _blocks[blockI]._tables.size())
            {
                tables.push_back(tableIndex);
                processedTables[tableIndex + numTables] = true;
            }
            else
            {
                processedTables[tableIndex + numTables] = false;
            }
        }
        numTables += _blocks[blockI]._tables.size();
    }

    for (unsigned int blockI = 0, numTables = 0; blockI < _blocks.size();
      ++blockI)
    {
        for (unsigned int tableI = 0; tableI < _blocks[blockI]._tables.size();
          ++tableI)
        {
            if (!processedTables[tableI + numTables])
            {
                tables.push_back(tableI);
            }
        }
        numTables += _blocks[blockI]._tables.size();
    }

    Write(cifo, tables, writeEmptyTables);

}


int CifFile::DataChecking(CifFile& ref)
{

    int ret = 0;

    ofstream log;

    log.open("Diagnostics.log", ios::out | ios::app);

    vector<string> refBlockNames;
    ref.GetBlockNames(refBlockNames);

    if (refBlockNames.size() > 1)
    {
        log << "WARNING - dictionary has " << refBlockNames.size() <<
          " datablocks" << endl;
    }

    Block& refBlock = ref.GetBlock(ref.GetFirstBlockName());

    vector<string> BlockNames;
    GetBlockNames(BlockNames);

    for (unsigned int blockI = 0; blockI < BlockNames.size(); ++blockI)
    {
        Block& block = GetBlock(BlockNames[blockI]);

        ostringstream buf;

        ret = DataChecking(block, refBlock, buf);

        string sBuf = buf.str();

        if (!sBuf.empty())
        {
            log << sBuf << endl;
        }
#ifdef VLAD_DEBUG
       if (ret == 5)
           log.close();
        return ret;
#endif
    }

    log.close();

    return(ret);

}


int CifFile::DataChecking(Block& block, Block& refBlock, ostringstream& buf)
{

    int ret1, ret2;

    ret1 = CheckCategories(block, refBlock, buf);
    CheckCategoryKey(block, buf);
    ret2 = CheckItems(block, refBlock, buf);

    CheckAndRectifyItemTypeCode(block, buf);

#ifdef VLAD_DEBUG
    if (ret2 == 5)
        return ret2;
#endif
    if (ret1 == 0)
        return ret2;
    else
        return ret1;

}


int CifFile::_IsQuotableText(const string& itemValue)
{

  if (itemValue.empty())
      return(0);

  if (itemValue[0] == '_')
      return(1);

  for (unsigned int i = 0; i < itemValue.size(); i++)
  {
      if (itemValue[i] == ' ') 
          return(1);
      else if (itemValue[i] == '\n') 
          return(1);
      else if (itemValue[i] == '\'' || itemValue[i] == '\"') 
          return(1);
  }

  if (String::IsCiEqual(itemValue.substr(0, 5), "data_"))
      return(1);
  if (String::IsCiEqual(itemValue.substr(0, 5), "loop_"))
      return(1);
  if (String::IsCiEqual(itemValue.substr(0, 5), "save_"))
      return(1);

  return(0);

}


CifFile::eIdentType CifFile::_FindPrintType(const vector<string>& values)
{

    // If at least one value is a number, all should be right justified.
    for (unsigned int valuesI = 0; valuesI < values.size(); ++valuesI)
    {
        if (String::IsNumber(values[valuesI]))
        {
            return(eRIGHT);
        }
    }

    return(eLEFT);

}


void CifFile::_PrintItemIdent(ostream& cifo, unsigned int& linePos)
{

    string ItemNameIdent = "    ";
 
    cifo << ItemNameIdent;

    linePos = ItemNameIdent.size();

}


void CifFile::_PrintItemName(ostream& cifo, const string& category,
  const string& keyword, unsigned int& linePos)
{

    string cifItem;

    CifString::MakeCifItem(cifItem, category, keyword);

    cifo << cifItem;

    linePos += cifItem.size();

}


void CifFile::_PrintPostItemSeparator(ostream& cifo, unsigned int& linePos,
  const bool ident, const unsigned int numSpaces)
{

    const string Space = " ";

    for (unsigned int spaceI = 0; spaceI < numSpaces; spaceI++)
    {
        cifo << Space;
    }

    linePos += numSpaces * Space.size();

    if (!ident)
    {
        return;
    }

    unsigned int start = 36;

    while (linePos > start)
        start += 10;

    for (unsigned int i = 0; i < start - linePos; i++)
        cifo << Space;

    linePos = start;

}


int CifFile::_PrintItemValue(ostream& cifo, const string& itemValue,
  unsigned int& linePos, const eIdentType identType,
  const unsigned int width)
{

    bool multipleLine, multipleWord, embeddedQuotes, embeddedSingleQuotes,
      embeddedDoubleQuotes;
    string Ident;
    int N = 0;
    // VLAD: with has different meaning for eNONE and others.

    if ((identType == eNONE) && (width != 0))
    {
        Ident = "          ";
    }

    if (linePos == 0)
    {
        cifo << Ident;
        linePos = Ident.size();
    }
  
    if (itemValue.empty())
    {
        if (_maxCifLineLength <= linePos + 2)
        {
            cifo << endl;
            linePos = 0;
        }

        N = _nullValue.size();

        if (identType == eRIGHT)
        {
            if (linePos + width - N < _maxCifLineLength)
                for (unsigned int k = 0; k < width - N; k++)
                    cifo << " ";
                linePos += width - N;
        }

        cifo << _nullValue;
        linePos += 1;

        if (identType == eLEFT)
        {
            if (linePos != 0)
            {
                if (linePos + width - N < _maxCifLineLength)
                    for (unsigned int k = 0; k < width - N; k++)
                        cifo << " ";
                    linePos += width - N;
            }
        }

        cifo << " ";
        linePos += 1;

        return(1);
    }
  
    unsigned int str_len = itemValue.size();

    multipleLine = false;
    multipleWord = false;
    embeddedQuotes = false;
    embeddedSingleQuotes = false;
    embeddedDoubleQuotes = false;

    string multipleWordQuotes = _quotes;

    for (unsigned int i = 0; i < str_len; i++)
    {
        if (itemValue[i] == ' ')
            multipleWord = true;
        else if (itemValue[i] == '\n')
            multipleLine = true;
        else if (itemValue[i] == '\'')
        {
            embeddedSingleQuotes = true;
            embeddedQuotes = true;
            multipleWordQuotes = "\"";
        }
        else if (itemValue[i] == '\"')
        {
            embeddedDoubleQuotes = true;
            embeddedQuotes = true;
            multipleWordQuotes = "\'";
        }
    }

    if (itemValue[0] == '_')
        multipleWord = true;
    if (itemValue[0] == ';')
        multipleWord = true;

    if (String::IsCiEqual(itemValue.substr(0, 5), "data_"))
        multipleWord = true;
    if (String::IsCiEqual(itemValue.substr(0, 5), "loop_"))
        multipleWord = true;
    if (String::IsCiEqual(itemValue.substr(0, 5), "save_"))
        multipleWord = true;

    if (embeddedQuotes && multipleWord)
        multipleLine = true;
    if (embeddedQuotes)
        multipleWord = true;

    if (embeddedSingleQuotes && embeddedDoubleQuotes)
        multipleWordQuotes = _quotes;

    if (str_len >= _maxCifLineLength || multipleLine)
    {
        if (linePos != 0)
            cifo << endl;
        cifo << ";" << itemValue;
        cifo << endl;
        cifo << ";" << endl;
        linePos = 0;
    }
    else
    {
        if ((!multipleWord && str_len + 2 + linePos > _maxCifLineLength) ||
	  (multipleWord && str_len + 4 + linePos > _maxCifLineLength))
        {
            cifo << endl;
            linePos = 0;
            cifo << Ident;
            linePos += Ident.size();
        }

        string fullItemValue;

        if (multipleWord)
        {
            fullItemValue = multipleWordQuotes + itemValue + multipleWordQuotes;
        }
        else
        {
            fullItemValue = itemValue;
        }

        N = fullItemValue.size();

        if (identType == eRIGHT)
        {
            if (linePos + width - N < _maxCifLineLength)
                for (unsigned int k = 0; k < width - N; k++)
                    cifo << " ";
                linePos += width - N;
        }

        cifo << fullItemValue;
        linePos += N;

        if ((identType == eNONE) || (identType == eLEFT))
        {
            cifo << " ";
            linePos++;
            N++;
        }
    }
 
    if ((identType == eNONE) && (!Ident.empty()))
    {
        if (linePos > Ident.size())
        {
            linePos = linePos + width - N;
            for (int i = 0; i < (int)(width - N); i++)
                cifo << " ";
        }
    }

    if (identType == eLEFT)
    {
        if (linePos != 0)
        {
            if (linePos + width - (N - 1) < _maxCifLineLength)
                for (unsigned int k = 0; k < width - (N - 1); k++)
                    cifo << " ";
                linePos += width - (N - 1);
        }
    }

    if (identType == eRIGHT)
    {
        cifo << " ";
        linePos++;
        N++;
    }

    if (multipleWord || multipleLine || embeddedQuotes) 
        return(str_len + 2);
    else
        return(str_len);
 
}


int CifFile::_PrintItemNameInHeader(ostream& cifo, const string& itemValue,
  unsigned int& linePos, const eIdentType identType,
  const unsigned int width)
{

    bool multipleLine, multipleWord, embeddedQuotes, embeddedSingleQuotes,
      embeddedDoubleQuotes;
    string Ident;
    int N = 0;
    // VLAD: with has different meaning for eNONE and others.

    if ((identType == eNONE) && (width != 0))
    {
        Ident = "          ";
    }

    if (linePos == 0)
    {
        cifo << Ident;
        linePos = Ident.size();
    }
  
    if (itemValue.empty())
    {
        if (_maxCifLineLength <= linePos + 2)
        {
            cifo << endl << "# ";
            linePos = 2;
        }

        N = _nullValue.size();

        if (identType == eRIGHT)
        {
            if (linePos + width - N < _maxCifLineLength)
                for (unsigned int k = 0; k < width - N; k++)
                    cifo << " ";
                linePos += width - N;
        }

        cifo << _nullValue;
        linePos += 1;

        if (identType == eLEFT)
        {
            if (linePos != 0)
            {
                if (linePos + width - N < _maxCifLineLength)
                    for (unsigned int k = 0; k < width - N; k++)
                        cifo << " ";
                    linePos += width - N;
            }
        }

        cifo << " ";
        linePos += 1;

        return(1);
    }
  
    unsigned int str_len = itemValue.size();

    multipleLine = false;
    multipleWord = false;
    embeddedQuotes = false;
    embeddedSingleQuotes = false;
    embeddedDoubleQuotes = false;

    string multipleWordQuotes = _quotes;

    for (unsigned int i = 0; i < str_len; i++)
    {
        if (itemValue[i] == ' ')
            multipleWord = true;
        else if (itemValue[i] == '\n')
            multipleLine = true;
        else if (itemValue[i] == '\'')
        {
            embeddedSingleQuotes = true;
            embeddedQuotes = true;
            multipleWordQuotes = "\"";
        }
        else if (itemValue[i] == '\"')
        {
            embeddedDoubleQuotes = true;
            embeddedQuotes = true;
            multipleWordQuotes = "\'";
        }
    }

    if (itemValue[0] == '_')
        multipleWord = true;
    if (itemValue[0] == ';')
        multipleWord = true;

    if (String::IsCiEqual(itemValue.substr(0, 5), "data_"))
        multipleWord = true;
    if (String::IsCiEqual(itemValue.substr(0, 5), "loop_"))
        multipleWord = true;
    if (String::IsCiEqual(itemValue.substr(0, 5), "save_"))
        multipleWord = true;

    if (embeddedQuotes && multipleWord)
        multipleLine = true;
    if (embeddedQuotes)
        multipleWord = true;
    if (embeddedSingleQuotes && embeddedDoubleQuotes)
        multipleWordQuotes = _quotes;

    if (str_len >= _maxCifLineLength || multipleLine)
    {
        if (linePos != 0)
            cifo << endl;
        cifo << ";" << itemValue;
        cifo << endl;
        cifo << ";" << endl;
        linePos = 0;
    }
    else
    {
        if ((!multipleWord && str_len + 2 + linePos > _maxCifLineLength) ||
	  (multipleWord && str_len + 4 + linePos > _maxCifLineLength))
        {
            cifo << endl << "# ";
            linePos = 2;
            cifo << Ident;
            linePos += Ident.size();
        }

        string fullItemValue;

        if (multipleWord)
        {
            fullItemValue = multipleWordQuotes + itemValue + multipleWordQuotes;
        }
        else
        {
            fullItemValue = itemValue;
        }

        N = fullItemValue.size();

        if (identType == eRIGHT)
        {
            if (linePos + width - N < _maxCifLineLength)
                for (unsigned int k = 0; k < width - N; k++)
                    cifo << " ";
                linePos += width - N;
        }

        cifo << fullItemValue;
        linePos += N;

        if ((identType == eNONE) || (identType == eLEFT))
        {
            cifo << " ";
            linePos++;
            N++;
        }
    }
 
    if ((identType == eNONE) && (!Ident.empty()))
    {
        if (linePos > Ident.size())
        {
            linePos = linePos + width - N;
            for (int i = 0; i < (int)(width - N); i++)
                cifo << " ";
        }
    }

    if (identType == eLEFT)
    {
        if (linePos != 0)
        {
            if (linePos + width - (N - 1) < _maxCifLineLength)
                for (unsigned int k = 0; k < width - (N - 1); k++)
                    cifo << " ";
                linePos += width - (N - 1);
        }
    }

    if (identType == eRIGHT)
    {
        cifo << " ";
        linePos++;
        N++;
    }

    if (multipleWord || multipleLine || embeddedQuotes) 
        return(str_len + 2);
    else
        return(str_len);
 
}


void CifFile::Write(ostream& cifo, vector<unsigned int>& tables,
  const bool writeEmptyTables)
{

    ISTable* tableP = NULL;
 
    unsigned int numColumn, numRow, linePos, cwid, ilen, numPostItemSpaces;

    vector<string> rowValues;
 
    vector<unsigned int> cwidth;
 
    if (_smartPrint)
    {
        numPostItemSpaces = SMART_PRINT_SPACING;
    }
    else
    {
        numPostItemSpaces = STD_PRINT_SPACING;
    }

 
    for (unsigned int blockI = 0, jt = 0;
      blockI < _blocks.size(); blockI++)
    {
        cifo << "data_" << _blocks[blockI].GetName() << endl;

        for (unsigned int tableI = 0; tableI < _blocks[blockI]._tables.size();
          ++tableI, ++jt)
        {
// VLAD - EXAMINE See this same loop in DICFileObjBase or DDLFileObjBase
// VLAD PERFORMANCE. The old BTreeObj code was doing a partial search
// on tids. It searched for "_0_ " in the tree. Even though the exact
// search was never found, it moved the pointer down the tree and reached
// the appropriate "_0_ blah" for example and then processed from that point
// onward the entries that are exact match. This kind of modified proprietory
// search is better than STL. We need to represent this subsearch using
// different STL structures to be able to have better performance than what
// we currently have.

            cifo << "# " << endl;

            tableP = _GetTablePtr(blockI, tables[jt]);

            numColumn = tableP->GetNumColumns();
            numRow = tableP->GetNumRows();

            if ((numRow == 0) && (!writeEmptyTables))
                continue;

            const vector<string>& colNames = tableP->GetColumnNames();

            if (numRow <= 1)
            {
                // cwid is maxColumnNameSize
                unsigned int longestNameIndex = 0;

                cwid = 0;
                for (unsigned int i = 0; i < numColumn; i++)
                {
                    if (colNames[i].size() > cwid)
                    {
                          cwid = colNames[i].size();
                          longestNameIndex = i;
                    }
                }
 
                string longestCifItem;

                CifString::MakeCifItem(longestCifItem,
                  tableP->GetName(), colNames[longestNameIndex]);

                if (numRow == 0)
                {
                    for (unsigned int i = 0; i < numColumn; i++)
                    {
                        rowValues.push_back(_nullValue);
                    }
                }
                else
                {
                    tableP->GetRow(rowValues, 0);
                }

                for (unsigned int i = 0; i < numColumn; i++)
                {
                    linePos = 0;
                    _PrintItemName(cifo,
                      tableP->GetName(),
                      colNames[i], linePos);
                    _PrintPostItemSeparator(cifo, linePos, false,
                      numPostItemSpaces + colNames[longestNameIndex].size() -
                      colNames[i].size());

                    linePos = longestCifItem.size() + numPostItemSpaces - 1;

                    _PrintItemValue(cifo, rowValues[i], linePos);

                    if (linePos != 0)
                        cifo << endl;
                }
                rowValues.clear();
            }
            else
            {
                cifo << "loop_" << endl;

                for (unsigned int i = 0; i < numColumn; i++)
                {
                    linePos = 0;
                    _PrintItemName(cifo,
                      tableP->GetName(),
                      colNames[i], linePos);
                    _PrintPostItemSeparator(cifo, linePos);
                    cifo << endl;
                }

                unsigned int addSpace;

                for (unsigned int i = 0; i < numColumn; i++)
                {
                    if (_smartPrint)
                    {
                        // In smart printing, the column width is determined
                        // based on column values as well as column name
                        if (i == 0)
                        {
                            // "# "
                            addSpace = 2;
                        }
                        else
                        {
                            addSpace = 0;
                        }
                        cwidth.push_back(colNames[i].size() + addSpace);
                    }
                    else
                    { 
                        // Even if values are empty, they should occupy at
                        // least one character.
                        cwidth.push_back(1);
                    }
                }

                for (unsigned int l = 0; l < numRow; l++)
                {
                    tableP->GetRow(rowValues, l);
                    for (unsigned int i = 0; i < numColumn; i++)
                    {
                        ilen = rowValues[i].size();
                        if (_IsQuotableText(rowValues[i]))
                            ilen += 2;
                        if (ilen > cwidth[i])
                        {
                            if ((i == 0) && _smartPrint)
                                cwidth[i] = ilen + 2;
                            else
                                cwidth[i] = ilen;
                        }
                    }
                    rowValues.clear();
                }

                // In smart print mode, determine justification type for
                // each column.
                vector<eIdentType> colPrintType;
                vector<string> col;
                for (unsigned int colI = 0; colI < numColumn; ++colI)
                {
                    if (_smartPrint)
                    {
                        tableP->GetColumn(col, colNames[colI]);

                        colPrintType.push_back(_FindPrintType(col));

                        col.clear();
                    }
                    else
                    {
                        colPrintType.push_back(eLEFT);
                    }
                }

                for (unsigned int l = 0; l < numRow; l++)
                {
                    if (_smartPrint)
                    {
                        if ((l == 0) || ((l % HEADER_SPACING) == 0))
                        {
                            _PrintHeaderedItems(cifo, colNames, cwidth,
                              colPrintType);
                        }
                        // VLAD - PRINT - Add here printing the header
                        // every certain number of lines
                    }
                    linePos = 0;
                    tableP->GetRow(rowValues, l);
                    for (unsigned int i = 0; i < numColumn; i++)
                    {
                        _PrintItemValue(cifo, rowValues[i], linePos,
                          colPrintType[i], cwidth[i]);
                    }

                    if (linePos != 0)
                        cifo << endl;

                    rowValues.clear();
                }

                cwidth.clear();
            }
        }
        cifo << "# " << endl;
    }

}

void CifFile::_PrintHeaderedItems(ostream& cifo,
  const vector<string>& colNames, const vector<unsigned int>& colWidths,
  const vector<eIdentType> colPrintType)
{

    unsigned int linePos = 0;
    unsigned int lessSpace = 0;

    cifo << "# ";

    linePos += 2;

    for (unsigned int colI = 0; colI < colNames.size(); ++colI)
    {
        if (colI == 0)
        {
            lessSpace = 2;
        }
        else
        {
            lessSpace = 0;
        }

        _PrintItemNameInHeader(cifo, colNames[colI], linePos,
          colPrintType[colI], colWidths[colI] - lessSpace);
    }

    cifo << endl;

}


int CifFile::CheckCategories(Block& block, Block& refBlock, ostringstream& log)
{

    int ret = 1;

    ISTable* refCatTableP = refBlock.GetTablePtr("category");

    vector<string> list;
    list.push_back("mandatory_code");

    vector<string> target;
    target.push_back("yes");

    vector<unsigned int> OutList;
    refCatTableP->Search(OutList, target, list);
    if (OutList.empty())
    {
        return(ret);
    }

    for (unsigned int i = 0; i < OutList.size(); i++)
    {
        // For all the mandatory categories
        const string& mandatoryCat = (*refCatTableP)(OutList[i], "id");
        if (!block.IsTablePresent(mandatoryCat))
        {
            log << "ERROR - category \"" << mandatoryCat <<
              "\" is mandatory, but it is not present in datablock \"" <<
              block.GetName() << "\"" << endl;
            ret = 0;
        }
    }

    return(ret);

}


void CifFile::CheckCategoryKey(Block& block, ostringstream& log)
{

    ISTable* catKeyTableP = block.GetTablePtr("category_key");

    ISTable* itemTableP = block.GetTablePtr("item");

    if ((catKeyTableP == NULL) || (itemTableP == NULL))
        return;

    vector<string> searchCols;
    searchCols.push_back("name");

    // For every item in the category key table, check if it is set as
    // mandatory or implicit
    for (unsigned int itemI = 0; itemI < catKeyTableP->GetNumRows(); ++itemI)
    {
        const string& cifItem = (*catKeyTableP)(itemI, "name");

        vector<string> targets;
        targets.push_back(cifItem);

        unsigned int foundIndex = itemTableP->FindFirst(targets, searchCols);
        if (foundIndex == itemTableP->GetNumRows())
        {
            log << "ERROR - In block \"" << block.GetName() <<
              "\", item \"" << cifItem << "\" exists in category_key " <<
              " but it is not defined" << endl;
#ifdef VLAD_DEBUG
            cout << "PROBLEMATIC SEARCH TABLE: " << (*itemTableP) << endl;
            exit(1);
#endif
        }
        else
        {
            const string& mandatoryCode = (*itemTableP)(foundIndex,
              "mandatory_code");

            if ((mandatoryCode != "yes") && (mandatoryCode != "implicit"))
            {
                log << "ERROR - In block \"" << block.GetName() <<
                  "\", key item \"" << cifItem << "\" is defined as" <<
                  " non-mandatory" << endl;
            }
        }
    }

}


int CifFile::CheckItems(Block& block, Block& refBlock, ostringstream& log)
{

    int ret = 1;
 

    // Get references to relevant dictionary categories

    ISTable* refCatTableP = refBlock.GetTablePtr("category");

    ISTable* catKeyTableP = refBlock.GetTablePtr("category_key");

    // Set search on "id" column to be case-insensitive, since the "id" item
    // of "category_key" category is implicitly set from the category name of
    // category save frame in the reference file and that category name in
    // category save frame is uppercase.
    catKeyTableP->SetFlags("id", ISTable::DT_STRING | ISTable::CASE_INSENSE);

    ISTable* itemTableP = refBlock.GetTablePtr("item");

    ISTable* itemLinkedTableP = refBlock.GetTablePtr("item_linked");

    ISTable* itemTypeTableP = refBlock.GetTablePtr("item_type");

    ISTable* itemTypeListTableP = refBlock.GetTablePtr("item_type_list");
  
    ISTable* itemRangeTableP = refBlock.GetTablePtr("item_range");

    ISTable* itemEnumTableP = refBlock.GetTablePtr("item_enumeration");

    vector<string> catNames;

    block.GetTableNames(catNames);
    for (unsigned int catI = 0; catI < catNames.size(); ++catI)
    {
#ifdef VLAD_PERF
        cout << "Checking category: " << catNames[catI] << "; " << catI <<
          " of " << catNames.size() << endl;
#endif
        ISTable* catTableP = block.GetTablePtr(catNames[catI]);

#ifdef VLAD_DEBUG
        if (catI == 24)
        {
            VladDebug = true;
        }
#endif

#ifdef VLAD_PERF
        log << "Category: \"" << catNames[catI] << "\"; Index: " << catI
          << " of " << catNames.size() << endl;
#endif
        bool skipCatCheck = false;

        // Check if the table is defined in the reference file
        if (!IsCatDefinedInRef(catNames[catI], *refCatTableP))
        {
            log << "ERROR - In block \"" << block.GetName() <<
              "\", category \"" << catNames[catI] << "\" is not " <<
              "defined in the reference file" << endl;
            skipCatCheck = true;
        }

        const vector<string>& colNames = catTableP->GetColumnNames();

        for (unsigned int itemI = 0; itemI < colNames.size(); ++itemI)
        {
            if (!IsItemDefinedInRef(catNames[catI], colNames[itemI],
              *itemTableP))
            {
                log << "ERROR - In block \"" << block.GetName() <<
                  "\", in category \"" << catNames[catI] << "\", item \"" <<
                  colNames[itemI] << "\" is not defined in the " <<
                  "reference file" << endl;
            }
        }

        if (skipCatCheck)
            continue;

        CheckKeyItems(block.GetName(), *catTableP, *catKeyTableP, log);

        CheckMandatoryItems(block.GetName(), *catTableP, *itemTableP, log);

        vector<string> itemNames;
        GetItemNames(itemNames, *catTableP);

        ret = CheckParentChild(block, *catTableP, itemNames,
          *itemLinkedTableP, log);

#ifdef VLAD_PERF
        cout << "Finished parent/child for category: " << catNames[catI] << endl;
#endif
#ifdef VLAD_DEBUG
        if (ret == 5)
            return(ret);
#endif
        for (unsigned int itemI = 0; itemI < itemNames.size(); ++itemI)
        {
#ifdef VLAD_PERF
            cout << "Processing item: " << itemI << " of " <<
              itemNames.size() << endl;
#endif
            ret = CheckRegExpRangeEnum(block, *catTableP, itemNames[itemI],
               *itemTypeTableP, *itemTypeListTableP, *itemRangeTableP,
               *itemEnumTableP, *itemLinkedTableP, log);
        }
        itemNames.clear();
    }

    catNames.clear();
 
    return(ret);

}


// This method converts the string with embedded escape sequences into
// the real string. For example, a two character combination "\t" is
// converted into one character representing horizontal tab.
void CifFile::ConvertEscapedString(const string& inString,
  string& outString)
{

    outString.clear();

    outString.reserve(inString.size());

    // Go through the characters of the input string
    for (unsigned int posI = 0; posI < inString.size(); )
    {
        if (inString[posI] == '\\')
        {
            if (inString[posI + 1] == '\\')
            {
                  outString.push_back('\\');
                  posI += 2;
            }
            else if (inString[posI + 1] == 't')
            {
                  outString.push_back('\t');
                  posI += 2;
            }
            else if (inString[posI + 1] == 'n')
            {
                  outString.push_back('\n');
                  posI += 2;
            }
            else
            {
                  outString.push_back(inString[posI]);
                  posI++;
            }
        } // Current character is '\'
        else
        {
            outString.push_back(inString[posI]);
            posI++;
        } // Curent character is not '\'
    } // for (characters in the input string)

} // CifFile::ConvertEscapedString()


void CifFile::GetItemNames(vector<string>& itemNames, ISTable& isTable)
{

    itemNames.clear();

    string tableName = isTable.GetName();

    const vector<string>& colNames = isTable.GetColumnNames();

    itemNames.reserve(colNames.size());

    string cifItem;

    for (unsigned int i = 0; i < colNames.size(); i++)
    {
        CifString::MakeCifItem(cifItem, tableName, colNames[i]);

        itemNames.push_back(cifItem);
    }

}


bool CifFile::IsCatDefinedInRef(const string& catName, ISTable& catTable)
{

    /*
    ** For a category, method looks into dictionary ("category" table)
    ** to find out whether it exists or not.
    */

    vector<string> keyTarget;
    keyTarget.push_back(catName);

    vector<string> keyList;
    keyList.push_back("id");

    unsigned int catIndex = catTable.FindFirst(keyTarget, keyList);
    if (catIndex == catTable.GetNumRows())
    {
        return(false);
    }

    return(true);
   
}


bool CifFile::IsItemDefinedInRef(const string& catName, const string& itemName,
  ISTable& refItemTable)
{

    /*
    ** For an item, method looks into dictionary ("item" table)
    ** to find out whether the item is defined in the category or not.
    */

    bool found = false;

    vector<string> refItemList;
    refItemList.push_back("category_id");
    refItemList.push_back("name");

    vector<string> refItemTarget;
    refItemTarget.push_back(catName);

    string cifItem;
    CifString::MakeCifItem(cifItem, catName, itemName);

    refItemTarget.push_back(cifItem);

    unsigned int index = refItemTable.FindFirst(refItemTarget, refItemList);

    if (index != refItemTable.GetNumRows())
    {
        found = true;
    }

    return(found);

}


void CifFile::CheckKeyItems(const string& blockName, ISTable& catTable,
  ISTable& catKeyTable, ostringstream& log)
{

    /*
    ** For a category, method looks into dictionary ("category_key" table)
    ** to find out what its key items are. It checks for existence of key
    ** items and checks if there are duplicate key values.
    */

    vector<string> keyTarget;
    keyTarget.push_back(catTable.GetName());

    vector<string> keyList;
    keyList.push_back("id");

    vector<unsigned int> OutList;
    catKeyTable.Search(OutList, keyTarget, keyList);
    if (OutList.empty())
    {
        log << "ERROR - In block \"" << blockName << "\", no key items " <<
          "found in category \"" << catTable.GetName() << "\"" << endl;
        return;
    }

    bool keyNotFound = false;

    vector<string> keyColumns;

    for (unsigned int k = 0; k < OutList.size(); k++)
    {
        string itemName;

        CifString::GetItemFromCifItem(itemName, catKeyTable(OutList[k],
          "name"));

        keyColumns.push_back(itemName);

        if (!catTable.IsColumnPresent(itemName))
        {
            log << "ERROR - In block \"" << blockName << "\", key item \"" <<
              itemName << "\" not found in category \"" <<
              catTable.GetName() << "\"" << endl;
            keyNotFound = true;
        }
    }

    if (!keyNotFound)
    {
        vector<pair<unsigned int, unsigned int> > duplRows;
        catTable.FindDuplicateRows(duplRows, keyColumns, true);
        if (!duplRows.empty())
        {
            for (unsigned int rowI = 0; rowI < duplRows.size(); rowI++)
            {
                bool report = true;
                for (unsigned int keyI = 0; keyI < keyColumns.size(); ++keyI)
                {
                    const string& cell = catTable(duplRows[rowI].first,
                      keyColumns[keyI]);
                 
                    if ((cell == ".") || (cell == "?") || (cell.empty()))
                    {
                        report = false;
                        break;
                    }
                }

                if (!report)
                    continue;

                log << "ERROR - In block \"" << blockName << "\", in " <<
                  "cateogory \"" << catTable.GetName() <<
                  "\", values for key item(s):" << endl;

                for (unsigned int keyI = 0; keyI < keyColumns.size(); ++keyI)
                {
                    log << "  \"" << keyColumns[keyI] << "\"," << endl; 
                }

                log << "  in row #" << duplRows[rowI].first <<
                  " are repeated in row #" << duplRows[rowI].second << endl;
            }
        }
    }

}


void CifFile::CheckMandatoryItems(const string& blockName, ISTable& catTable,
  ISTable& refItemTable, ostringstream& log)
{

    /*
    ** For a category, method looks into dictionary ("item" table)
    ** to find out what its mandatory items are. Then it checks for existence
    ** of those items in the category.
    */

    vector<string> refItemList;
    refItemList.push_back("category_id");
    refItemList.push_back("mandatory_code");

    vector<string> refItemTarget;
    refItemTarget.push_back(catTable.GetName());
    refItemTarget.push_back("yes");

    vector<unsigned int> OutList;
    refItemTable.Search(OutList, refItemTarget, refItemList);
    if (OutList.empty())
    {
        refItemTarget[1] = "implicit";
 
        refItemTable.Search(OutList, refItemTarget, refItemList);
        if (OutList.empty())
        {
            log << "ERROR - In block \"" << blockName <<
              "\", no mandatory items found in category \"" <<
              catTable.GetName() << "\"" << endl;
        }

        return;
    }

    for (unsigned int k = 0; k < OutList.size(); ++k)
    {
        string cell = refItemTable(OutList[k], "name");
        string itemName;
        CifString::GetItemFromCifItem(itemName, cell);
        if (!catTable.IsColumnPresent(itemName))
        {
            log << "ERROR - In block \"" << blockName <<
              "\", mandatory item \"" << itemName <<
              "\" is not in category \"" << catTable.GetName() <<
              "\"" << endl;
        }
    }

}


int CifFile::CheckParentChild(Block& block, ISTable& catTable,
  const vector<string>& cifItemNames, ISTable& parChildTable, ostringstream& log)
{

    int ret = 1;

    map<string, unsigned int> maxKeyGroups;

    ISTable* keysTableP = CreateKeysTable(cifItemNames, parChildTable,
      maxKeyGroups);

#ifdef VLAD_DEBUG
    if (VladDebug) 
        cout << "Keys Table: " << endl << (*keysTableP) << endl;
#endif

    if (keysTableP->GetNumRows() == 0)
    {
        // No parents. Just return.
        return(ret);
    }


    for (map<string, unsigned int>::iterator pos = maxKeyGroups.begin();
      pos != maxKeyGroups.end(); ++pos)
    {
        // pos->first is the parent category name
        // pos->second is the maximum group number for the parent

        if (!block.IsTablePresent(pos->first))
        {
            // Parent table does not exist. Ignore this fact.
            continue;
        }

        ISTable* parentCatTableP = block.GetTablePtr(pos->first);
        if (parentCatTableP->GetNumRows() == 0)
        {
            // Parent table has no rows. Ignore this fact.
            continue;
        }

        vector<vector<string> > comboKeys;
        vector<string> parKeys;

        GetComboKeys(pos->first, pos->second, *keysTableP, comboKeys,
          parKeys);

#ifdef VLAD_DEBUG
        if (VladDebug) 
        {
            for (unsigned int ind1 = 0; ind1 < comboKeys.size(); ++ind1)
            {
                cout << "Combo key at index " << ind1 << endl;
                for (unsigned int ind2 = 0; ind2 < comboKeys[ind1].size();
                  ++ind2)
                {
                    cout << "  Subkey at index " << ind2 << " is: " <<
                      comboKeys[ind1][ind2] << endl;
                }
                cout << endl;
            }
        }
#endif

        vector<string> parKeyItems;
        for (unsigned int parKeyItemsI = 0; parKeyItemsI < parKeys.size();
          ++parKeyItemsI)
        {
            string tmpItem;
            CifString::GetItemFromCifItem(tmpItem, parKeys[parKeyItemsI]);
            parKeyItems.push_back(tmpItem);
        }

        // Check existence in the parent
        for (unsigned int parKeyI = 0; parKeyI < parKeyItems.size(); ++parKeyI)
        {
            if (!parentCatTableP->IsColumnPresent(parKeyItems[parKeyI]))
            {
                log << "ERROR - In block \"" << block.GetName() <<
                  "\", parent key \"" << parKeyItems[parKeyI] <<
                  "\" is not in parent category \"" << pos->first <<
                  "\"" << endl;
                ret = 0;

                delete(keysTableP);

                return(ret);
            }
        }
        
        for (unsigned int comboKeyI = 0; comboKeyI < comboKeys.size();
          ++comboKeyI)
        {
#ifdef VLAD_PERF
            cout << "Processing parent \"" << pos->first << "\" combo key: " << comboKeyI << " of " <<
              comboKeys.size() << endl;
#endif
            for (unsigned int l = 0; l < catTable.GetNumRows(); ++l)
            {
#ifdef VLAD_PERF
                cout << "  Processing combo key row: " << l << " of " <<
                  catTable.GetNumRows() << endl;
#endif
                vector<string> parKeyValues;

                for (unsigned int keyI = 0; keyI < comboKeys[comboKeyI].size();
                  ++keyI)
                {
                    string itemName;
                    CifString::GetItemFromCifItem(itemName,
                      comboKeys[comboKeyI][keyI]);

                    // loop for all rows
                    const string& parKeyValue = catTable(l, itemName);
                    if ((parKeyValue != ".") && (parKeyValue != "?") &&
                      (!parKeyValue.empty()))
                    {
                        parKeyValues.push_back(parKeyValue);
                    }
                    else
                    {
                        parKeyValues.clear();
                        break;
                    }
                } // For every key item in a combo key

                if (parKeyValues.empty())
                    continue;

                unsigned int searchIndex =
                  parentCatTableP->FindFirst(parKeyValues, parKeyItems);
                if (searchIndex == parentCatTableP->GetNumRows())
                {
#ifdef VLAD_DEBUG
                    log << "Category Table:" << endl <<
                      catTable << endl;

                    log << "Parent Category Table:" << endl <<
                      (*parentCatTableP) << endl;
#endif
                    log << "ERROR - In block \"" << block.GetName() <<
                      "\"," << endl << "  value(s) of child key item(s) " <<
                      "were not found in the parent foreign key " << endl <<
                      "  item(s) (child key=parent key=value):" << endl;

                    for (unsigned int keyValI = 0; keyValI <
                      parKeyValues.size(); ++keyValI)
                    {
                        log << "  \"" << comboKeys[comboKeyI][keyValI] <<
                          "\"=\"" << parKeys[keyValI] << "\"=\"" <<
                          parKeyValues[keyValI] << "\"" << endl;
                    }
#ifdef VLAD_DEBUG
                    return(5);
#endif

                    ret = 0;
                }
            } // For every row in child
        } // For every combo key for one parent
    } // For every parent

    delete(keysTableP);

    return(ret);

}


int CifFile::CheckRegExpRangeEnum(Block& block, ISTable& catTable,
  const string& cifItemName, ISTable& itemTypeTable, 
  ISTable& itemTypeListTable, ISTable& itemRangeTable,
  ISTable& itemEnumTable, ISTable& parChildTable, ostringstream& log)
{

    /************** REGEX **************/
    /* 
    ** For an item, a check is being made whether that item is in _item_type
    ** table. If it is not, method ends. If it is, its "code" is taken
    ** from _item_type table and regular expression (from "construct" item
    ** in item_type_list category) table for that "code" is taken and verified
    ** if it is a proper regular expression. Then every value for that item
    ** (that is of regular expression kind) is checked against that regular
    ** expression.
    */

    /************** RANGE **************/
    /* 
    ** Cell type is first determinted and then floating range check or
    ** integer range check is performed. For each found range, it is
    ** checked if cell value is withing that range. If one of the limits
    ** is set to unknown or any, the range against that limit is not checked.
    */

    /************** ENUMERATION **************/
    /* 
    ** Cell type is first determinted and then floating enumeration check or
    ** integer enumeration check is performed. Cell value is checked against
    ** all enumerated values. If value is not one of the enumerated values,
    ** check fails.
    */


    /* compile a regex */
    string patternString;
    regex_t preg;
#define NS 10
    regmatch_t pmatch[NS];
    int icmp = 0;
    int ret = 1;
    vector<string> maxlist;
    vector<string> minlist;
    int errCodeRange = -1;
    vector<string> enumlist;
    int errCodeEnumeration = -1;

    int errCodeRegex = -1;
    vector<string> target;
    target.push_back(cifItemName);

    vector<string> nameList;
    nameList.push_back("name");

    string typeCode;
    GetItemTypeCode(typeCode, cifItemName, itemTypeTable);

    if (typeCode.empty())
    {
        return(1);
    }

    /************** REGEX step 1**************/
    errCodeRegex = 0;

    vector<string> ItemTypeLTarget;
    ItemTypeLTarget.push_back(typeCode);

    vector<string> ItemTypeLList;
    ItemTypeLList.push_back("code");

    unsigned int iOut = itemTypeListTable.FindFirst(ItemTypeLTarget,
      ItemTypeLList);

    string primCode = itemTypeListTable(iOut, "primitive_code");

    ConvertEscapedString(itemTypeListTable(iOut, "construct"),
      patternString); 

    icmp = rcsb_regcomp(&preg, patternString.c_str(), REG_EXTENDED);

    rcsb_regfree(&preg);


    /************** RANGE step 1**************/
    vector<unsigned int> OutList;
    itemRangeTable.Search(OutList, target, nameList);
    if (!OutList.empty())
    {
        errCodeRange = 0;
        for (unsigned int l = 0; l < OutList.size(); ++l)
        {
            maxlist.push_back(itemRangeTable(OutList[l], "maximum"));
            minlist.push_back(itemRangeTable(OutList[l], "minimum"));
        }
    }

    /************ ENUMERATION step 1************/
    itemEnumTable.Search(OutList, target, nameList);
    if (!OutList.empty())
    {
        errCodeEnumeration = 0;
        for (unsigned int l = 0; l < OutList.size(); ++l)
        {
            enumlist.push_back(itemEnumTable(OutList[l], "value"));
        }
    }

    string itemName;
    CifString::GetItemFromCifItem(itemName, cifItemName);
    for (unsigned int l = 0; l < catTable.GetNumRows(); l++)
    {
#ifdef VLAD_PERF
                cout << "  Processing regex row: " << l << " of " <<
                  catTable.GetNumRows() << endl;
#endif
        // loop for all rows
        const string& cell = catTable(l, itemName);

        /*************** REGEX step 2**************/
        if ((!cell.empty()) && (cell != ".") && (cell != "?"))
        {
            if ((errCodeRegex >= 0) && (icmp == 0))
            {
                rcsb_regcomp(&preg, patternString.c_str(), REG_EXTENDED);
                ret = rcsb_regexec(&preg, cell.c_str(), NS, pmatch, 0);
                rcsb_regfree(&preg);
                if (ret != 0)
                {
                    log << "ERROR - In block \"" << block.GetName() <<
                       "\", value \"" << cell << "\" for \"" <<
                      cifItemName << "\" does not match the pattern." << endl;
                    ret = 0;
                }
                else
                {
                    int len = pmatch[0].rm_eo - pmatch[0].rm_so;
                    if ((len != (int)(cell.size())) && (typeCode != "text"))
                    {
                        log << "ERROR - In block \"" << block.GetName() <<
                          "\", value \"" << cell << "\" for \"" <<
                          cifItemName << "\" does not match the pattern; " <<
                          endl;
                        log << "        Only \"" <<
                          cell.substr(pmatch[0].rm_so) << "\" is matched." <<
                          endl;
                        ret = 0;
                    }
                }
            }

            if (errCodeRange >= 0)
            {
                ret = CheckCellRange(cell, typeCode, minlist, maxlist);
                if (ret == 0)
                {
                    log << "ERROR - In block \"" << block.GetName() <<
                      "\", value \"" << cell << "\" for \"" << cifItemName <<
                      "\" is out of range" << endl;
                }
            }

            if (errCodeEnumeration >= 0)
            {
                ret = CheckCellEnum(cell, typeCode, primCode, enumlist);
                if (ret == 0)
                {
                    log << "ERROR - In block \"" << block.GetName() <<
                      "\", value \"" << cell << "\" for \"" << cifItemName <<
                      "\" is not defined in the enumeration list" << endl;
                }
            }
        }
    }

    return(ret);
}


int CifFile::CheckCellRange(const string& cell, const string& typeCode, 
  const vector<string>& minlist, const vector<string>& maxlist)
{

    int matched = 1;

    if (typeCode == "float")
    {
        matched = CheckCellFloatRange(cell, minlist, maxlist);
    }
    else
    {
        if (typeCode == "int")
        {
            matched = CheckCellIntRange(cell, minlist, maxlist);
        }
    }

    return(matched);

}


int CifFile::CheckCellEnum(const string& cell, const string& typeCode,
  const string& primCode, const vector<string>& enumlist)
{

    int matched = 1;

    if (typeCode == "float")
    {
        matched = CheckCellFloatEnum(cell, enumlist);
    }
    else
    {
        if (typeCode == "int")
        {
            matched = CheckCellIntEnum(cell, enumlist);
        }
        else
        {
            matched = CheckCellOtherEnum(cell, primCode, enumlist);
        }
    }

    return(matched);

}


int CifFile::CheckCellFloatRange(const string& cell,
  const vector<string>& minlist, const vector<string>& maxlist)
{

    int matched = 0;

    double dmin = 0;
    double dmax = 0;

    double dval = String::StringToDouble(cell);

    unsigned int m = 0;
    while (m < minlist.size() && !matched)
    {

        bool checkMin = false;
        if ((minlist[m] != "?") && (minlist[m] != ".") &&
          (!minlist[m].empty()))
        {
            checkMin = true;
            dmin = String::StringToDouble(minlist[m]);
        }

        bool checkMax = false;
        if ((maxlist[m] != "?") && (maxlist[m] != ".") &&
          (!maxlist[m].empty()))
        {
            checkMax = true;
            dmax = String::StringToDouble(maxlist[m]);
        }

        if ((!checkMin) && (checkMax))
        {
            if (dval < dmax)
                matched = 1;
        }

        if ((!checkMax) && (checkMin))
        {
            if (dval > dmin)
                matched = 1;
        }

        if (checkMax && checkMin)
        {
            if (dmin == dmax)
            {
                if (dval == dmax)
                    matched = 1;
            }
            else
            {
                if ((dval > dmin) && (dval < dmax))
                    matched = 1;
            }
        }

        m++;
    }

    return(matched);

}


int CifFile::CheckCellIntRange(const string& cell,
  const vector<string>& minlist, const vector<string>& maxlist)
{

    int matched = 0;

    int imin = 0;
    int imax = 0;

    int ival = String::StringToInt(cell);

    unsigned int m = 0;
    while (m < minlist.size() && !matched)
    {

        bool checkMin = false;
        if ((minlist[m] != "?") && (minlist[m] != ".") &&
          (!minlist[m].empty()))
        {
            checkMin = true;
            imin = String::StringToInt(minlist[m]);
        }

        bool checkMax = false;
        if ((maxlist[m] != "?") && (maxlist[m] != ".") &&
          (!maxlist[m].empty()))
        {
            checkMax = true;
            imax = String::StringToInt(maxlist[m]);
        }

        if ((!checkMin) && (checkMax))
        {
            if (ival < imax)
                matched = 1;
        }

        if ((!checkMax) && (checkMin))
        {
            if (ival > imin)
                matched = 1;
        }

        if (checkMax && checkMin)
        {
            if (imin == imax)
            {
                if (ival == imax)
                    matched = 1;
            }
            else
            {
                if ((ival > imin) && (ival < imax))
                    matched = 1;
            }
        }

        m++;
    }

    return(matched);

}


int CifFile::CheckCellFloatEnum(const string& cell,
  const vector<string>& enumlist)
{

    double de;
    double dval;

    int matched = 0;

    dval = String::StringToDouble(cell);

    unsigned int m = 0;
    while (m < enumlist.size() && !matched)
    {
        de = String::StringToDouble(enumlist[m]);
        if (dval == de)
            matched = 1;
        m++;
    }

    return(matched);

}


int CifFile::CheckCellIntEnum(const string& cell,
  const vector<string>& enumlist)
{

    int matched = 0;

    int ival = String::StringToInt(cell);

    unsigned int m = 0;
    while (m < enumlist.size() && !matched)
    {
        int ie = String::StringToInt(enumlist[m]);
        if (ival == ie)
            matched = 1;
        m++;
    }

    return(matched);

}


int CifFile::CheckCellOtherEnum(const string& cell, const string& primCode,
  const vector<string>& enumlist)
{

    int matched = 0;

    unsigned int m = 0;
    while (m < enumlist.size() && !matched)
    {
        if (primCode == "uchar")
        {
            if (String::IsCiEqual(cell, enumlist[m]))
                matched = 1;
        }
        else
        {
            if (cell == enumlist[m])
                matched = 1;
        }
        m++;
    }

    return(matched);

}


ISTable* CifFile::CreateKeysTable(const vector<string>& cifItemNames,
  ISTable& parChildTable, map<string, unsigned int>& maxKeyGroups)
{

    // Prepare the table that will contain all the needed keys information
    // This table will have four columns: child key CIF item, group number
    // parent key CIF item and parent category name

    ISTable* keysTableP = new ISTable();

    keysTableP->AddColumn("childKeyCifItem");
    keysTableP->AddColumn("keyGroup");
    keysTableP->AddColumn("parKeyCifItem");
    keysTableP->AddColumn("parCategory");

    vector<string> childNameCol;
    childNameCol.push_back("child_name");

    vector<string> keyList;
    keyList.push_back("parKeyCifItem");


    for (unsigned int cifItemI = 0; cifItemI < cifItemNames.size();
      ++cifItemI)
    {
        vector<string> childCifItem;
        childCifItem.push_back(cifItemNames[cifItemI]);

        vector<unsigned int> parLoc;

        parChildTable.Search(parLoc, childCifItem, childNameCol);

        if (parLoc.empty())
        {
            // This item is not a key
            continue;
        }

        for (unsigned int parLocI = 0; parLocI < parLoc.size(); ++parLocI)
        {
            string parCatName;

            string parCifItemName = parChildTable(parLoc[parLocI],
              "parent_name");
            CifString::GetCategoryFromCifItem(parCatName, parCifItemName);

            vector<string> keyTarget;
            keyTarget.push_back(parCifItemName);

            vector<unsigned int> parents;

            keysTableP->Search(parents, keyTarget, keyList);

            unsigned int newKeyGroupInt = 1;
            if (!parents.empty())
            {
                newKeyGroupInt = String::StringToInt((*keysTableP)(
                  parents[parents.size() - 1], "keyGroup"));
                ++newKeyGroupInt;
            }

            if (newKeyGroupInt > maxKeyGroups[parCatName])
                maxKeyGroups[parCatName] = newKeyGroupInt;

            string newKeyGroup = String::IntToString(newKeyGroupInt);

            vector<string> row;

            row.push_back(cifItemNames[cifItemI]);
            row.push_back(newKeyGroup);
            row.push_back(parCifItemName);
            row.push_back(parCatName);

            keysTableP->AddRow(row);
        } // For all item's parent items (that can be in different parents)
    } // For all items in the category

    return(keysTableP);

}


void CifFile::GetComboKeys(const string& parCatName,
  const unsigned int maxKeyGroup, ISTable& keysTable,
  vector<vector<string> >& comboKeys, vector<string>& parKeys)
{

    // First determine the order of keys
    // Look for all keys that have this parent and which group number is 1
    // Return result will have keys that may have single or multipe occurences

    vector<string> keyList;
    keyList.push_back("keyGroup");
    keyList.push_back("parCategory");

    vector<string> keyTarget;
    keyTarget.push_back("1");
    keyTarget.push_back(parCatName);

    vector<string> groupList;
    groupList.push_back("keyGroup");
    groupList.push_back("parKeyCifItem");

    vector<unsigned int> groupOne;

    keysTable.Search(groupOne, keyTarget, keyList);

    vector<string> firstComboKey;

    // Fill in the first combo key and find out the order of parent keys
    for (unsigned int itemI = 0; itemI < groupOne.size(); ++itemI)
    {
        firstComboKey.push_back(keysTable(groupOne[itemI],
          "childKeyCifItem"));

        parKeys.push_back(keysTable(groupOne[itemI],
          "parKeyCifItem"));
    }

    comboKeys.push_back(firstComboKey);

    for (unsigned int groupI = 2; groupI <= maxKeyGroup; ++groupI)
    {
        vector<string> nextComboKey;

        for (unsigned int parKeyI = 0; parKeyI < parKeys.size(); ++parKeyI)
        {
            vector<string> groupTarget;
            groupTarget.push_back(String::IntToString(groupI));
            groupTarget.push_back(parKeys[parKeyI]);

            unsigned int nextGroupIndex;

            nextGroupIndex = keysTable.FindFirst(groupTarget, groupList);
            if (nextGroupIndex == keysTable.GetNumRows())
            {
                // Use it from the first combo key
                nextComboKey.push_back(firstComboKey[parKeyI]);
            }
            else
            {
                nextComboKey.push_back(keysTable(nextGroupIndex,
                  "childKeyCifItem"));
            }
        }
        comboKeys.push_back(nextComboKey);
    }

}


void CifFile::CheckAndRectifyItemTypeCode(Block& block, ostringstream& log)
{

    ISTable* parChildTableP = block.GetTablePtr("item_linked");

    ISTable* itemTypeTableP = block.GetTablePtr("item_type");

    ISTable* itemTableP = block.GetTablePtr("item");


    if ((parChildTableP == NULL) || (itemTypeTableP == NULL) ||
      (itemTableP == NULL))
        return;

#ifdef VLAD_DEBUG
    if (block.GetName() == "mmcif_pdbx.dic")
        log << "Item Table for Debugging: " << endl << (*itemTableP) << endl;
#endif
    for (unsigned int cifItemI = 0; cifItemI < itemTableP->GetNumRows();
      ++cifItemI)
    {
        const string& cifItemName = (*itemTableP)(cifItemI, "name");

        // Get item's type code
        string itemTypeCode;
        bool hasItemTypeCode = false;
        vector<string> target;
        target.push_back(cifItemName);

        vector<string> nameList;
        nameList.push_back("name");

        unsigned int iOut = itemTypeTableP->FindFirst(target, nameList);
        if (iOut != itemTypeTableP->GetNumRows())
        {
            itemTypeCode = (*itemTypeTableP)(iOut, "code");
            hasItemTypeCode = true;
        }

        // Check item's parent's type code
        string parTypeCode;

        bool hasParTypeCode = false;

        vector<string> childCifItem;
        childCifItem.push_back(cifItemName);

        vector<string> childNameCol;
        childNameCol.push_back("child_name");

        bool hasParents = true;

        vector<unsigned int> parLoc;

        parChildTableP->Search(parLoc, childCifItem, childNameCol);

        if (parLoc.empty())
        {
            // This item does not have any parents
            hasParents = false;
        }

        if (parLoc.size() > 1)
        {
            cout << "BIG TROUBLE - CHILD HAS MULTIPLE PARENTS" << endl;
            exit(1);
        }

        for (unsigned int parLocI = 0; parLocI < parLoc.size(); ++parLocI)
        {
            target.clear();
            target.push_back((*parChildTableP)(parLoc[parLocI], "parent_name"));

            unsigned int iOut = (*itemTypeTableP).FindFirst(target, nameList);
            if (iOut != itemTypeTableP->GetNumRows())
            {
                hasParTypeCode = true;
                parTypeCode = (*itemTypeTableP)(iOut, "code");
                break;
            }
        }

        if (!hasParents && !hasItemTypeCode)
        {
            // No type code in either the item or the parent
            log << "ERROR - In block \"" << block.GetName() <<
              "\", \"" << cifItemName << "\"" << 
              " does not have item type code defined" << endl;
            continue;
        }

        if ((!hasParTypeCode) && (!hasItemTypeCode))
        {
            log << "ERROR - In block \"" << block.GetName() <<
              "\", parent item \"" << target[0] << "\"" <<
              " of item \"" << cifItemName << "\"," <<
              " does not have item type code defined" << endl;
            continue;
        }

        if (hasItemTypeCode)
        {
            if (hasParTypeCode && (itemTypeCode != parTypeCode))
            {
                log << "ERROR - In block \"" << block.GetName() <<
                  "\", child item \"" << cifItemName <<
                  "\" has item type code \"" << itemTypeCode <<
                  "\", while its parent \"" << target[0] <<
                  "\" has item type code \"" << parTypeCode << "\"" << endl;
            }
        }
        else
        {
            // Add it to the "item_type" table of the item
            itemTypeTableP->AddRow();
            unsigned int lastRowInd = itemTypeTableP->GetNumRows() - 1;

            itemTypeTableP->UpdateCell(lastRowInd, "name",
              cifItemName);
            itemTypeTableP->UpdateCell(lastRowInd, "code",
              parTypeCode);

            // VLAD - ADD INFO MESSAGE HERE THAT WE ARE MODIFYING AFTER
            //  DIC PARSING
#ifdef VLAD_TRACE_IMPLEMENT
            log << "TRACE - In block \"" << block.GetName() <<
              "\", adding type code \"" << parTypeCode << "\" for " <<
              "item \"" << cifItemNames << "\"" << endl;
#endif
        }
    }

}


void CifFile::GetItemTypeCode(string& typeCode, const string& cifItemName, 
  ISTable& itemTypeTable)
{

    typeCode.clear();

    vector<string> target;
    target.push_back(cifItemName);

    vector<string> nameList;
    nameList.push_back("name");

    unsigned int iOut = itemTypeTable.FindFirst(target, nameList);
    if (iOut != itemTypeTable.GetNumRows())
    {
        typeCode = itemTypeTable(iOut, "code");
    }

}
