/*
FILE:     DDLParserBase.C
*/
/*
VERSION:  3.004
*/
/*
DATE:     9/11/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


/*!
** \file DDLParserBase.C
**
** \brief Implementation file for DDLParser class.
*/


/* 
  PURPOSE:    A DDL 2.1 compliant CIF file parser.
*/

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "GenString.h"
#include "CifString.h"
#include "DDLScannerBase.h"
#include "DDLParserInt.h"
#include "DDLParserBase.h"


extern FILE* ddlparser_in;
extern int ddlparser_parse();
extern void ddlparser_restart(FILE*);

string Glob_tBufKeywordSaveDDL;
char* Glob_pBufValueDDL;
char* Glob_dataBlockNameDDL;

DDLParser* DDLParserP = NULL;

DDLParser::DDLParser(DdlFile* fo, bool verbose)
{

    if (DDLParserP != NULL)
    {
        // Attempting to create a new parser, during the lifetime of
        // an existing parser.
        throw AlreadyExistsException("Cannot create a new parser, since "\
          "one already exists.", "DDLParser::DDLParser");
    }

    Clear();

    if (fo != NULL)
        _fobj = fo;
    else
        throw EmptyValueException("fo is a NULL pointer",
          "DDLParser::DDLParser");

    _verbose=verbose;
    format=fo->GetFormatTable();
    errorLog.clear();
    OpenLog("DDLParser.log",_verbose);

    DDLParserP = this;
}

void DDLParser::Parse(const string& fileName, string& diagnostics)
{
    diagnostics.clear();

    FILE* ddlIn;

    if ((ddlIn = fopen(fileName.c_str(), "r")) == NULL )
    {
        ndb_log_message(NDB_MSG_DEBUG4,"Unable to open file %s", fileName.c_str());
        diagnostics = "Unable to open file.";
        return;
    }

    ddlparser_in = ddlIn;

    ddlparser_restart(ddlparser_in);

    ddlparser_parse();

    fclose(ddlIn);

    if (this->errorLog.size() > 0)
    {
        diagnostics = this->errorLog;
    }
}

void ddlparser_error(const char *s)
{
    DDLParserP->Error(s);
}

DDLParser::~DDLParser()
{
    Reset();
    DDLParserP = NULL;
}


void DDLParser::Error(const char *s)
/* 
 * Purpose:  yyerror() Print errors in DDLScanner log.
 */
{
  errorLog += s;
  errorLog += " near line ";
  errorLog += String::IntToString(NDBlineNo);
  errorLog += '\n';
  log << s << " near line " <<  NDBlineNo << endl;
} 

void DDLParser::Reset()
{
  if (_tbl && (_curItemNo > 0)) { // write the current table / management of _tbl by _fobj 
      if (_nTablesInBlock) {
      Block& block = _fobj->GetBlock(_curDataBlockName);
      block.WriteTable(_tbl);
      } else {
      Block& block = _fobj->GetBlock(_prevDataBlockName);
      block.WriteTable(_tbl);
      }
      _nTablesInBlock++;
  }
  _fieldList.clear();
  _fieldListSave.clear();
  if (_savetbl != NULL)
  {
      delete _savetbl;
      _savetbl = NULL;
  }
}
void DDLParser::Clear()
{
  _nTablesInBlock=0;
  _numDataBlocks=0;
  _afterLoop = false;
  _curItemNo=0;
  _curRow=0;
  _curValueNo=0;
  _fieldListAlloc=100;
  _fieldList.reserve(_fieldListAlloc);
  _pBufValue.clear();
  _fobj=NULL;
  _tbl=NULL;
  isSave = 0;
  _nTablesInBlockSave=0;
  _curItemNoSave=0;
  _saveobj=NULL;
  _savetbl=NULL;
  _fieldListAllocSave=100;
  _fieldListSave.reserve(_fieldListAllocSave);
  _uString = ".";
  _mString = CifString::NullValue;
  _tBufKeyword.clear();
  _curCategoryName.clear();
  _curDataBlockName = "MISSING";
  _prevDataBlockName = "MISSING";
  _curCategoryNameSave.clear();
  _curDataBlockNameSave.clear();
  _prevDataBlockNameSave.clear();
  vector<string> list;
/* table format keeps information about what tables are item related, what
   category related, and what datablock related
   column dbName has datablock name in it, since file can have more then one
   datablock, column type has type of table such as datablock, category or
   item; and column catName has table name such as dictionary_history,
   item_type_list... - datablock related; category, category_key... - category
   related, item_type, item_linked... - item related
   This table is nessery to write ddl in a save frame format
*/
}


void DDLParser::ProcessLoopDeclaration(void)
/* ----------------------------------------------------------------------
     Purpose: DDLParser::ProcessLoopDeclaration(void)

              Handles initialization for a new loop, by creating a new
              category and adding the current item name to this category.
   ---------------------------------------------------------------------- */
{
  string categoryName;

#if DEBUG
  if (_verbose) log << "Processing loop declaration at line " << NDBlineNo << " value " <<_tBufKeyword << endl;
#endif
  _afterLoop = true;
  if (! CifString::GetCategoryFromCifItem(categoryName, _tBufKeyword) ) {
    log << "Error in category name at line " << NDBlineNo << " value " << _tBufKeyword << endl;
    errorLog += "Error in category name at line ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += " value ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
  }

  Block& block = _fobj->GetBlock(_curDataBlockName);

  if (block.IsTablePresent(categoryName)) {
    if (!isSave) {
      log << "Warning - Duplicate category name " << categoryName << " at line " << NDBlineNo << endl;
      /* this is just a warning - rows are concatenated, this is
         happens only out of save frame, and in that case isSave=0.
         Because of that "else" is removed.
      */
    }
    if (_tbl) { // write the current table / management of _tbl by _fobj
      if (_nTablesInBlock) {
        Block& block = _fobj->GetBlock(_curDataBlockName);
        block.WriteTable(_tbl);
      } else {
        Block& block = _fobj->GetBlock(_prevDataBlockName);
        block.WriteTable(_tbl);
      }
      _nTablesInBlock++;
    }

    Block& block = _fobj->GetBlock(_curDataBlockName);

    _tbl = block.GetTablePtr(categoryName);

    _curRow = _tbl->GetNumRows();
    ProcessItemNameList();
    _curCategoryName = categoryName;
  } else {
    if (_tbl) { // write the current table / management of _tbl by _fobj
      if (_nTablesInBlock) {
        Block& block = _fobj->GetBlock(_curDataBlockName);
        block.WriteTable(_tbl);
        } else {
          Block& block = _fobj->GetBlock(_prevDataBlockName);
          block.WriteTable(_tbl);
        }
      _nTablesInBlock++;
    }
    _tbl = new ISTable(categoryName);
    _curRow = 0;
    _curCategoryName = categoryName;
    ProcessItemNameList();
  }
}
void DDLParser::ProcessItemNameList(void)
/* ----------------------------------------------------------------------
   Purpose: DDLParser::ProcessItemNameList(void)

            Registers the item keyword for the the current item in the
            current category.  Maintains an index array of "valid" keyword
            names in fieldList[].  This array is used to indirectly
            reference between keywords and values ...
 * ----------------------------------------------------------------------*/
{

  string keywordName;
  string categoryName;

#if DEBUG
  if (_verbose) log << "Processing item name list at line " <<  NDBlineNo << " keyword " <<  _tBufKeyword << endl;
#endif

  if (! CifString::GetCategoryFromCifItem(categoryName, _tBufKeyword) ) {
    log << "Error in category name at line " << NDBlineNo << " value " << _tBufKeyword << endl;
    errorLog += "Error in category name at line ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += " value ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
  }

  if ( _curItemNo > _fieldListAlloc - 1) {
    _fieldListAlloc = _curItemNo + _fieldListAlloc;
    _fieldList.reserve(_fieldListAlloc);
  }

  vector<string> colNames, target;
  colNames.push_back("dbName");
  colNames.push_back("type");
  colNames.push_back("catName");
  target.push_back(_curDataBlockName);
  if (isSave == 0) {
    target.push_back("data");
  }
  if (isSave == 1) {
    target.push_back("category");
  }
  if (isSave == 2) {
    target.push_back("item");
  }
  target.push_back(categoryName);
  unsigned int resIndex = format->FindFirst(target, colNames);
  if (resIndex == format->GetNumRows()) {
    format->AddRow(target);
  }

    if (isSave) {
  if ( (categoryName == _curCategoryName) && CifString::GetItemFromCifItem(keywordName, _tBufKeyword)) {
#if DEBUG
    if (_verbose) log << "Line " << NDBlineNo << " keyword is " << keywordName << endl;
#endif
    if (!_tbl->IsColumnPresent(keywordName))
        _tbl->AddColumn(keywordName);
    if (_curItemNo >= (int)_fieldList.size())
      _fieldList.push_back(keywordName);
    else
      _fieldList[_curItemNo] = keywordName;
  }
  else {
    CifString::GetItemFromCifItem(keywordName, _tBufKeyword);
    if (_curItemNo >= (int)_fieldList.size())
      _fieldList.push_back(keywordName);
    else
      _fieldList[_curItemNo] = keywordName;
  }
    }
    else {
      _curCategoryName = categoryName;
  if ( (categoryName == _curCategoryName) && CifString::GetItemFromCifItem(keywordName, _tBufKeyword)) {
    if (_tbl->IsColumnPresent(keywordName)) {
      log << "Warning - Duplicate item name " <<_tBufKeyword  << " at line " << NDBlineNo << endl;
    }
/*    else*/ {
#if DEBUG
    if (_verbose) log << "Line " << NDBlineNo << " keyword is " << keywordName << endl;
#endif
    _tbl->AddColumn(keywordName);

    if (_curItemNo >= (int)_fieldList.size())
      _fieldList.push_back(keywordName);
    else
      _fieldList[_curItemNo] = keywordName;

    }
  } else {
    log << "Syntax error at line " << NDBlineNo << " at item " << _tBufKeyword << endl;
    errorLog += "*Syntax error at line ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += " item ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
    _fieldList[_curItemNo].clear();
  }
    }
  _curItemNo++;

}
void DDLParser::ProcessValueList(void)
/* ----------------------------------------------------------------------
     Purpose:  DDLParser::ProcessValueList(void)

               Add the current value to the appropriate column in the
               the current row.  Start a new row if necessary.
 * ----------------------------------------------------------------------*/
{
#if DEBUG
  if (_verbose) {
    if (!_pBufValue.empty())
      log << "Processing value at line " << NDBlineNo << " value [" << _pBufValue << "]" << endl;
  }
#endif

  if (!_fieldList[_curValueNo].empty()) {

    if (_curValueNo == 0)
    {
      vector<string> rowBuf(_tbl->GetNumColumns(), String::Empty);
      _curRow++;
      _tbl->AddRow(rowBuf);
    }

    if (!_pBufValue.empty())
    {
         _tbl->UpdateCell(_curRow-1, _fieldList[_curValueNo], _pBufValue);
    }
    else
    {
         _tbl->UpdateCell(_curRow-1, _fieldList[_curValueNo], _mString);
    }
    
  }
  _curValueNo++;

  if (_curValueNo == _curItemNo) {
#if DEBUG
  if (_verbose) {
    log << "Loading row " << _curRow -1 << " with " <<  rowBuf.size() << " elements" << endl;
    for (int i=0; i < rowBuf.size(); i++) {
      log << "Column [" << i << "] value "<< rowBuf[i].c_str() << endl;
    }
  }
#endif
    _curValueNo = 0;

  }
}
void DDLParser::ProcessItemValuePair(void)
/* ----------------------------------------------------------------------
      Purpose: ndb_cif_process_item_name_value_pair()

               Assign the current value to its associated item name.
 * ----------------------------------------------------------------------*/
{
  string categoryName;
  string keywordName;

  _curItemNo  = 1;
  _curValueNo = 0;
#if DEBUG
  if (_verbose)  {
    if (!_pBufValue.empty())
      log << "Processing " << _tBufKeyword << " at " <<  NDBlineNo << " value " << _pBufValue  << endl;
  }
#endif

  if (! CifString::GetCategoryFromCifItem(categoryName, _tBufKeyword) ) {
    log << "Error in category name at line " << NDBlineNo << " value " << _tBufKeyword << endl;
    errorLog += "Error in category name at line ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += " value ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
    return;
  }
  vector<string> colNames, target;
  colNames.push_back("dbName");
  colNames.push_back("type");
  colNames.push_back("catName");
  target.push_back(_curDataBlockName);
  if (isSave == 0) {
    target.push_back("data");
  }
  if (isSave == 1) {
    target.push_back("category");
  }
  if (isSave == 2) {
    target.push_back("item");
  }
  target.push_back(categoryName);
  unsigned int resIndex = format->FindFirst(target, colNames);
  if (resIndex == format->GetNumRows()) {
    format->AddRow(target);
  }

  Block& block = _fobj->GetBlock(_curDataBlockName);

  if (block.IsTablePresent(categoryName)) { //  duplicates a persistent table?
    if (!isSave) {
      log << "Warning - Duplicate category name " << categoryName << " at line " << NDBlineNo << endl;
    }
  if ((categoryName != _curCategoryName)|| _afterLoop) {
    if (_tbl) { // write the current table / management of _tbl by _fobj
      if (_nTablesInBlock) {
        Block& block = _fobj->GetBlock(_curDataBlockName);
        block.WriteTable(_tbl);
      } else {
        Block& block = _fobj->GetBlock(_prevDataBlockName);
        block.WriteTable(_tbl);
      }
      _nTablesInBlock++;
    }
      Block& block = _fobj->GetBlock(_curDataBlockName);

      _tbl = block.GetTablePtr(categoryName);

      _tbl->AddRow();
      _curCategoryName = categoryName;
  }
  _afterLoop=false;

  _curRow = _tbl->GetNumRows();
  CifString::GetItemFromCifItem(keywordName, _tBufKeyword);
  if (!_tbl->IsColumnPresent(keywordName)) {
    {
      // Created a block to be able to re-use emptyColumn
    vector<string> emptyColumn;
    for (int i=0; i<_curRow; i++)
      emptyColumn.push_back(String::Empty);
    _tbl->AddColumn(keywordName);
    _tbl->FillColumn(keywordName, emptyColumn);
    emptyColumn.clear();
    }
  }

  if (!_pBufValue.empty()) {
    _tbl->UpdateCell(_curRow-1, keywordName, _pBufValue);
  } else {
    _tbl->UpdateCell(_curRow-1, keywordName, _mString);
  }

  }
  else {
  if (categoryName != _curCategoryName) {
    if (_tbl) { // write the current table / management of _tbl by _fobj
      if (_nTablesInBlock) {
        Block& block = _fobj->GetBlock(_curDataBlockName);
        block.WriteTable(_tbl);
      } else {
        Block& block = _fobj->GetBlock(_prevDataBlockName);
        block.WriteTable(_tbl);
      }
      _nTablesInBlock++;
    }
    _tbl = new ISTable(categoryName);
    _curCategoryName = categoryName;
  }

  if ( CifString::GetItemFromCifItem(keywordName, _tBufKeyword) ) {
#if DEBUG
    if (_verbose) log << "Line " << NDBlineNo << " keyword is " << keywordName << endl;
#endif
    _tbl->AddColumn(keywordName);
  }
  else {
    log << "Syntax error line " << NDBlineNo << " at item " << _tBufKeyword << endl;
    errorLog += "Syntax2 error at line ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += " item ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
  }

  if (!_pBufValue.empty())
  {
    if (_tbl->GetNumRows() == 0)
        _tbl->AddRow();
    
    _tbl->UpdateCell(_tbl->GetNumRows() - 1, keywordName, _pBufValue);
  }
  else
  {
    if (_tbl->GetNumRows() == 0)
        _tbl->AddRow();
    
    _tbl->UpdateCell(_tbl->GetNumRows() - 1, keywordName, _mString);
  }
#ifdef VLAD_DEBUG
  cout << "VLAD-TABLE-CONTENT" << endl << (*_tbl) << endl;
#endif
  }  
}

/* These methods are the same as methods in CifParser; rules are the same,
   except these methods are applied to temporary _saveobj object to examine
   correctness of save frame
*/
//******************************

void DDLParser::ProcessLoopDeclarationSave(void)
/* ----------------------------------------------------------------------
     Purpose: DDLParser::ProcessLoopDeclarationSave(void)

              Handles initialization for a new loop, by creating a new
              category and adding the current item name to this category.
   ---------------------------------------------------------------------- */
{
  string categoryName;

  _curItemNoSave = 0;  _curValueNoSave = 0; 
  if (! CifString::GetCategoryFromCifItem(categoryName, _tBufKeyword) ) {
    log << "Error in category name at line (save frame) " << NDBlineNo << " value " << _tBufKeyword << endl;
    errorLog += "Error in category name at line (save frame) ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += " value ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
  }

  bool tablePresent = false;

  try
  {
      Block& block = _saveobj->GetBlock(_curDataBlockNameSave);

      tablePresent = block.IsTablePresent(categoryName);
  }
  catch (NotFoundException)
  {
      tablePresent = false;
  }

  if (tablePresent) {
    log << "Duplicate category name in a save frame " << categoryName << " at line " << NDBlineNo << endl;
    errorLog += "Duplicate category name in a save frame ";
    errorLog += categoryName;
    errorLog += " at line ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += '\n';
  } else {
    if (_savetbl) { // write the current table /
      if (_nTablesInBlockSave) {
      try {
          Block& block = _saveobj->GetBlock(_curDataBlockNameSave);
          block.WriteTable(_savetbl);
          _savetbl=NULL;
          }
          catch (NotFoundException)
          {
          } 
      } else {
      try {
          Block& block = _saveobj->GetBlock(_prevDataBlockNameSave);
          block.WriteTable(_savetbl);
          _savetbl=NULL;
          }
          catch (NotFoundException)
          {
          }
      }
      _nTablesInBlockSave++;
    }
    if (_savetbl != NULL)
    {
        delete _savetbl;
        _savetbl = NULL;
    }
    _savetbl = new ISTable(categoryName);
    _curRowSave = 0;
    _curCategoryNameSave = categoryName;
    ProcessItemNameListSave();
  }
}

void DDLParser::ProcessItemNameListSave(void)
/* ----------------------------------------------------------------------
   Purpose: DDLParser::ProcessItemNameListSave(void)

            Registers the item keyword for the the current item in the
            current category.  Maintains an index array of "valid" keyword
            names in fieldList[].  This array is used to indirectly
            reference between keywords and values ...
 * ----------------------------------------------------------------------*/
{

  string keywordName;
  string categoryName;

  if (! CifString::GetCategoryFromCifItem(categoryName, _tBufKeyword) ) {
    log << "Error in category name at line (save frame) " << NDBlineNo << " value " << _tBufKeyword << endl;
    errorLog += "Error in category name at line (save frame) ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += " value ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
  }

  if ( _curItemNoSave > _fieldListAllocSave - 1) {
    _fieldListAllocSave = _curItemNoSave + _fieldListAllocSave;
    _fieldListSave.reserve(_fieldListAllocSave);
  }

  if ( (categoryName == _curCategoryNameSave) && CifString::GetItemFromCifItem(keywordName, _tBufKeyword)) {
    _savetbl->AddColumn(keywordName);

    if (_curItemNoSave >= (int)_fieldListSave.size())
      _fieldListSave.push_back(keywordName);
    else
      _fieldListSave[_curItemNoSave] = keywordName;

  } else {
    log << "Syntax error at line (save frame) " << NDBlineNo << " at item " << _tBufKeyword << endl;
    errorLog += "Syntax error at line (save frame) ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += " item ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
    _fieldListSave[_curItemNoSave].clear();
  }
  _curItemNoSave++;

}
void DDLParser::ProcessValueListSave(void)
/* ----------------------------------------------------------------------
     Purpose:  DDLParser::ProcessValueListSave(void)

               Add the current value to the appropriate column in the
               the current row.  Start a new row if necessary.
 * ----------------------------------------------------------------------*/
{

  if (!_fieldListSave[_curValueNoSave].empty()) {

    if (_curValueNoSave == 0)
    {
      vector<string> rowBufSave(_savetbl->GetNumColumns(), String::Empty);
      _curRowSave++;
      _savetbl->AddRow(rowBufSave);
    }
    if (!_pBufValue.empty())
    {
         _savetbl->UpdateCell(_curRowSave-1, _fieldListSave[_curValueNoSave],
           _pBufValue);
    } else {
         _savetbl->UpdateCell(_curRowSave-1, _fieldListSave[_curValueNoSave],
           _mString);
      }
  }
  _curValueNoSave++;

  if (_curValueNoSave == _curItemNoSave) {
    _curValueNoSave = 0;
  }
}





void DDLParser::ProcessItemValuePairSave(void)
/* ----------------------------------------------------------------------
      Purpose: ProcessItemValuePairSave(void)

               Assign the current value to its associated item name.
 * ----------------------------------------------------------------------*/
{
  string categoryName;
  string keywordName;

  _curItemNoSave  = 1;
  _curValueNoSave = 0;

  if (! CifString::GetCategoryFromCifItem(categoryName, _tBufKeyword) ) {
    log << "Error in category name at line (save frame) " << NDBlineNo << " value " << _tBufKeyword << endl;
    errorLog += "Error in category name at line (save frame) ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += " value ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
    return;
  }

  bool tablePresent = false;

  try
  {
      Block& block = _saveobj->GetBlock(_curDataBlockNameSave);

      tablePresent = block.IsTablePresent(categoryName);
  }
  catch (NotFoundException)
  {
      tablePresent = false;
  }

  if (tablePresent) { //  duplicates a persistent table?
    log << "Duplicate category name in a save frame " << categoryName << " at line " << NDBlineNo << endl;
    errorLog += "Duplicate category name in a save frame ";
    errorLog += categoryName;
    errorLog += " at line ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += '\n';
    return;
  }

  if (categoryName != _curCategoryNameSave) {

    if (_savetbl) { // write the current table
      if (_nTablesInBlockSave) {
      try {
          Block& block = _saveobj->GetBlock(_curDataBlockNameSave);
          block.WriteTable(_savetbl);
          _savetbl=NULL;
          }
          catch (NotFoundException)
          {
          }
      } else {
      try {
          Block& block = _saveobj->GetBlock(_prevDataBlockNameSave);
          block.WriteTable(_savetbl);
          _savetbl=NULL;
          }
          catch (NotFoundException)
          {
          }
      }
      _nTablesInBlockSave++;
    }
    if (_savetbl != NULL)
    {
        delete _savetbl;
        _savetbl = NULL;
    }
    _savetbl = new ISTable(categoryName);
    _curCategoryNameSave = categoryName;
  }
  if ( CifString::GetItemFromCifItem(keywordName, _tBufKeyword) ) {
    _savetbl->AddColumn(keywordName);
  }
  else {
    log << "Syntax error line (save frame) " << NDBlineNo << " at item " << _tBufKeyword << endl;
    errorLog += "Syntax error at line (save frame) ";
    errorLog += String::IntToString(NDBlineNo);
    errorLog += " item ";
    errorLog += _tBufKeyword;
    errorLog += '\n';
  }
  if (!_pBufValue.empty())
  {
    if (_savetbl->GetNumRows() == 0)
        _savetbl->AddRow();
    
    _savetbl->UpdateCell(_savetbl->GetNumRows() - 1, keywordName, _pBufValue);
  }
  else
  {
    if (_savetbl->GetNumRows() == 0)
        _savetbl->AddRow();
    
    _savetbl->UpdateCell(_savetbl->GetNumRows() - 1, keywordName, _mString);
  }

}

void DDLParser::ProcessAssignments(void)
{
  if (_curValueNo!=0)
  log <<"Number of data values is not exact multiples of the number of data names   (look above line)"<<NDBlineNo<<endl;

}

void DDLParser::ProcessOneAssignment(void)
{
/* isSave is set to 1 if the parser is in category save frame, to 2
   if parser in item save frame, and to 0 if there is no save frame
   Sytax checking is done into save frames for temporary Tbl object;
*/
  if (isSave)
    ProcessItemValuePairSave();
  ProcessItemValuePair();
}

void DDLParser::ProcessItemNameListLoop(void)
{
 if (isSave)
    ProcessLoopDeclarationSave();
  ProcessLoopDeclaration();
}

void DDLParser::ProcessItemNameListName(void)
{
  if (isSave)
    ProcessItemNameListSave();
  ProcessItemNameList();
}

void DDLParser::ProcessValueListItem(void)
{
  if (isSave)
    ProcessValueListSave();
  ProcessValueList();
}

void DDLParser::ProcessItemName(void)
{
  _tBufKeyword = Glob_tBufKeywordSaveDDL;
}

void DDLParser::ProcessLoop(void)
{
    _curItemNo = 0;  _curValueNo = 0; 
    _curItemNoSave = 0;  _curValueNoSave = 0; 
}

void DDLParser::ProcessItemValue(void)
{
    _pBufValue = Glob_pBufValueDDL;
}

void DDLParser::ProcessLsItemValue(void)
{
    _pBufValue = Glob_pBufValueDDL;
}

void DDLParser::ProcessUnknownValue(void)
{
   _pBufValue = _uString;
}

void DDLParser::ProcessMissingValue(void)
{
   _pBufValue = _mString;
}

void DDLParser::ProcessSaveBegin(void)
{
/* making a new Tbl object for save frame; save frame has same meaning
   as datablock in cif data file; this is actually only for save frame
   syntax checking and this information is temporary just for that chacking
*/
  _nTablesInBlockSave=1;
  _saveobj = new DdlFile();
  _prevDataBlockNameSave = _curDataBlockNameSave;
  _curDataBlockNameSave = &(Glob_dataBlockNameDDL)[5];
  _savetbl=NULL;
  _pBufValue = Glob_dataBlockNameDDL;
}

void DDLParser::ProcessSaveEnd(void)
{
  // Destroy all the tables
  vector<string> blockNames;
  _saveobj->GetBlockNames(blockNames);
  for (unsigned int blockI = 0; blockI < blockNames.size(); ++blockI)
  {
      Block& block = _saveobj->GetBlock(blockNames[blockI]);
      vector<string> tableNames;
      block.GetTableNames(tableNames);
      for (unsigned int tableI = 0; tableI < tableNames.size(); ++tableI)
      {
          ISTable* tableP = block.GetTablePtr(tableNames[tableI]);
          if (tableP != NULL)
              delete tableP;
      }
  }
  delete _saveobj;
  if (_savetbl != NULL)
  {
      delete _savetbl;
      _savetbl = NULL;
  }
 _pBufValue = Glob_pBufValueDDL;
}

void DDLParser::ProcessDataBlockName(void)
{
  _numDataBlocks++;
  _nTablesInBlock=0;
  if (&(Glob_dataBlockNameDDL)[5] && (strlen(&(Glob_dataBlockNameDDL)[5])>0)) {
    _curDataBlockName = &(Glob_dataBlockNameDDL)[5];
    _fobj->AddBlock(_curDataBlockName);
  } else {
    char bCount[10];
    sprintf(bCount,"%d",_numDataBlocks);
    _curDataBlockName = "UNNAMED-";
    _curDataBlockName += bCount;
  }
  if (_prevDataBlockName == "MISSING")   _prevDataBlockName = _curDataBlockName;

  if (_tbl && (_curItemNo > 0)) { // write the current table / management of _tbl by _fobj
      Block& block = _fobj->GetBlock(_prevDataBlockName);
      block.WriteTable(_tbl);
      _nTablesInBlock++;
      _tbl=NULL;
  }
  if (_prevDataBlockName != _curDataBlockName)
         _curCategoryName.clear();

#if DEBUG
  if (_verbose) log << " Previous data block is now  " <<  _prevDataBlockName << endl;
  if (_verbose) log << " Current  data block is now  " <<  _curDataBlockName << endl;
#endif
}

void ProcessAssignmentsFromDDLParser()
{
    DDLParserP->ProcessAssignments();
}

void ProcessOneAssignmentFromDDLParser()
{
    DDLParserP->ProcessOneAssignment();
}

void ProcessItemNameListLoopFromDDLParser()
{
    DDLParserP->ProcessItemNameListLoop();
}

void ProcessItemNameListNameFromDDLParser()
{
    DDLParserP->ProcessItemNameListName();
}

void ProcessValueListItemFromDDLParser()
{
    DDLParserP->ProcessValueListItem();
}

void ProcessItemNameFromDDLParser()
{
    DDLParserP->ProcessItemName();
}

void ProcessLoopFromDDLParser()
{
    DDLParserP->ProcessLoop();
}

void ProcessItemValueFromDDLParser()
{
    DDLParserP->ProcessItemValue();
}

void ProcessLsItemValueFromDDLParser()
{
    DDLParserP->ProcessLsItemValue();
}

void ProcessUnknownValueFromDDLParser()
{
    DDLParserP->ProcessUnknownValue();
}

void ProcessMissingValueFromDDLParser()
{
    DDLParserP->ProcessMissingValue();
}

void ProcessSaveBeginFromDDLParser()
{
    DDLParserP->ProcessSaveBegin();
}

void ProcessSaveEndFromDDLParser()
{
    DDLParserP->ProcessSaveEnd();
}

void ProcessDataBlockNameFromDDLParser()
{
    DDLParserP->ProcessDataBlockName();
}

