/*
FILE:     ndb_sysutil.C
*/
/*
VERSION:  3.004
*/
/*
DATE:     9/11/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <dirent.h>

#include "ndb_misclib.h"


#define	MAXLINE	4096			/* max line length */

#define LOCK_POLL_INTERVAL  10           /* In seconds. */
#define STRUC_LOCK_FILE_EXT ".lock"



#if 0
static void	err_doit(int, const char *, va_list);
#endif

/*char	*pname = NULL;		 caller can set this from argv[0] */



void ndb_execute_command(char *cmdstring)
/*
 * Execute the command string under the csh interpreter and wait for completion
 */
{
  int status;
  pid_t pid;

  /*
   * Get a child process.
   */
  
  if ((pid = fork()) < 0) {
    ndb_log_message(NDB_MSG_ERR,(char*)"Error occurred during fork() %s",strerror(errno));
    exit(1);
  }
  
  /*
     * The child executes the code inside the if.
     */
  if (pid == 0) {
    execl("/bin/csh", "csh", "-f", "-c", cmdstring, (char *) 0 ) ;
    ndb_log_message(NDB_MSG_ERR,(char*)"Error occurred during execl() %s",strerror(errno));
    exit(1);
  }
  
  /*
   * The parent executes the wait.
   */
  while (wait(&status) != pid)
    /* empty */ ;
}


int lock_reg(int fd, int cmd, int type, off_t offset, int whence, off_t len)
{
	struct flock	lock;

	lock.l_type = type;		/* F_RDLCK, F_WRLCK, F_UNLCK */
	lock.l_start = offset;	/* byte offset, relative to l_whence */
	lock.l_whence = whence;	/* SEEK_SET, SEEK_CUR, SEEK_END */
	lock.l_len = len;		/* #bytes (0 means to EOF) */

	return( fcntl(fd, cmd, &lock) );
}


pid_t lock_test(int fd, int type, off_t offset, int whence, off_t len)
{
	struct flock	lock;

	lock.l_type = type;		/* F_RDLCK or F_WRLCK */
	lock.l_start = offset;	/* byte offset, relative to l_whence */
	lock.l_whence = whence;	/* SEEK_SET, SEEK_CUR, SEEK_END */
	lock.l_len = len;		/* #bytes (0 means to EOF) */

	if (fcntl(fd, F_GETLK, &lock) < 0)
		ndb_log_message(NDB_MSG_ERR,(char*)"fcntl error");

	if (lock.l_type == F_UNLCK)
		return(0);		/* false, region is not locked by another proc */
	return(lock.l_pid);	/* true, return pid of lock owner */
}


int lock_file(char* fileNameP, int maxWaitTime, int* retCodeP)
{

    /*
    ** fileNameP - absolute file name. Must start with a '/'.
    ** maxWaitTime - maximum wait time in seconds.
    ** retCodeP - pointer to the return code.
    */

    char* lastForwSlashP = NULL;
    int dirNameSize;
    char* dirNameP = NULL;
    DIR* dirP = NULL;
    int fd;
    struct flock slock;
    int lockObtainedB;
    int waitTime;


    /* Assume success. */
    *retCodeP = LOCK_NO_ERROR;

    /* Get the directory where file is located and try to access it. */
    lastForwSlashP = strrchr(fileNameP, '/');
    if (lastForwSlashP == NULL)
    {
        /* No forward slash found. */
        *retCodeP = LOCK_DIR_ACCESS_ERROR;
        return(-1);
    }

    dirNameSize = lastForwSlashP - fileNameP;
    dirNameP = (char*)malloc(dirNameSize + 1);
    if (dirNameP == NULL)
    {
        /* Memory allocation failure. */
        *retCodeP = LOCK_DIR_ACCESS_ERROR;
        return(-1);
    }

    strncpy(dirNameP, fileNameP, dirNameSize);
    dirNameP[dirNameSize] = '\0';

    /* Try to open the directory. */
    dirP = opendir(dirNameP);

    free(dirNameP);

    if (dirP == NULL)
    {
        /* Cannot access the directory. */
        *retCodeP = LOCK_DIR_ACCESS_ERROR;
        return(-1);
    }
    else
    {
        closedir(dirP);
    }

    /* Open the file. */
    fd = open(fileNameP, O_RDWR);

    if (fd == -1)
    {
        /* Cannot access the file. */
        *retCodeP = LOCK_FILE_ACCESS_ERROR;
        return(-1);
    }

    /*
    ** This code gets the lock on the file.
    ** It is concluded that flock() does not work if a file is located on
    ** GPFS/NFS mounted file system. The workaround is to use fcntl()
    ** function. However, fcntl() also does not work properly with GPFS/NFS
    ** if blocking is desired (using F_SETLKW), but works if non-blocking
    ** is desired (using F_SETLK). So the code uses non-blocking call to
    ** fcntl() in conjunction with polling, in order to simulate the
    ** blocking call.
    */

    /* Get exclusive write lock. */
    slock.l_type = F_WRLCK;
    slock.l_whence = SEEK_SET;
    slock.l_start = 0;
    slock.l_len = 0;
    slock.l_pid = 0;

    /* Set the flag that the lock is not yet obtained. */
    lockObtainedB = 0;

    /* Initialize overall wait time. */
    waitTime = 0;

    do
    {
        /* Try to get the lock, without blocking. */
        if ((fcntl(fd, F_SETLK, &slock)) == 0)
        {
            /* Set the flag that the lock is obtained. */
            lockObtainedB = 1;
        }
        else
        {
            /* Wait and try to get the lock again. */
            sleep(LOCK_POLL_INTERVAL);
            waitTime += LOCK_POLL_INTERVAL;
        }
    }
    while ((lockObtainedB == 0) && (waitTime < maxWaitTime));
    /* while (lock not obtained and wait time less than max time) */

    if (lockObtainedB == 0)
    {
        close(fd);

        /* Timeout. */
        *retCodeP = LOCK_TIMEOUT_ERROR;
        return(-1);
    }

    /* Lock obtained. */
    *retCodeP = LOCK_NO_ERROR;

    return(fd);

} /* End of lock_file() */


int unlock_file(int fd)
{

    struct flock slock;
    int ret;


    /* Prepare locking structure for unlocking. */
    slock.l_type = F_UNLCK;
    slock.l_whence = SEEK_SET;
    slock.l_start = 0;
    slock.l_len = 0;
    slock.l_pid = 0;

    /* Release the lock. */
    ret = fcntl(fd, F_SETLK, &slock);
    if (ret == -1)
    {
        return(LOCK_UNLOCK_ERROR);
    }

    close(fd);

    return(LOCK_NO_ERROR);

} /* End of unlock_file() */


int lock_file_aux(char* fileNameP, int maxWaitTime, int* retCodeP)
{

    int size;
    char* lockFileNameP = NULL;
    int fd;
    int ret;


    if (fileNameP == NULL)
    {
        /* No file name. */
        *retCodeP = LOCK_FILE_ACCESS_ERROR;
        return(-1);
    }

    size = strlen(fileNameP) + strlen(STRUC_LOCK_FILE_EXT) + 1;

    lockFileNameP = (char*)malloc(size);
    if (lockFileNameP == NULL)
    {
        /* No memory. */
        *retCodeP = LOCK_FILE_ACCESS_ERROR;
        return(-1);
    }

    strcpy(lockFileNameP, fileNameP);
    strcat(lockFileNameP, STRUC_LOCK_FILE_EXT);

    fd = open(lockFileNameP, (O_WRONLY | O_CREAT | O_TRUNC),
      (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH));
    if (fd == -1)
    {
        free(lockFileNameP);

        /* Cannot create lock file. */
        *retCodeP = LOCK_FILE_ACCESS_ERROR;
        return(-1);
    }

    close(fd);

    ret = lock_file(lockFileNameP, maxWaitTime, retCodeP);

    free(lockFileNameP);

    return(ret);

} /* End of lock_file_aux() */


#if 0
/* Nonfatal error related to a system call.
 * Print a message and return. */

void err_ret(const char *fmt, ...)
{
	va_list		ap;

	va_start(ap, fmt);
	err_doit(1, fmt, ap);
	va_end(ap);
	return;
}

/* Fatal error related to a system call.
 * Print a message and terminate. */

void err_sys(const char *fmt, ...)
{
	va_list		ap;

	va_start(ap, fmt);
	err_doit(1, fmt, ap);
	va_end(ap);
	exit(1);
}

/* Fatal error related to a system call.
 * Print a message, dump core, and terminate. */

void err_dump(const char *fmt, ...)
{
	va_list		ap;

	va_start(ap, fmt);
	err_doit(1, fmt, ap);
	va_end(ap);
	abort();		/* dump core and terminate */
	exit(1);		/* shouldn't get here */
}

/* Nonfatal error unrelated to a system call.
 * Print a message and return. */

void err_msg(const char *fmt, ...)
{
	va_list		ap;

	va_start(ap, fmt);
	err_doit(0, fmt, ap);
	va_end(ap);
	return;
}

/* Fatal error unrelated to a system call.
 * Print a message and terminate. */

void err_quit(const char *fmt, ...)
{
	va_list		ap;

	va_start(ap, fmt);
	err_doit(0, fmt, ap);
	va_end(ap);
	exit(1);
}

/* Print a message and return to caller.
 * Caller specifies "errnoflag". */

static void err_doit(int errnoflag, const char *fmt, va_list ap)
{
	int		errno_save;
	char	buf[MAXLINE];

	errno_save = errno;		/* value caller might want printed */
	vsprintf(buf, fmt, ap);
	if (errnoflag)
		sprintf(buf+strlen(buf), ": %s", strerror(errno_save));
	strcat(buf, "\n");
	fflush(stdout);		/* in case stdout and stderr are the same */
	fputs(buf, stderr);
	fflush(NULL);		/* flushes all stdio output streams */
	return;
}
#endif
