#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include "TableFile.h"
#include "CifFile.h"
#include "CifParserBase.h"
#include "match.h"
#include "util.h"


struct cryst
{
    char *space_group, *a, *b, *c;
    char *alpha, *beta, *gamma;
}cell_unit;


struct coord
{
    char **group_pdb, **atomnum, **atomname, **alt;
    char **resname, **chainID, **resnum, **x;
    char **y, **z, **occup, **biso, **footnote;
    char **auth_asym, **pdbx_PDB_model_num;
    
}xyz;

extern char METHOD[10];

int strcmp_case (const char *s1, const char *s2);
int strncmp_case (const char *s1, const char *s2, int n);


void cif2pdb(char *iFile, char *oFile)
// only for the purpose of getting cells and residue names
{

    char symm[80];
    string diags, blockId;
    int i=0,  j=0;
    long n0=0,n1=0,n2=0,n3=0,n4=0,n5=0,n6=0;
    long n7=0,n8=0,n9=0, n10=0, n11=0, n12=0, natom=0;

    CifFile * fobjR = NULL;
    FILE *fp, *fout;
    
    if((fp = fopen(iFile, "r"))==NULL){        
        printf("Can not open the file %s (routine:is_category)\n",iFile);
        return ;
    }
    if((fout = fopen(oFile, "w"))==NULL){        
        printf("Can not write the file %s (routine:is_category)\n",oFile);
        return ;
    }

  
    CifParser* cifParserR = NULL;

    fobjR = new CifFile(true);

    cifParserR = new CifParser(fobjR, fobjR->GetVerbose());

    cifParserR->Parse(iFile, diags);

    delete(cifParserR);

    if (!diags.empty()) {
        fprintf(stderr, "Diags for file %s: %s\n",iFile,diags.c_str());
//        exit(0);
    }
    
    blockId = fobjR->GetFirstBlockName();

    xyz.group_pdb=get_attribute_values(fobjR,blockId.c_str(),"atom_site", "group_PDB", &n1);
    xyz.atomnum = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "id", &n2);
    xyz.atomname = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "label_atom_id", &n3);
    xyz.alt = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "label_alt_id", &n4);
    xyz.resname = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "label_comp_id", &n5);
    xyz.chainID = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "label_asym_id", &n6);
    xyz.resnum = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "auth_seq_id", &n7);
    xyz.x = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "Cartn_x", &n8  );
    xyz.y = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "Cartn_y", &n9  );
    xyz.z = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "Cartn_z", &n10 );
    xyz.occup = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "occupancy", &n11 );
    xyz.biso = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "B_iso_or_equiv", &n12 );
    xyz.pdbx_PDB_model_num = get_attribute_values(fobjR, blockId.c_str(), "atom_site", "pdbx_PDB_model_num", &n0);


    if(n2<1 || n3<1 || n5<1 || n7<1 ||  n8<1 || n9<1 || n10<1) {
        printf("mmCIF to PDB conversion error! please check your input mmCIF file.\n");
        exit(0);
    }
    
    if(strcmp_case(METHOD, "NMR")){
        cell_unit.a = get_attribute_value(fobjR, blockId.c_str(), "cell", "length_a");
        cell_unit.b = get_attribute_value(fobjR, blockId.c_str(), "cell", "length_b");
        cell_unit.c = get_attribute_value(fobjR, blockId.c_str(), "cell", "length_c");
        cell_unit.alpha = get_attribute_value(fobjR, blockId.c_str(), "cell", "angle_alpha");
        cell_unit.beta = get_attribute_value(fobjR, blockId.c_str(), "cell", "angle_beta");
        cell_unit.gamma = get_attribute_value(fobjR, blockId.c_str(), "cell", "angle_gamma");
        cell_unit.space_group = get_attribute_value(fobjR, blockId.c_str(), "symmetry", "space_group_name_H-M");
        (cell_unit.space_group)?strcpy(symm, cell_unit.space_group):strcpy(symm," ");
        

        if(cell_unit.a && cell_unit.b && cell_unit.c && cell_unit.alpha && cell_unit.beta && cell_unit.gamma){
            fprintf(fout,"CRYST1%9.3f%9.3f%9.3f%7.2f%7.2f%7.2f %10s \n",
                    atof(cell_unit.a) , atof(cell_unit.b) , atof(cell_unit.c) ,
                    atof(cell_unit.alpha) , atof(cell_unit.beta) , atof(cell_unit.gamma),symm);
        }
    }
      
    for(i=0; i<n8; i++){
      if(n0>1){// model id exist
          if(xyz.pdbx_PDB_model_num[i][0] != '.' && xyz.pdbx_PDB_model_num[i][0] != '?'){
             j=atoi(xyz.pdbx_PDB_model_num[i]); 
             if(j!=1) continue; //only keep model 1
          }
          
      }
      
      (n1>0) ? fprintf(fout,"ATOM  ") :  fprintf(fout,"ATOM  ");
      (n2>0) ? fprintf(fout,"%5s ", xyz.atomnum[i])  : fprintf(fout,"%5s ", " ");

      if((n3>0 && strlen(xyz.atomname[i])==4 ) ){
          (n3>0 && strlen(xyz.atomname[i])==4 )
              ? fprintf(fout,"%4s", xyz.atomname[i])  : fprintf(fout,"%4s", " ");
      }else if((n3>0 && strlen(xyz.atomname[i])==3 ) ){
          (n3>0 && strlen(xyz.atomname[i])==3 )
              ? fprintf(fout," %3s", xyz.atomname[i])  : fprintf(fout,"%4s", " ");
          
      }else if((n3>0 && strlen(xyz.atomname[i])==2 ) ){
          (n3>0 && strlen(xyz.atomname[i])==2 )
              ? fprintf(fout," %2s ", xyz.atomname[i])  : fprintf(fout,"%4s", " ");
          
      }else if((n3>0 && strlen(xyz.atomname[i])==1 ) ){
          (n3>0 && strlen(xyz.atomname[i])==1 )
              ? fprintf(fout," %1s  ", xyz.atomname[i])  : fprintf(fout,"%4s", " ");
          
      }
      
      (n4>0 && xyz.alt[i][0] !='.' && xyz.alt[i][0] !='?' )
          ? fprintf(fout,"%1s", xyz.alt[i])  : fprintf(fout,"%1s", " ");
      (n5>0) ? fprintf(fout,"%3s ", xyz.resname[i])  : fprintf(fout,"%4s", " ");
      (n6>0 && xyz.chainID[i][0] !='.' && xyz.chainID[i][0] !='?' )
          ? fprintf(fout,"%1s", xyz.chainID[i])  : fprintf(fout,"%1s", " ");
          
      if (n7>0 && strlen(xyz.resnum[i])<=4 ){
          (n7>0 && strlen(xyz.resnum[i])<=4 )
              ? fprintf(fout,"%4s ",xyz.resnum[i]) : fprintf(fout,"%5s", " ");
      }else if (n7>0 && strlen(xyz.resnum[i])>4 ){
          (n7>0 && strlen(xyz.resnum[i])>4 )
              ? fprintf(fout,"%5s",xyz.resnum[i])  : fprintf(fout,"%5s", " ");
      }
      
      (n8>0 && n9>0 && n10>0)
          ? fprintf(fout,"   %8s%8s%8s", xyz.x[i], xyz.y[i], xyz.z[i])  : fprintf(fout,"%29s", " ");
       
      (n11>0&& n12>0) ? fprintf(fout,"%6s%6s",  xyz.occup[i], xyz.biso[i]): fprintf(fout,"%12s", " ");
 
      fprintf(fout,"\n");
      natom++;
      
  }
  fclose(fp);
  fclose(fout);
  if (fobjR) delete fobjR;
  if(natom<2) {
      printf("Problem for mmcif to PDB conversion. Please check you input mmcif file!\n");
      exit(0);
  }
  

}




           
