#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "rcsb_math.h"
#include "structure.h"
#include "util.h"
#include "harv.h"


extern AXIS axis;
extern EXPTL_CRYSTAL exptl_crystal ;
extern PDBX_CRYSTAL_ALIGNMENT pdbx_crystal_alignment;

void extract_HKL_log_index( char *inpfile )
// Extracting index information  from LOG files of HKL or SCALEPACK 
{
    char str[256], tmp1[80], tmp2[80];
    int  n=0, n1=0, n2=0,n3=0;
    float a1, a2, a3;
    FILE *fp;
    

    if((fp = fopen(inpfile, "r"))==NULL) {
        printf("Can not open the file (%s)\n", inpfile);
        return ;
    }
    
    while (fgets(str, sizeof str, fp) != NULL) {
        n = sscanf(str,"%s%s%f%f%f", tmp1, tmp2, &a1,&a2,&a3);
        if(n==5 && !strcmp_case(tmp1, "FIT")){
            
            if(!strcmp_case(tmp2, "CRYSX")){
                pdbx_crystal_alignment.crysx_spindle = a1;
                pdbx_crystal_alignment.crysx_spindle_esd = a3;
                
            }else if( !strcmp_case(tmp2, "CRYSY")){
                pdbx_crystal_alignment.crysy_vertical = a1;
                pdbx_crystal_alignment.crysy_vertical_esd = a3;
                
            }else if( !strcmp_case(tmp2, "CRYSZ")){
                pdbx_crystal_alignment.crysz_beam = a1;
                pdbx_crystal_alignment.crysz_beam_esd = a3;
                
            }else if( !strcmp_case(tmp2, "XBEAM")){
                pdbx_crystal_alignment.xbeam = a1;
                pdbx_crystal_alignment.xbeam_esd = a3;
                
            }else if( !strcmp_case(tmp2, "YBEAM")){
                pdbx_crystal_alignment.ybeam = a1;
                pdbx_crystal_alignment.ybeam_esd = a3;
                
            }else if( !strcmp_case(tmp2, "CROSX")){
                pdbx_crystal_alignment.crossfire_x = a1;
                pdbx_crystal_alignment.crossfire_x_esd = a3;
                
            }else if( !strcmp_case(tmp2, "CROSY")){
                pdbx_crystal_alignment.crossfire_y = a1;
                pdbx_crystal_alignment.crossfire_y_esd = a3;
                
            }else if( !strcmp_case(tmp2, "CROSXY")){
                pdbx_crystal_alignment.crossfire_xy = a1;
                pdbx_crystal_alignment.crossfire_xy_esd = a3;
                
            }else if( !strcmp_case(tmp2, "DISTANCE")){
                pdbx_crystal_alignment.crystal_to_detector_distance = a1;
                pdbx_crystal_alignment.crystal_to_detector_distance_esd = a3;
                 
            }else if( !strcmp_case(tmp2, "MOSAICITY")){
                exptl_crystal.pdbx_mosaicity = a1;
                exptl_crystal.pdbx_mosaicity_esd = a3;
                
            }
            
        }else if(n==5 && !strcmp(tmp1, "AXIS") && !strcmp(tmp2, "PHI")){
            sscanf(strstr(str,"PHI")+3, "%f%f%f", &axis.phi[0], &axis.phi[1], &axis.phi[2]);
            n1=1;
        }else if(n==5 && !strcmp(tmp1, "AXIS") && !strcmp(tmp2, "KAPPA")){
            sscanf(strstr(str,"KAPPA")+5, "%f%f%f", &axis.kappa[0], &axis.kappa[1], &axis.kappa[2]);
            n2=1;
        }else if(n==5 && !strcmp(tmp1, "AXIS") && !strcmp(tmp2, "OMEGA")){
            sscanf(strstr(str,"OMEGA")+5, "%f%f%f", &axis.omega[0], &axis.omega[1], &axis.omega[2]);
            n3=1;
            
                 
        }else if(!strcmp_case(tmp1, "CrysX") &&
                 pattern5(str, "(spindle)", ".", "shift", "error", ".")){
            pdbx_crystal_alignment.crysx_spindle = get_float(str,")", "shift");
            pdbx_crystal_alignment.crysx_spindle_esd = get_float(str, "error","\n");
            
        }else if(!strcmp_case(tmp1, "CrysY") &&
                 pattern5(str, "(vertical)", ".", "shift", "error", ".")){
            pdbx_crystal_alignment.crysy_vertical = get_float(str,")", "shift");
            pdbx_crystal_alignment.crysy_vertical_esd = get_float(str, "error","\n");
            
        }else if(!strcmp_case(tmp1, "CrysZ") &&
                 pattern5(str, "(beam)", ".", "shift", "error", ".")){
            pdbx_crystal_alignment.crysz_beam = get_float(str,")", "shift");
            pdbx_crystal_alignment.crysz_beam_esd = get_float(str, "error","\n");
            
        }else if(!strcmp_case(tmp1, "X") &&
                 pattern5(str, "beam", ".", "shift", "error", ".")){
            pdbx_crystal_alignment.xbeam = get_float(str,"beam", "shift");
            pdbx_crystal_alignment.xbeam_esd = get_float(str, "error","\n");
            
        }else if(!strcmp_case(tmp1, "Y") &&
                 pattern5(str, "beam", ".", "shift", "error", ".")){
            pdbx_crystal_alignment.ybeam = get_float(str,"beam", "shift");
            pdbx_crystal_alignment.ybeam_esd = get_float(str, "error","\n");

            
        }else if(!strcmp_case(tmp1, "Crossfire") &&
                 pattern5(str, " x ", ".", "shift", "error", ".")){
            pdbx_crystal_alignment.crossfire_x = get_float(str," x ", "shift");
            pdbx_crystal_alignment.crossfire_x_esd = get_float(str, "error","\n");
            
        }else if(!strcmp_case(tmp1, "Crossfire") &&
                 pattern5(str, " y ", ".", "shift", "error", ".")){
            pdbx_crystal_alignment.crossfire_y = get_float(str," y ", "shift");
            pdbx_crystal_alignment.crossfire_y_esd = get_float(str, "error","\n");

        }else if(!strcmp_case(tmp1, "Crossfire") &&
                 pattern5(str, " xy ", ".", "shift", "error", ".")){
            pdbx_crystal_alignment.crossfire_xy = get_float(str," xy ", "shift");
            pdbx_crystal_alignment.crossfire_xy_esd = get_float(str, "error","\n");

            
        }else if(!strcmp_case(tmp1, "distance") &&
                 pattern5(str, "distance", ".", "shift", "error", ".")){
            pdbx_crystal_alignment.crystal_to_detector_distance = get_float(str,"distance", "shift");
            pdbx_crystal_alignment.crystal_to_detector_distance_esd = get_float(str, "error","\n");

        }else if(!strcmp_case(tmp1, "MOSAICITY") ){
            if(sscanf(str, "%*s%*s%f%f", &a1, &a2)==2){
                exptl_crystal.pdbx_mosaicity = a1;
                exptl_crystal.pdbx_mosaicity_esd = a2;
            }
            
        }
        
        
    }
    
          
    fclose(fp);
    if(n1==1 && n2==1 && n3==1){
        axis.key=3;
    }
        
}

void extract_DTREK_log_index( char *inpfile )
// Extracting index information  from LOG files of D*trek 
{
    char str[256], tmp1[80], tmp2[80];
    int  n=0;
    float a1, a2, a3;
    FILE *fp;
    

    if((fp = fopen(inpfile, "r"))==NULL) {
        printf("Can not open the file (%s)\n", inpfile);
        return ;
    }
    
    while (fgets(str, sizeof str, fp) != NULL) {
        n = sscanf(str,"%s%s%f%*s%f%*s%f", tmp1, tmp2, &a1,&a2,&a3);
        if(n==5){
             if(!strcmp_case(tmp1, "CrysX") && !strcmp_case(tmp2,"(spindle)")){
                 pdbx_crystal_alignment.crysx_spindle = a1;
                 pdbx_crystal_alignment.crysx_spindle_esd = a3;
                
             }else if(!strcmp_case(tmp1, "CrysY")&& !strcmp_case(tmp2, "(vertical)")){
                pdbx_crystal_alignment.crysy_vertical = a1;
                pdbx_crystal_alignment.crysy_vertical_esd = a3;
                
            }else if(!strcmp_case(tmp1, "CRYSZ") &&!strcmp_case(tmp2, "(beam)")){
                pdbx_crystal_alignment.crysz_beam = a1;
                pdbx_crystal_alignment.crysz_beam_esd = a3;
                
            }else if( !strcmp_case(tmp1, "X")&& !strcmp_case(tmp2, "BEAM")){
                pdbx_crystal_alignment.xbeam = a1;
                pdbx_crystal_alignment.xbeam_esd = a3;
                
            }else if( !strcmp_case(tmp1, "Y")&& !strcmp_case(tmp2, "BEAM")){
                pdbx_crystal_alignment.ybeam = a1;
                pdbx_crystal_alignment.ybeam_esd = a3;
                
            }else if( !strcmp_case(tmp1, "Crossfire")&& !strcmp_case(tmp2, "X")){
                pdbx_crystal_alignment.crossfire_x = a1;
                pdbx_crystal_alignment.crossfire_x_esd = a3;
                
            }else if( !strcmp_case(tmp1, "Crossfire")&&!strcmp_case(tmp2, "Y")){
                pdbx_crystal_alignment.crossfire_y = a1;
                pdbx_crystal_alignment.crossfire_y_esd = a3;
                
            }else if( !strcmp_case(tmp1, "Crossfire")&& !strcmp_case(tmp2, "XY")){
                pdbx_crystal_alignment.crossfire_xy = a1;
                pdbx_crystal_alignment.crossfire_xy_esd = a3;
            }
        }else if(!strcmp_case(tmp1, "DISTANCE")){
            n=sscanf(str,"%s%f%s%f%*s%f", tmp1, &a1, tmp2,&a2,&a3);
            if(n==5 && !strcmp_case(tmp2, "shift")){
                pdbx_crystal_alignment.crystal_to_detector_distance = a1;
                pdbx_crystal_alignment.crystal_to_detector_distance_esd = a3;
                 
            }
        }else if( !strcmp_case(tmp1, "MOSAICITY")){
            n=sscanf(str,"%s%f%s%f%*s%f", tmp1, &a1, tmp2,&a2,&a3);
            if(n==5 && !strcmp_case(tmp2, "shift")){
                exptl_crystal.pdbx_mosaicity = a1;
                exptl_crystal.pdbx_mosaicity_esd = a3;
            }
        }
    }
    
               
    fclose(fp);
    
}

void extract_MOSFLM_log_index(char *inpfile )
// Extracting index information  from LOG files of D*trek 
{
    char str[256], tmp1[80];
    float a1, a2;
    FILE *fp;
    

    if((fp = fopen(inpfile, "r"))==NULL) {
        printf("Can not open the file (%s)\n", inpfile);
        return ;
    }
    
    while (fgets(str, sizeof str, fp) != NULL) {
        if(pattern5(str,"Mosaic spread (MOSAIC)", ".","degrees", "","")){
            exptl_crystal.pdbx_mosaicity = get_float(str,"(MOSAIC)","degrees");
            
        }else if(pattern5(str,"Minimum and maximum X coordinate (XMIN,XMAX)", ".",".","mm","")){
            get_data_set_id(str, ")", "mm", tmp1);
            
            if(sscanf(tmp1, "%f%f", &a1, &a2)==2){
                pdbx_crystal_alignment.xbeam = a2;
            }
            
         }else if(pattern5(str,"Minimum and maximum Y coordinate (YMIN,YMAX)", ".",".","mm","")){
            get_data_set_id(str, ")", "mm", tmp1);
            
            if(sscanf(tmp1, "%f%f", &a1, &a2)==2){
                pdbx_crystal_alignment.ybeam = a2;
            }
            
         }else if(pattern5(str,"Crystal to detector distance (DIST)", ".","mm","","")){
             pdbx_crystal_alignment.crystal_to_detector_distance = get_float(str,")","mm");
             
         }
    }
    
               
    fclose(fp);
    
}

