/*
FILE:     nmr.C
*/
/*
VERSION:  3.004
*/
/*
DATE:     9/11/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
/* the input coordinate file either is PDF format or CIF */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "rcsb_math.h"

#include "harv.h"
#include "util.h"
void pdb_ensemble_to_mmcif(FILE *fout, char *inpfile);
void write_PDB_item(FILE *fout, char *str, long len, int format, int digit);
void write_PDB_30_70(FILE *fout, char *str, int n) ; //column 30_70
void write_PDB_6_27(FILE *fout, char *str);
void extract_remarks(FILE *fp, FILE *fout);
void process_log(FILE *fout, char *inpfile, char *prog);
void extract_log_cyana(FILE *fout, FILE *fp);
void get_software_info_nmr(FILE *fout,int argc, char **argv);
//extern int NPDB;


void proc_nmr_structure(char *outfile, int argc, char **argv)
{
    char prog[80];
    int i=0, nl=0;
    FILE *fout=NULL;
    

    fout = fopen(outfile,"w");
    printf("\nExtracting information about NMR structure\n\n");
    fprintf(fout,"\ndata_UNNAMED\n\n");
    fprintf(fout,"\n_exptl.entry_id          UNNAMED \n");
    fprintf(fout,"_exptl.method   NMR\n\n");

    get_software_info_nmr(fout, argc, argv);
    strcpy(prog,"");
    for(i=1; i<argc-1; i++){
        if(!strcmp_case(argv[i],"-r")){
            strcpy(prog, argv[i+1]);
            break;
        }
    }
    pdb_extract_info(fout);
    
    if(strlen(prog)<=1) {
        printf("Warnning! You did not provide the program name for refinement\n");
    }
    
//    NPDB=pdb2cif_maxit(argc,argv,pdb_cif);

//    printf("NPDB=%d\n", NPDB);
    
    for(i=1; i<argc-1; i++){  
        if(!strcmp_case(argv[i],"-IENT")){
            printf("\nAdditional_file; file_format=CIF: file_name=%s\n", argv[i+1]);
            if(is_cif(argv[i+1])>1){
                write_cif(fout,  argv[i+1], "CIF");
            }else {
                parse_adit_text_to_mmcif(fout, argv[i+1], 0);
            }
        }else if((!strcmp_case(argv[i],"-IPDB")||!strcmp_case(argv[i],"-PDB"))){
            printf("\nCoordinate file; file_format=PDB: file_name=%s\n", argv[i+1]);
            pdb_ensemble_to_mmcif(fout, argv[i+1]);
            
        }else if(!strcmp_case(argv[i],"-ILOG")||!strcmp_case(argv[i],"-LOG")){
            printf("\nLOG file name=%s\n", argv[i+1]);
            process_log(fout, argv[i+1], prog);
        }
        
    }

    fclose(fout);
/*    
    if(NPDB==1) {
        sprintf(command, "/bin/csh  -c \" /usr/bin/perl  -i -p -e 's/rcsb_/pdbx_/g; s/ndb_/pdbx_/g;' %s\" ", pdb_cif);
        system(command);
        merge_files( outfile, pdb_cif);
    }
*/    
    cifparse(outfile, outfile,90);
        
    nl=num_of_line(outfile);
    
    if(nl<=30){
        printf("Output file=  %s (%d lines)\n\n", outfile, nl);
        printf("\nThe output file is suspicious!\n");
        printf("%s %s\n\n","Please check the supported version of ",
               "pdb_extract or the input format");
    }
    printf("Output file=  %s (%d lines)\n\n", outfile, nl);

    
}

void get_software_info_nmr(FILE *fout,int argc, char **argv)
{
    int i, yes=0;
    
    fprintf(fout, "#\nloop_\n");
    fprintf(fout, "_pdbx_nmr_software.name \n");
    fprintf(fout, "_pdbx_nmr_software.version\n");
    fprintf(fout, "_pdbx_nmr_software.classification\n");
    fprintf(fout, "_pdbx_nmr_software.authors\n");
    
    for(i=1; i<argc-1; i++){
        if(!strcmp_case(argv[i],"-r")){
            if(!strcmp_case(argv[i+1],"cns") ||
               !strcmp_case(argv[i+1],"xplor")||
               !strcmp_case(argv[i+1],"x-plor")||
               !strcmp_case(argv[i+1],"x_plor")||
               !strcmp_case(argv[i+1],"cnx")){
                yes=1;
                fprintf(fout, "'%s' ?  refinement 'Brunger A. T. et.al.'\n", argv[i+1]);
            }else if(!strcmp_case(argv[i+1],"cyana") ||
               !strcmp_case(argv[i+1],"dyana")){
                yes=1;
                fprintf(fout, "'%s' ?  refinement 'Güntert P.'\n", argv[i+1]);
            }
            
        }
    }
    if(yes==0)fprintf(fout, "? ? ? ?\n");
    
}

void process_log(FILE *fout, char *inpfile, char *prog)
{
    FILE *fp;
    
    if((fp = fopen(inpfile,"r"))==NULL){
        open_file_error(inpfile, "process_log");
    }
    if(!strcmp_case(prog,"CYANA") || !strcmp_case(prog,"DYANA")){
        extract_log_cyana(fout, fp);
    }
    fclose(fp); 
}

void extract_log_cyana(FILE *fout, FILE *fp)
{
    char str[512], tmp[80], tmp1[256], **str_tok=NULL ;
    int nstr = 0, ndis_v=-1, nang_v=-1, rmsd=-1;
    float rmsd_bone_sig= -1, rmsd_bone= -1, rmsd_heavy_sig= -1, rmsd_heavy= -1;
    
    
    
    while (fgets(str, sizeof str, fp) != NULL) {
        strcpy(tmp1, str);
        sscanf(str,"%s", tmp);
        if(!strcmp_case(tmp,"Ave")){
            str_tok=string_token(tmp1, " ", &nstr);
            if(nstr==14){
//                printf("function= %s %s \n", str_tok[1],str_tok[2]);
            }
            if(nstr>0) free_memory_2d(str_tok,nstr);
            
        }else  if(!strcmp_case(tmp,"+/-")){
            str_tok=string_token(tmp1, " ", &nstr);
            if(nstr==14){
//                printf("sigma= %s %s \n", str_tok[1],str_tok[2]);
            }
            if(nstr>0) free_memory_2d(str_tok,nstr);
            
        }else if(strstr_case(str,"violated distance constraints.")){
            if(sscanf(str,"%s", tmp) ==1){
                ndis_v= atoi(tmp);
            }
            
        }else if(strstr_case(str,"violated angle constraints.")){
            if(sscanf(str,"%s", tmp) ==1){
                nang_v= atoi(tmp);
            }
            
        }else if(pattern5(str,"Average backbone RMSD to mean   : ", "+/-", " A ", "", "")){
            rmsd_bone = get_float (str,":", "+/-");
            rmsd_bone_sig = get_float (str, "+/-", " A ");
            rmsd=1;
            

        }else if(pattern5(str,"Average heavy atom RMSD to mean : ", "+/-", " A ", "", "")){
            rmsd_heavy = get_float (str,":", "+/-");
            rmsd_heavy_sig = get_float (str, "+/-", " A ");
            rmsd=1;
            
        }
    }

    fprintf(fout,"_pdbx_nmr_ensemble.entry_id  UNNAMED\n");
    if(ndis_v>=0) fprintf(fout,"_pdbx_nmr_ensemble.average_distance_constraint_violation %d\n", ndis_v);
    if(nang_v>=0) fprintf(fout," _pdbx_nmr_ensemble.average_torsion_angle_constraint_violation %d\n", nang_v);

    fprintf(fout,"%s\n %s\n %s\n %s\n %s\n ", "loop_ ",
            "_pdbx_nmr_ensemble_rms.entry_id ",
            "_pdbx_nmr_ensemble_rms.atom_type ",
            "_pdbx_nmr_ensemble_rms.distance_rms_dev ",
            "_pdbx_nmr_ensemble_rms.distance_rms_dev_error ");
    
    if(rmsd>0){
        if(rmsd_bone>=0 && rmsd_bone_sig>=0)
            fprintf(fout,"UNNAMED  bacbone %.2f  %.2f\n", rmsd_bone, rmsd_bone_sig);
        if(rmsd_heavy>=0 && rmsd_heavy_sig>=0) 
            fprintf(fout,"UNNAMED  'heavy atom'  %.2f  %.2f\n", rmsd_heavy, rmsd_heavy_sig);
    }else{
        fprintf(fout, "? ? ? ?\n"); 
    }
    
}

    

void pdb_ensemble_to_mmcif(FILE *fout, char *inpfile)
{
    char str[100];
    int n = 0, m = 0,em = 0, k=0, ek=1, len=0;
    FILE *fp;

    if((fp = fopen(inpfile,"r"))==NULL){
        open_file_error(inpfile, "pdb_ensemble_to_mmcif");
    }  

    while (fgets(str, sizeof str, fp) != NULL) {
        if(!strncmp_case(str, "ATOM  ",6)||!strncmp_case(str,"HETATM",6) ||
           !strncmp_case(str,"MODEL",5)){
            n++;
        }
        if(!strncmp_case(str,"MODEL",5))m++;
        if(!strncmp_case(str,"ENDMDL",6))em++;
    }
    rewind(fp);
    if(n>0){
        fprintf(fout,"%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n", 
                "loop_", 
                "_atom_site.group_PDB",
                "_atom_site.id",
                
                "_atom_site.auth_atom_id",
                "_atom_site.label_atom_id",
                "_atom_site.label_alt_id",
                "_atom_site.auth_comp_id",
                "_atom_site.label_comp_id",
                "_atom_site.auth_asym_id",
                "_atom_site.label_asym_id",
                "_atom_site.auth_seq_id" ,
                "_atom_site.label_seq_id" ,
                "_atom_site.pdbx_PDB_ins_code",
                
                "_atom_site.Cartn_x", 
                "_atom_site.Cartn_y", 
                "_atom_site.Cartn_z", 
                "_atom_site.occupancy", 
                "_atom_site.B_iso_or_equiv",
                "_atom_site.pdbx_PDB_model_num"
                
                );
        
        
        while (fgets(str, sizeof str, fp) != NULL) {
            if(!strncmp_case(str,"MODEL",5))k++;
            if(!strncmp_case(str,"ENDMDL",6))ek++;
            len = strlen(str);    
            if ((!strncmp_case(str, "ATOM  ", 6) ||
                 !strncmp_case(str, "HETATM", 6)) && len>54 ) {

                write_PDB_item(fout, str+0, 6, 1, 0); //_atom_site.group_PDB
                write_PDB_6_27(fout, str);  //column 6_27
                write_PDB_30_70(fout, str, 1);//column 30_70
                if(m>0){
                    fprintf(fout,"  %d\n", k);
                }else if(em>0 && m==0){
                    fprintf(fout,"  %d\n", ek);
                    
                }else{
                    fprintf(fout,"  1\n");
                }
                    
            }
        }
        
    }
    rewind(fp);
    extract_remarks(fp, fout);
}

void  extract_remarks(FILE *fp, FILE *fout)
// mostly the CNS file
{
    char str[100], **line=NULL, **str_tok, tmp[256];
    int n=0, i=0, nstr=0,ct=-1;
    float a4,a5, rms[20], force[20];

    for(i=0; i<20; i++){
        rms[i]=-1; force[i]=-1; 
    }
    
    
    line = new char *[200];
    while (fgets(str, sizeof str, fp) != NULL) {
        if(!strncmp(str,"REMARK",6)){
            if(!strncmp(str,"ATOM  ",6)) break;
            line[n]=new char[strlen(str)+1];
            strcpy(line[n], str);
            n++;
        }
    }
    
    for(i=0; i<n; i++){
        
//pdbx_nmr_constraints
        if(pattern5(line[i]+7,"NOEs", "in", "class", "", "")){
            ct = get_int(line[i], "REMARK", "NOEs");
            fprintf(fout, "_pdbx_nmr_constraints.entry_id   UNNAMED\n");
            if(ct>=0)fprintf(fout, "_pdbx_nmr_constraints.NOE_constraints_total %d\n", ct);
        }else if (pattern5(line[i]+7,"dihedral restraints with scale factors", ";", ";", ";", ";")){
            ct = get_int(line[i], "REMARK", "dihedral");
            if(ct>=0)fprintf(fout, "_pdbx_nmr_constraints.protein_other_angle_constraints_total_count  %d\n", ct);
            

//pdbx_nmr_ensemble
        }else if(pattern5(line[i]+7, "Accepted structure", "of", " structures", "", "")){
            nstr = get_int(line[i], "structures", "of");
            if(nstr<=0) nstr = get_int(line[i], "structure", "of");
            if(nstr>0){  
                fprintf(fout, "_pdbx_nmr_ensemble.entry_id  UNNAMED\n"); 
                fprintf(fout, "_pdbx_nmr_ensemble.conformers_submitted_total_number  %d\n",nstr); 
            }
            nstr = get_int(line[i], "of", "structures");
            fprintf(fout, "_pdbx_nmr_ensemble.conformers_calculated_total_number %d\n",nstr); 
            

//pdbx_nmr_ensemble_rms
        }else if(pattern5(line[i]+7, "bond, angles, improp, vdw", "dihed","","","") ||
                 pattern5(line[i]+7, "bonds,  angles, improp, noe,", "cdih","","","")){
            nstr = sscanf(line[i+2]+19,"%f%f%f%f%f", &rms[0],&rms[1],&rms[2],&a4,&a5);
            if(nstr==4){
                rms[3]=a4;
            }else if(nstr==5){
                rms[3]=a5;
            }
            
//or
        }else if(pattern5(line[i], "REMARK Bond   :", "+/-","+/-","","")){
            nstr=sscanf(line[i]+35,"%f",&rms[0]);
            
        }else if(pattern5(line[i], "REMARK Angl   :", "+/-","+/-","","")){
            nstr=sscanf(line[i]+35,"%f",&rms[1]);
            
        }else if(pattern5(line[i], "REMARK Impr   :", "+/-","+/-","","")){
            nstr=sscanf(line[i]+35,"%f",&rms[2]);
            
        }else if(pattern5(line[i], "REMARK Dihed  :", "+/-","+/-","","")){
            nstr=sscanf(line[i]+35,"%f",&rms[3]);
//or
        }else if(pattern5(line[i], "REMARK    bonds=", "A","","","")){
            rms[0] =  get_float_after_id(line[i]+10,"=");
        }else if(pattern5(line[i], "REMARK    angles=", "degrees","","","")){
            rms[1] =  get_float_after_id(line[i]+10,"=");
        }else if(pattern5(line[i], "REMARK    impropers=", "degrees","","","")){
            rms[2] =  get_float_after_id(line[i]+10,"=");
        }else if(pattern5(line[i], "REMARK    dihedrals=", "degrees","","","")){
            rms[3] =  get_float_after_id(line[i]+10,"=");
//or
        }else if(strstr(line[i], "REMARK            bonds,angles,impropers,noe,cdih,plan")){
            if(pattern5(line[i+1],"REMARK rms-d:", ",",  ",",  ",",  ",")){
                strcpy(tmp,strchr(line[i+1], ':')+1);
                str_tok = string_token(tmp, ",", &nstr);
                if(nstr>=4){
                    rms[0]= atof(str_tok[0]);
                    rms[1]= atof(str_tok[1]);
                    rms[2]= atof(str_tok[2]);
                    rms[3]= atof(str_tok[3]);
                }
                if(str_tok) free_memory_2d(str_tok, nstr);
            }
            
//pdbx_nmr_force_constants        
        }else if(!strncmp(line[i]+7,"overall =", 9) ){
            force[0]=get_float_after_id(line[i]+7, "=");
            
        }else if(!strncmp(line[i]+7,"bond    =", 9) || strstr(line[i],"REMARK bon  =") ){
            force[1]=get_float_after_id(line[i]+7, "=");
            
        }else if(!strncmp(line[i]+7,"angle   =", 9) || strstr(line[i],"REMARK ang  =")){
            force[2]=get_float_after_id(line[i]+7, "=");
            
        }else if(!strncmp(line[i]+7,"improp  =", 9) || strstr(line[i],"REMARK imp  =")){
            force[3]=get_float_after_id(line[i]+7, "=");
            
        }else if(!strncmp(line[i]+7,"vdw     =", 9) || strstr(line[i],"REMARK vdw  =")){
            force[4]=get_float_after_id(line[i]+7, "=");
            
        }else if(!strncmp(line[i]+7,"noe     =", 9) || strstr(line[i],"REMARK noe  =")){
            force[5]=get_float_after_id(line[i]+7, "=");
            
        }else if(!strncmp(line[i]+7,"coup    =", 9) || strstr(line[i],"REMARK coup  =")){
            force[6]=get_float_after_id(line[i]+7, "=");
       
        }else if(!strncmp(line[i]+7,"cdih    =", 9) || strstr(line[i],"REMARK dih  =") ){
            force[7]=get_float_after_id(line[i]+7, "=");
            
        }else if(strstr(line[i], "REMARK            overall,bonds,angles,improper,vdw,noe,cdih,plan")){
            if(pattern5(line[i+1],"REMARK energies:", ",",  ",",  ",",  ",")){
                strcpy(tmp,strchr(line[i+1], ':')+1);
                str_tok = string_token(tmp, ",", &nstr);
                if(nstr>=7){
                    
                    force[0]= atof(str_tok[0]);
                    force[1]= atof(str_tok[1]);
                    force[2]= atof(str_tok[2]);
                    force[3]= atof(str_tok[3]);
                    force[4]= atof(str_tok[4]);
                    force[5]= atof(str_tok[5]);
                    force[7]= atof(str_tok[6]);
                }
                if(str_tok) free_memory_2d(str_tok, nstr);
                
            }
        }
        
    }

    if(rms[0]>=0 || rms[1]>=1 || rms[2]>=0 || rms[3]>=0 ){
        fprintf(fout, " _pdbx_nmr_ensemble_rms.entry_id   UNNAMED \n");
    }
    if(rms[0]>=0) fprintf(fout, " _pdbx_nmr_ensemble_rms.covalent_bond_rms_dev  %.4f\n",rms[0]);
    if(rms[1]>=0) fprintf(fout, " _pdbx_nmr_ensemble_rms.bond_angle_rms_dev  %.3f\n",rms[1]);
    if(rms[2]>=0) fprintf(fout, " _pdbx_nmr_ensemble_rms.improper_torsion_angle_rms_dev  %.3f\n",rms[2]);
    if(rms[3]>=0) fprintf(fout, " _pdbx_nmr_ensemble_rms.dihedral_angles_rms_dev  %.3f\n",rms[3]);


    if(force[0]>=0 || force[1]>=0 || force[2]>=0 || force[3]>=0 || force[4]>=0 || force[5]>=0 )
        fprintf(fout,"_pdbx_nmr_force_constants.entry_id   UNNAMED \n");
    if(force[0]>=0) fprintf(fout,"_pdbx_nmr_force_constants.pdbx_overall %.4f\n", force[0]);
    if(force[1]>=0) fprintf(fout,"_pdbx_nmr_force_constants.covalent_geom_bond_term %.4f\n", force[1]);
    if(force[2]>=0) fprintf(fout,"_pdbx_nmr_force_constants.covalent_geom_angles_term %.4f\n", force[2]);
    if(force[3]>=0) fprintf(fout,"_pdbx_nmr_force_constants.covalent_geom_impropers_term %.4f\n", force[3]);
    if(force[4]>=0) fprintf(fout,"_pdbx_nmr_force_constants.non-bonded_inter_van_der_Waals_term %.4f\n", force[4]);
    if(force[5]>=0) fprintf(fout,"_pdbx_nmr_force_constants.pdbx_noe %.4f\n", force[5]);
    if(force[6]>=0) fprintf(fout,"_pdbx_nmr_force_constants.exptl_J_coupling_term %.4f\n", force[6]);
    if(force[7]>=0) fprintf(fout,"_pdbx_nmr_force_constants.exptl_torsion_angles_term %.4f\n", force[7]);
    
}

