/*
FILE:     phase.C
*/
/*
VERSION:  3.004
*/
/*
DATE:     9/11/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
/*  a program to extract information form the CCP4 program */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "structure.h"
#include "harv.h"
#include "phase.h"
#include "nrutil.h"
#include "util.h"


#define LEN 256
# define MAXL 512
#define NW  100
#define NBA 100

typedef struct
{
    char name[40];
    float resl_sh[NBIN];
    float resh_sh[NBIN];
    int ref_c_sh[NBIN];
    float item_c_sh[NBIN];
    int ref_a_sh[NBIN];
    float item_a_sh[NBIN];
    int ref_sh[NBIN];
    float item_sh[NBIN];
        
    float resl;
    float resh;
    int ref_c;
    float item_c;
    int ref_a;
    float item_a;
    int ref;
    float item;
    int nbin;
        
}PHASE;


void mlphare_log(FILE *fout, char *inpfile, char *method);
void get_phase_stat(PHASE *RK, char **line, int nline, int i, int nset, int *j);
void get_data_new_v2(FILE *fp, char *rc, double **res_L, double **res_H, 
                     double **rc_ai, double **rc_ci, long **nrc_ai, long **nrc_ci, 
                     double **rc_aa, double **rc_ca, long **nrc_aa, long **nrc_ca, 
                     double *rc_ai_all, double *rc_ci_all, long *nrc_ai_all, long *nrc_ci_all,
                     double *rc_aa_all, double *rc_ca_all, long *nrc_aa_all, long *nrc_ca_all,
                     long nba, long *nb);

void SHARP_phasing_new_v2(FILE *fout, char *inpfile, char *method);


extern  CELL cell ;
extern SYMMETRY symmetry;
PHASING_MAD phasing_mad;
PDBX_PHASING_MAD_SET pdbx_phasing_mad_set;
PDBX_PHASING_MAD_SET_SHELL pdbx_phasing_mad_set_shell[NBIN];
PDBX_PHASING_MAD_SET_SITE pdbx_phasing_mad_set_site;
PDBX_PHASING_MAD_SHELL pdbx_phasing_mad_shell[NBIN];
PHASING_MAD_SET phasing_mad_set;

PHASING_MIR phasing_mir;
PHASING_MIR_DER phasing_mir_der;
PHASING_MIR_DER_SHELL phasing_mir_der_shell;
PHASING_MIR_DER_SITE phasing_mir_der_site;
PHASING_MIR_SHELL phasing_mir_shell;



void CCP4_phase_file_assign(FILE *fout, char *inpfile, char *format,
                            char *method)
/* assign the files (from phasing) to different subroutines */
{
    FILE *fp;
    
    if((fp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s (routine:CCP4_phase_file_assign)\n",
               inpfile);
        return;
    }


    if(!strcmp_case(format, "PDB")) { /*take as heavy atom XYZ */
        heavy_coord_pdb(fout, inpfile, method);
        return;
    }
/* must be prephadata log file and mlphare log file*/
    if(!strcmp_case(format, "LOG")) { 
        CCP4_MAD_wave(fout, inpfile, method);
        mlphare_log(fout, inpfile,method);
   
        return;
    }
    
    if(!strcmp_case(format, "CIF")) {
        CCP4_mlphare_cif(fout, inpfile, method);
        return;
    }
    fclose(fp);
    
}

       
    
void mlphare_log(FILE *fout, char *inpfile, char *method)
{
    char  tmp[80], **str1=NULL,**str=NULL, **line=NULL,  **deriv=NULL, bfact[30];
    int  n=0,m,i,k, j,nline,nline1, ntoken=0, nbin=0;
    
    int refl_sh[NBIN],refl_a_sh[NBIN], refl_c_sh[NBIN], refl=0,refl_a=0, refl_c=0;
    float fom_sh[NBIN],fom_a_sh[NBIN], fom_c_sh[NBIN],resL_sh[NBIN],resH_sh[NBIN];
    
    int ref_a_sh[500][NBIN], ref_c_sh[500][NBIN], ref_a[500], ref_c[500];
    float loc_a_sh[500][NBIN], php_a_sh[500][NBIN], rc_a_sh[500][NBIN];
    float loc_c_sh[500][NBIN], php_c_sh[500][NBIN], rc_c_sh[500][NBIN];
    float loc_a[500], php_a[500], rc_a[500], loc_c[500], php_c[500], rc_c[500];
    
    float fom=0, fom_a=0, fom_c=0, *resL_set,*resH_set ;
    int n1, n2,n3, n4,n5,n6, nsite, nder, len;
    
    char  atom[10];
    float a1, a2, a3, a4, a5, a6;
    FILE *fp;
 
// 500 max derivatives
    
    const char *phasing[9] =  //
        {
            "entry_id",
            "d_res_high",
            "d_res_low",
            "reflns",
            "fom",
            "reflns_centric",
            "FOM_centric",
            "reflns_acentric",
            "FOM_acentric"
        };
        
    const char *phasing_shell[8] = 
        {
            "d_res_high",
            "d_res_low",
            "reflns",
            "fom",
            "reflns_centric",
            "FOM_centric",
            "reflns_acentric",
            "FOM_acentric"
        };
        
    const char *items_site[8] = 
        {
            "id",
            "atom_type_symbol",
            "fract_x",
            "fract_y",
            "fract_z",
            "B_iso",
            "Occupancy",
            "der_id"
        };
        
    const char *item_der_shell[12] = 
        {
            "id",
            "d_res_high",
            "d_res_low",
            "pdbx_reflns_acentric",
            "pdbx_loc_acentric",
            "pdbx_power_acentric",
            "pdbx_R_cullis_acentric",
            "pdbx_reflns_centric",
            "pdbx_loc_centric",
            "pdbx_power_centric",
            "pdbx_R_cullis_centric",
            "der_id"
        };
    
 
    if((fp=fopen(inpfile, "r"))==NULL)
       open_file_error(inpfile, "mlphare_log");
    
    str1  = get_lines_from_file(inpfile, &nline1);
    str=new char *[nline1+1];
    for(i=0; i<nline1-1; i++){
        if(pattern5(str1[i], "## ", "CCP4", "MLPHARE ", "version", "##")){
            n=0;
            for(j=i; j<nline1-1; j++){
                if(strstr_case(str1[j], "MLPHARE:  Normal termination")){
                    i=j;
                    break;
                }
                
                len=strlen(str1[j]);
                str[n]=new char[len+1];
                strcpy(str[n], str1[j]);
                n++;
            }
        }
    }
    if (nline1>0) free_memory_2d(str1,nline1);
    
    nline=n;
    
    phasing_mir_shell.nbin = 0;
    
    for(i=0; i<nline-1; i++){
//cell        
        if (strstr_space(str[i], "* Cell Dimensions :")){
            n=sscanf(str[i+1],"%f%f%f%f%f%f", &a1, &a2, &a3, &a4, &a5, &a6);
            
            if( n==6 ){
                cell.length_a = a1;
                cell.length_b = a2;
                cell.length_c = a3;
                cell.angle_alpha = a4;
                cell.angle_beta = a5;
                cell.angle_gamma = a6;        
            }
//symmetry
        }else if(pattern5(str[i], "Space group =", "(number", ")", "", "") ){
            strcpy(tmp, "");
            strcpy(tmp, get_substr(str[i],"Space group =", "(number"));
            char2space(tmp, '\'');
            if(strlen(tmp)>0) strcpy(symmetry.space_group_name_H_M, tmp);
            strcpy(tmp, "");
            strcpy(tmp, get_substr(str[i],"number", ")"));
            if(strlen(tmp)>0) symmetry.Int_Tables_number= atoi(tmp);

//reso            
        }else if(strstr_space(str[i], "Protein Resolution Limits are from") &&
                 pattern5(str[i], "from" , "to", "Angstroms", "", "")){
            phasing_mad.pdbx_d_res_low  = get_float(str[i],"from","to");
            phasing_mad.pdbx_d_res_high = get_float(str[i],"to", "Angstroms");
            
//resol shell
        }else if(strstr_space(str[i], "Resolution of each shell in angstroms:")){
            line = string_token(str[i+1]," " , &ntoken);
            nbin = ntoken;
            resL_sh[0] = phasing_mad.pdbx_d_res_low;
            resH_sh[0] = atof(line[0]);
            for(k=1; k<ntoken; k++){
                resL_sh[k] = atof(line[k-1]);
                resH_sh[k] = atof(line[k]);
            }
            
            
            phasing_mir_shell.nbin = ntoken;
            phasing_mir_shell.d_res_low[0] = phasing_mad.pdbx_d_res_low;
            phasing_mir_shell.d_res_high[0] = atof(line[0]);
            
            for(k=1; k<ntoken; k++){
                phasing_mir_shell.d_res_low[k] = atof(line[k-1]);
                phasing_mir_shell.d_res_high[k] = atof(line[k]);
            }
            
//FOM shell  -ACENTRIC          
        }else if(strstr_space(str[i],"Number of Measurements phased -ACENTRIC  TOTAL")&&
                 strstr_space(str[i+2],"Mean Figure of Merit") ){
            
            line = string_token(str[i+1]," " , &n1);
            if(nbin == n1-1){
                for(k=0; k<n1-1; k++){
                    refl_a_sh[k] = atoi(line[k]);
                }
                refl_a = atoi(line[n1-1]);
            }
            
            line = string_token(str[i+3]," " , &n2);
            if(nbin == n2 -1){
                for(k=0; k<n2 -1; k++){
                    fom_a_sh[k] = atof(line[k]);
                }
                fom_a = atof(line[n2 -1]);
            }
            
//FOM shell  -CENTRIC          
            
        }else if(strstr_space(str[i],"Number of Measurements phased -CENTRIC  TOTAL")&&
                 strstr_space(str[i+2],"Mean Figure of Merit")){
            
            line = string_token(str[i+1]," " , &n3 );
            if(nbin == n3 -1){
                for(k=0; k<n3 -1; k++){
                    refl_c_sh[k] = atoi(line[k]);
                }
                refl_c = atoi(line[n3 -1]);
            }

            line = string_token(str[i+3]," " , &n4 );
            if(nbin == n4 -1){
                for(k=0; k<n4 -1; k++){
                    fom_c_sh[k] = atof(line[k]);
                }
                fom_c = atof(line[n4 -1]);
            }
            
//FOM shell  -ALL          
            
        }else if(strstr_space(str[i],"Number of Measurements phased -ALL  TOTAL")&&
                 strstr_space(str[i+2],"Mean Figure of Merit")){
            
            line = string_token(str[i+1]," " , &n5 );
            if(nbin == n5 -1){
                for(k=0; k<n5 -1; k++){
                    refl_sh[k] = atoi(line[k]);
                }
                refl = atoi(line[n5 -1]);
            }
            
            line = string_token(str[i+3]," " , &n6 );
            if(nbin == n6 -1){
                for(k=0; k<n6 -1; k++){
                    fom_sh[k] = atof(line[k]);
                }
                fom = atof(line[n6 -1]);
            }
        }
    }

//
    
  
    if(nbin ==n1-1 && nbin>0){
        if(!strcmp_case(method,"mir"))
           print_mmcif_head(fout, "_phasing_MIR_shell", phasing_shell, 8);
        else
           print_mmcif_head(fout, "_pdbx_phasing_MAD_shell", phasing_shell, 8);
            
        for(i=0; i<nbin; i++){
            
            fprintf(fout, "%.2f %.2f  %d %.3f  %d %.3f  %d %.3f\n",
                    resH_sh[i], resL_sh[i],refl_sh[i],fom_sh[i],
                    refl_c_sh[i],fom_c_sh[i],refl_a_sh[i],fom_a_sh[i]);
        }
        
        if(!strcmp_case(method,"mir"))
            print_mmcif_head(fout, "_phasing_MIR", phasing, 9);
        else
            print_mmcif_head(fout, "_phasing_MAD", phasing, 9);
        
            
        fprintf(fout, "%s %.2f %.2f  %d %.3f  %d %.3f  %d %.3f\n",DATA_BLOCK,
                resH_sh[nbin-1], resL_sh[0],refl,fom,refl_c,fom_c,refl_a,fom_a);
    }
    
   
//atom site coordinates

    if(!strcmp_case(method,"mir"))
        print_mmcif_head(fout,"_phasing_MIR_der_site", items_site, 8);
    else{
        print_mmcif_head(fout,"_pdbx_phasing_MAD_set_site", items_site, 8);
    }
    
    deriv = new char *[500];
    resL_set = new float [500];
    resH_set = new float [500];

    nsite=0;
    nder=0;
    for(i=100; i<nline-1; i++){
        if (strstr_case(str[i],"MLPHARE:  Normal termination"))break;
        sscanf(str[i],"%s", tmp);
        
        if(!strcmp(tmp, "DERIV") && 
           strstr(str[i+1], "DCYCLE PHASE ALL REFCYC ALL KBOV ALL") &&
           strstr(str[i+2], "RESO ")){
            sscanf(strstr(str[i], "DERIV")+5, "%s",tmp);
            
//            get_data_set_id(str[i], "DERIV", "\n",tmp);
//            rid_of_front_end_space(tmp);

            deriv[nder]= new char [strlen(tmp)+1];
            strcpy(deriv[nder], tmp);
            
            n=sscanf(str[i+2],"%*s%f%f", &a1, &a2);
            if(n==2){
                resL_set[nder] = a1;
                resH_set[nder] = a2;
            }            
            
            for(j=i+1; j<nline-1; j++){
                rid_of_front_space(str[j]);
                if(!strncmp(str[j],  "DERIV ", 6) ){ //next
                    i=j-1;
                    break;
                }
                
                if(pattern5(str[j],"ATOM", ".",".",".", "BFAC")){
                    
                    n=sscanf(str[j], "%*s %s %f %f %f %f ",
                             atom, &a1, &a2, &a3, &a4);
                    strcpy(bfact, get_string_after_id_number(str[j], "bfac",1));
                    if (n==5 && bfact ){
                        nsite++;
                        fprintf(fout, "%d  '%s' %.3f %.3f %.3f  %s  %.3f \"%s\"\n",
                                nsite, atom, a1, a2, a3, bfact, a4, deriv[nder]);
                    }
                }
                if (strstr_case(str[j],"MLPHARE:  Normal termination"))break;
               
            }
            
            nder++;

        }
    }
    
    if(nsite==0)fprintf(fout,"1 ? ? ? ? ? ? ?\n");
    
    
//  Phasing statistics: phasing Power, Lack_of_closure, R_Cullis(censtric, acentric)
    m=0;
    n=0;
    
    for(i=0; i<nline-1; i++){
        if(strstr_case(str[i],"1/resol^2  Resol   Nref_a DISO_a LOC_a PhP_a CullR_a Nref_c DISO_c LOC_cPhP_c CullR_c")  ){
            n=0;
            if (strstr_case(str[i],"END OF MLPHARE PROCEDURE")||
                strstr_case(str[i],"MLPHARE:  Normal termination")){
                nbin=n;
                break;
            }
            
            
            for(k=i+1; k<nline-1; k++){
                ref_a_sh[m][n] = get_int_by_length(str[k],17, 24);
                loc_a_sh[m][n]  = get_float_by_length(str[k],32,38);
                php_a_sh[m][n]  = get_float_by_length(str[k],39,43);
                rc_a_sh[m][n]  = get_float_by_length(str[k],44,51);
                ref_c_sh[m][n]  = get_int_by_length(str[k],51,58);
                loc_c_sh[m][n]  = get_float_by_length(str[k],66,72);
                php_c_sh[m][n]  = get_float_by_length(str[k],73, 77);
                rc_c_sh[m][n]  = get_float_by_length(str[k],78, 84);
                n++;
                
                
                if(strstr_case(str[k], "total")){

                    ref_a[m] = get_int_by_length(str[k],17, 24);
                    loc_a[m] = get_float_by_length(str[k],32,38);
                    php_a[m]  = get_float_by_length(str[k],39,43);
                    rc_a[m]  = get_float_by_length(str[k],44,51);
                    ref_c[m]  = get_int_by_length(str[k],51,58);
                    loc_c[m] = get_float_by_length(str[k],66,72);
                    php_c[m]  = get_float_by_length(str[k],73, 77);
                    rc_c[m]  = get_float_by_length(str[k],78, 84);
                    i=k;
                    
                    break;
                }
            }   
            m++;       
        }
    }
    
    
    if( m>0){
        if(!strcmp_case(method,"mir"))
            print_mmcif_head(fout, "_phasing_MIR_der_shell", item_der_shell, 12);
        else
            print_mmcif_head(fout, "_pdbx_phasing_MAD_set_shell", item_der_shell, 12);      
        for(j=0; j<m; j++){
            for(k=0; k<nbin; k++){  
                fprintf(fout, "%d %.2f %.2f  %d %.3f %.3f %.3f    %d %.3f %.3f %.3f  \"%s\"\n",
                        j+1,resH_sh[k], resL_sh[k], 
                        ref_a_sh[j][k],loc_a_sh[j][k],php_a_sh[j][k],rc_a_sh[j][k],
                        ref_c_sh[j][k],loc_c_sh[j][k],php_c_sh[j][k],rc_c_sh[j][k],deriv[j]);
            }
        }
    }
        
    if(nder==m && m>0 ){
        if(!strcmp_case(method,"mir"))
            print_mmcif_head(fout, "_phasing_MIR_der", item_der_shell, 12);
        else
            print_mmcif_head(fout, "_pdbx_phasing_MAD_set", item_der_shell, 12);
            
        for(j=0; j<m; j++){
            fprintf(fout, "%d %.2f %.2f  %d %.3f %.3f %.3f   %d %.3f %.3f %.3f  \"%s\"\n",
                    j+1,resH_set[j],resL_set[j],
                    ref_a[j],loc_a[j],php_a[j],rc_a[j],
                    ref_c[j],loc_c[j],php_c[j],rc_c[j], deriv[j]);
        }
          
    }

    delete [] deriv;
    delete [] resL_set;
    delete [] resH_set;
    if (nline>0) free_memory_2d(str,nline);
    
    fclose(fp); 
    
}

void CCP4_MAD_wave(FILE *fout, char inpfile[], char method[])
/* extract the wavelength information from *_prephadata.log file */
{
    char str[512];
    int num=0,id[20], n=0 , i;
    float wave[20], fprime[20], f2prime[20];
    FILE *fp;

    if(!strcmp_case(method,"SIR") || !strcmp_case(method,"MIR"))return;

    fprintf(fout,"\n######## the wavelength information ########\n\n");
            
    if((fp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s(routine:CCP4_MAD_wave)\n",inpfile);
        return;
    }

    while (fgets(str, sizeof str, fp) != NULL) {
       
        if(strstr_space(str, "Ndata Wavelengths      Fprime   Fdoublepr")){
            n=0;
            while (fgets(str, sizeof str, fp) != NULL){
                num=sscanf(str,"%d %f %f %f",
                           &id[n], &wave[n], &fprime[n], &f2prime[n]);
                
                if(strstr_space(str,"Data will be excluded if the") ||
                   strstr_space(str,"file opened on unit" ))
                    break;
                if(num ==4 ){
                    n++;
                }  
            }
            break;
        }
    }

    if(n>0) {     
        fprintf(fout,"\nloop_\n"
                "_phasing_MAD_set.expt_id\n"
                "_phasing_MAD_set.clust_id\n"
                "_phasing_MAD_set.set_id\n"
                "_phasing_MAD_set.wavelength\n"
                "_phasing_MAD_set.f_prime\n"
                "_phasing_MAD_set.f_double_prime\n"
                );
        
        for( i=0; i<n; i++){
            fprintf(fout, "1  '%d wavelength' %d %.3f %.3f %.3f\n",
                    n, id[i], wave[i], fprime[i], f2prime[i] );
        }
    }
    
           
    fprintf(fout, "#\n\n");
    fclose(fp);
}

void CCP4_mlphare_cif(FILE *fout, char inpfile[], char method[])
/* append the *.mlphare to the cif file; do not need data block and entry ID */ 
{

    char tmp[512], str[512], category[512],item[512], tmpfile[512];
    int ncat=0;
    
    FILE *fp;
    
    fprintf(fout, "\n\n########## Mlphare phasing information #########\n\n");
    if( is_cif(inpfile)< 5) {
        printf("The input file (%s) for mlphare is not in mmCIF format",inpfile);
        return;
    }

    sprintf(tmpfile, "%s_TEMP", inpfile);
    cifparse(inpfile, tmpfile,500);
    if((fp = fopen(tmpfile, "r"))==NULL) {        
        printf("Can not open the file %s(routine:CCP4_mlphare_cif)\n",inpfile);
        return;
    }
    
    if(!strcmp_case(method,"MIR")){
        while (fgets(str, sizeof str, fp) != NULL) {
            strcpy(tmp, str);
            upperstr(tmp);
            rid_of_front_space(tmp);
            if(!strncmp_case(tmp, "DATA_",5)||
               !strncmp_case(tmp, "_audit.creation_date",20) || 
               !strncmp_case(tmp, "_software.",10) || 
               !strncmp_case(tmp, "_ENTRY.ID",9))
                continue;
            fprintf(fout, "%s", str);
        }
        fclose(fp);
        return;
        
    }else if (!strcmp_case(method,"MAD")){
        while (fgets(str, sizeof str, fp) != NULL) {
            strcpy(tmp, str);
            upperstr(tmp);
            rid_of_front_space(tmp);
            if(!strncmp_case(tmp, "DATA_", 5) ||
               !strncmp_case(tmp, "_audit.creation_date",20) ||
               !strncmp_case(tmp, "_software.",10) || 
               !strncmp_case(tmp, "_ENTRY.ID", 9))
                continue;
            
            ncat = check_category(str, category, item);
            
            strcpy(tmp,str);

            if(!strcmp_case(category,  "_PHASING_MIR_SHELL") && ncat>0){
               replace_first_str(tmp,"_PHASING_MIR","_pdbx_phasing_MAD",str);
                         
            }else if(!strcmp_case(category,  "_PHASING_MIR") && ncat>0){
                if(!strcmp_case(item,"entry_id") || !strcmp_case(item,"method")
                   || !strcmp_case(item,"details")){
                    replace_first_str(tmp,"_PHASING_MIR","_phasing_MAD",str);
                }else{
                    replace_first_str(tmp,"_PHASING_MIR.","_phasing_MAD.pdbx_",str);
                }
                
            }else if(!strcmp_case(category, "_PHASING_MIR_DER_SITE")&& ncat>0){
                replace_first_str(tmp,"_PHASING_MIR_DER_SITE",
                                  "_pdbx_phasing_MAD_set_site",str);
                if(!strcmp_case(item, "der_id")){
                    replace_first_str(tmp,"_PHASING_MIR_DER_SITE.der_id",
                                      "_pdbx_phasing_MAD_set_site.set_id",str);
                    
                }else if(!strcmp_case(item, "Occupancy_anom")){
                    replace_first_str(tmp,"_phasing_MIR_der_site.Occupancy_anom",
                                      "_pdbx_phasing_MAD_set_site.Occupancy",str);
                    
                }else if(!strcmp_case(item, "Occupancy_anom_esd")){
                    replace_first_str(tmp,"_phasing_MIR_der_site.occupancy_anom_esd",
                                      "_pdbx_phasing_MAD_set_site.occupancy_esd",str);
                }
                
            }else if(!strcmp_case(category,"_PHASING_MIR_DER") && ncat>0){
                replace_first_str(tmp,"_PHASING_MIR_DER",
                                  "_pdbx_phasing_MAD_set",str);
                
                if(!strcmp_case(item, "Reflns_anomalous")){
                    replace_first_str(tmp,"_phasing_MIR_der.Reflns_anomalous",
                                      "_pdbx_phasing_MAD_set.Reflns",str);
                    
                }else if(!strcmp_case(item, "R_cullis_anomalous")){
                    replace_first_str(tmp,"_phasing_MIR_der.R_cullis_anomalous",
                                      "_pdbx_phasing_MAD_set.R_cullis",str);
                }
            }
            if(strstr(str,"\n"))
                fprintf(fout, "%s", str);
            else
                fprintf(fout, "%s\n", str);
        }
    }
    delete_file(tmpfile);
    
    fclose(fp);
   
}



void CCP4_phaser_ep(FILE *fout, char *inpfile,char *format, char *method)
{
    char **line, atom[10], str[256],tmp[80];
    int i,j, n1, n, n2, nline,n3, natom,nbin=0, nwave, nfom;
    float  x,y,z,occ,b,a2,a1, a3,a4,a5,a6;
    
    
    const char *items_site[7] = 
        {
            "id",
            "atom_type_symbol",
            "fract_x",
            "fract_y",
            "fract_z",
            "Occupancy",
            "B_iso"
        };
    const char *items_wave[6] =
        {
            "expt_id",
            "set_id",
            "pdbx_atom_type",
            "wavelength",
            "f_prime",
            "f_double_prime"
        };
    const char *phasing[9] = 
        {
            "entry_id",
            "d_res_high",
            "d_res_low",
            "reflns_acentric",
            "FOM_acentric",
            "reflns_centric",
            "FOM_centric",
            "reflns",
            "fom"
        };
        
    const char *phasing_shell[8] = 
        {
            "d_res_high",
            "d_res_low",
            "reflns_acentric",
            "FOM_acentric",
            "reflns_centric",
            "FOM_centric",
            "reflns",
            "fom"
        };
        
    line=get_lines_from_file(inpfile, &nline);
    
    natom=0;
    for(i=0; i<nline; i++){// atom site
        if(strlen(line[i])>250) continue;
        
        strcpy(str,line[i]);
        rid_of_front_space(str);
        if(!strncmp(str,"ATOM ", 5) &&
           pattern5(str,"CRYST", "ELEMENT", "FRAC", "OCC", "ISOB")){
            natom++;
            if(natom==1){
                print_mmcif_head(fout,"_pdbx_phasing_MAD_set_site", items_site, 7);
            }
            sscanf(strstr(str,"ELEMENT")+ 7, "%s", atom);
            n=sscanf(strstr(str,"FRAC")+ 4, "%f%f%f", &x, &y, &z );
            sscanf(strstr(str," OCC")+ 4, "%f", &occ );
            sscanf(strstr(str,"ISOB")+ 4, "%f", &b );
            if(n==3)fprintf(fout, "%2d %2s %8.4f %8.4f %8.4f %6.3f %6.2f\n",
                            natom, atom, x,y,z,occ,b);
            if(natom>0 && n!=3) break;
            
            
        }else if (strstr_space(str, "Unit Cell:")){
            n=sscanf(str+10,"%f%f%f%f%f%f", &a1, &a2, &a3, &a4, &a5, &a6);
            if( n==6 ){
                cell.length_a = a1;
                cell.length_b = a2;
                cell.length_c = a3;
                cell.angle_alpha = a4;
                cell.angle_beta = a5;
                cell.angle_gamma = a6;        
            }
//symmetry
        }else if (pattern5(str, "Space-Group Name (Hall Symbol):", "(", ")","","")){
            
            get_data_set_id(str, ":", "(", tmp);
            if(strlen(tmp)>1) strcpy(symmetry.space_group_name_H_M, tmp);
            
        }else if( strstr_space(str,"Space-Group Number:")){
            n=get_int_after_id(str, ":");
            if(n>0)symmetry.Int_Tables_number=n;
        }
    }
    
    nfom=0;
    nwave=0;
    for(i=nline-1; i>0; i--){ //FOM, shells
        if(nfom==0 &&
           strstr_space(line[i], "Number FOM   Number FOM   Number FOM   Number FOM")){
            nbin=0;
//            start=nline-i;
            
            for(j=i+1; j<nline; j++){
                replace_first_str(line[j],"- ", " ", str);
                n=sscanf(str, "%s %f %f %d  %f %d %f %*s%*s %d %f",
                         tmp, &a1,&a2,&n1, &a3, &n2 , &a4, &n3, &a5);
                
                if(n==9){
                    nbin++;
                    if(nbin==1){
                        print_mmcif_head(fout, "_pdbx_phasing_MAD_shell", phasing_shell, 8);
                        nfom=1;
                    }
                    if(!strcmp_case(tmp,"ALL")){
                        print_mmcif_head(fout, "_pdbx_phasing_MAD", phasing, 9);
                        fprintf(fout,"%s %5.2f %5.2f %4d %6.3f %4d %6.3f %4d %6.3f\n",
                               DATA_BLOCK, a2,a1,n1, a3,n2,a4,n3,a5);
                        break;
                    }
                    fprintf(fout,"%5.2f %5.2f %4d %6.3f %4d %6.3f %4d %6.3f\n",
                            a2,a1,n1, a3,n2,a4,n3,a5);
                }
            }
        }else if(nwave==0 &&
                 !strncmp_case(line[i],"ATOM ", 5) &&  //wavelength
                 pattern5(line[i]," f", " (f')", "", "", "")){
            
            n=sscanf(line[i+1], "%s%f%f", atom, &x, &y);
            if(n==3){
                print_mmcif_head(fout,"_phasing_MAD_set", items_wave, 6);
                fprintf(fout," 1  1 %s ? %.4f %.4f\n", atom, x, y);
                nwave=1;
            }
        }        
    }
    
    if(nline && line)free_memory_2d(line, nline);
    
    
}


void get_phase_stat(PHASE *rk, char **line, int nline, int i, int nset, int *k)
{

    char tmp[80];
    int j,m, n, n1, n2, n3, nbin=0;
    float a1, a2,a3,a4,a5;

    *k=i;

    sscanf(strstr(line[i],"Dataset")+7, "%s", rk[nset].name);
    nbin=0;
    for(j=i+1; j<nline; j++){
        n=sscanf(line[j],"%s %f%f %*s %d%f %d%f %d%f",
                 tmp, &a1,&a2,&n1, &a3, &n2 , &a4, &n3, &a5);
        if(n==9){
            rk[nset].resl_sh[nbin]=a1;
            rk[nset].resh_sh[nbin]=a2;
            rk[nset].ref_c_sh[nbin]=n1;
            rk[nset].item_c_sh[nbin]=a3;
            rk[nset].ref_a_sh[nbin]=n2;
            rk[nset].item_a_sh[nbin]=a4;
            rk[nset].ref_sh[nbin]=n3;
            rk[nset].item_sh[nbin]=a5;
            nbin++;
                
        }else if(n==1 && strstr(line[j],"TOTAL")){
            m=sscanf(line[j+1]," %f%f %d%f %d%f %d%f",
                     &a1,&a2,&n1, &a3, &n2 , &a4, &n3, &a5);
            if(m==8) {
            rk[nset].resl=a1;
            rk[nset].resh=a2;
            rk[nset].ref_c=n1;
            rk[nset].item_c=a3;
            rk[nset].ref_a=n2;
            rk[nset].item_a=a4;
            rk[nset].ref=n3;
            rk[nset].item=a5;
            rk[nset].nbin=nbin;
            }
            
            *k=j;
            break;
        }
    }
}


void BP3_phase(FILE *fout, char *inpfile,char *format, char *method)
{
    char **line, atom[10], tmp[80];
    float  x,y,z,occ,b;
    float a2,a1,a3,a4,a5,dat[6];
    int i,j, m, n1, n, n2, nline,n3,nbin=0, nwave, nfom=0;
    int nrc=0, nrk=0, npp=0;

    const char *items_wave[6] =
        {
            "expt_id",
            "set_id",
            "pdbx_atom_type",
            "wavelength",
            "f_prime",
            "f_double_prime"
        };
    const char *phasing[9] = 
        {
            "entry_id",
            "d_res_high",
            "d_res_low",
            "reflns_acentric",
            "FOM_acentric",
            "reflns_centric",
            "FOM_centric",
            "reflns",
            "fom"
        };
        
    const char *phasing_shell[8] = 
        {
            "d_res_high",
            "d_res_low",
            "reflns_acentric",
            "FOM_acentric",
            "reflns_centric",
            "FOM_centric",
            "reflns",
            "fom"
        };
    
//

    PHASE RC[10];
    PHASE RK[10];
    PHASE PP[10];
    
      
    line=get_lines_from_file(inpfile, &nline);
    
    nwave=0;
    strcpy(atom, "?");
    for(i=0; i<nline-1; i++){// atom site
        if(strstr_space(line[i], "Scattering factors for atom")){
            sscanf(strstr(line[i],"atom")+4, "%s", atom);
        }
        if(strstr_space(line[i], "Wavelength #") &&
           pattern5(line[i+1], "fp =", "fpp =", ".", ".", "")){
            nwave++;
            if(nwave==1){
                print_mmcif_head(fout,"_phasing_MAD_set", items_wave, 6);
            }
            
            a1=get_float_after_id(line[i+1], "fp =");
            a2=get_float_after_id(line[i+1], "fpp =");
            if(nwave<10)fprintf(fout,"1 %d  %s ? %.4f %.4f\n", nwave, atom, a1, a2);
            i=i+1;
            
        }else if (strstr_space(line[i],"Space group name:")){
            
            get_data_set_id(line[i], ":", "\n", tmp);
            if(strlen(tmp)>1) strcpy(symmetry.space_group_name_H_M, tmp);
            
        }else if( strstr_space(line[i],"Space group number:")){
            n=get_int_after_id(line[i], ":");
            if(n>0)symmetry.Int_Tables_number=n;
            
        }else if(strstr_space(line[i],"Cell Dimensions")){
            n=sscanf(line[i+1],"%f%f%f%f%f%f", &dat[0],&dat[1],&dat[2],
                     &dat[3],&dat[4],&dat[5]);
            if(n==6){
                cell.length_a = dat[0];
                cell.length_b = dat[1];
                cell.length_c = dat[2];
                cell.angle_alpha = dat[3];
                cell.angle_beta = dat[4];
                cell.angle_gamma = dat[5];
            }
            
        
        }else if(nfom==0 &&
                 strstr_space(line[i], "Figure of merit vs. 1/Res^2-centric, acentric, overall:")){
            nbin=0;
            for(j=i+1; j<nline; j++){
                n=sscanf(line[j],"%s %f%f%*s%d%f %d%f %d%f",
                     tmp, &a1,&a2,&n1, &a3, &n2 , &a4, &n3, &a5);
                
                if(n==9){
                    nbin++;
                    if(nbin==1){
                        print_mmcif_head(fout, "_pdbx_phasing_MAD_shell", phasing_shell, 8);
                        nfom=1;
                    }
                    fprintf(fout,"%5.2f %5.2f %4d %6.3f %4d %6.3f %4d %6.3f\n",
                            a2,a1,n1, a3,n2,a4,n3,a5);
                
                }else if(n==1 && strstr(line[j],"TOTAL")){
                    m=sscanf(line[j+1]," %f%f %d%f %d%f %d%f",
                             &a1,&a2,&n1, &a3, &n2 , &a4, &n3, &a5);
                    if(m==8){
                    
                        print_mmcif_head(fout, "_pdbx_phasing_MAD", phasing, 9);
                
                        fprintf(fout,"%s %5.2f %5.2f %4d %6.3f %4d %6.3f %4d %6.3f\n",
                            DATA_BLOCK, a2,a1,n1, a3,n2,a4,n3,a5);
                        break;
                    }
                }
            }
            
        }else if(pattern5(line[i], "Dataset ", "R Kraut vs. 1/Res^2-centric, acentric, overall,", "", "","")){
            get_phase_stat(RK, line, nline,i,nrk,  &j);
            i=j;
            nrk++;
        }else if(pattern5(line[i], "Dataset ", "R Cullis vs. 1/Res^2-centric, acentric, overall,", "", "","")){
            get_phase_stat(RC, line, nline, i, nrc,&j);
            i=j;
            nrc++;
        }else if(pattern5(line[i], "Dataset ", "Phasing Power vs. 1/Res^2-centric, acentric, overall,", "", "","")){
            get_phase_stat(PP, line, nline,i, npp,  &j);
            i=j;
            npp++;
        }else if(npp>0 && nrk>0 && i<nline-4&& strstr(line[i],"Atom ") && strstr_space(line[i+1],"Site #") &&
                 strstr_space(line[i+2]," X      Y      Z")&&
                 strstr_space(line[i+4],"O        B")){
            sscanf(strstr(line[i],"Atom ")+6, "%s", atom);
            sscanf(strstr(line[i+1],"Site #")+6, "%d", &n);
            sscanf(line[i+3], "%f%f%f", &x, &y, &z);
            sscanf(line[i+5], "%f%f", &occ, &b);
            if(n==1)
                write_mmcif_head(fout,"_pdbx_phasing_MAD_set_site",_pdbx_phasing_MAD_set_site);
            fprintf(fout, "%d . %s . . . %8.3f %8.3f %8.3f %8.3f %8.3f\n",
                   n, atom,  x, y,z, occ,b);
        }
    }

    if(nrk==nrc && nrc== npp && nrk>0){
        write_mmcif_head(fout,"_pdbx_phasing_MAD_set_shell",_pdbx_phasing_MAD_set_shell);
        for(i=0; i<nrk; i++){
            for(j=0; j<RK[i].nbin; j++){
                fprintf(fout, "%s %5.2f %5.2f  %6.3f %6.3f %6.3f  %6.3f %6.3f %6.3f  %6.3f %6.3f %6.3f  %4d %4d %4d . . . . . .  \n",
                       RK[i].name, RK[i].resh_sh[j], RK[i].resl_sh[j],
                       PP[i].item_sh[j], PP[i].item_a_sh[j], PP[i].item_c_sh[j],
                       RC[i].item_sh[j], RC[i].item_a_sh[j], RC[i].item_c_sh[j],
                       RK[i].item_sh[j], RK[i].item_a_sh[j], RK[i].item_c_sh[j],
                       RK[i].ref_sh[j],  RK[i].ref_a_sh[j],  RK[i].ref_c_sh[j]);
            }
        }
        write_mmcif_head(fout,"_pdbx_phasing_MAD_set",_pdbx_phasing_MAD_set);
        for(i=0; i<nrk; i++){
            fprintf(fout, "%s %5.2f %5.2f . %6.3f %6.3f %6.3f  %6.3f %6.3f %6.3f  %6.3f %6.3f %6.3f  %4d %4d %4d . . . . . .  \n",
                    RK[i].name, RK[i].resh, RK[i].resl,
                    PP[i].item, PP[i].item_a, PP[i].item_c,
                    RC[i].item, RC[i].item_a, RC[i].item_c,
                    RK[i].item, RK[i].item_a, RK[i].item_c,
                    RK[i].ref,  RK[i].ref_a,  RK[i].ref_c);
        }
        
    }
    
    if(nline && line)free_memory_2d(line, nline);
    
    
}


void CNS_phase_file_assign(FILE *fout, char *inpfile,char *format,char *method)
{
    char str[512];
    FILE *fp;
    
    if((fp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s(routine:CNS_phase_file_assign)\n",
               inpfile);
        return;
    }
/* Note: if CIF, write everythings; if PDB write heavy atom cood.*/
    
    if(!strcmp_case(format, "CIF") || !strcmp_case(format, "PDB")){
        write_cif_or_pdb(fout,  inpfile, format,method);
        return;
    }
        
    while (fgets(str, sizeof str, fp) != NULL) {
        upperstr(str);
        
        if( pattern5(str, "==>}", "SITE.X_", ";", "", "") >0 ){
            fclose(fp);
            CNS_heavy_cood(fout, inpfile, method);
            break;
        }else if(strstr_space(str, "==== lack-of-closure expression") >0){
            fclose(fp);
            CNS_phasing_summary(fout, inpfile, method);
            break;
        }else if(strstr_space(str, "From Friedel(F)->Fref LOCs f' f\'' type")){
            fclose(fp);
            CNS_phasing_wave(fout, inpfile, method);
            break;
        }
    }
    
}

void CNS_phasing_wave(FILE *fout, char inpfile[], char method[])
/* extract wavelength */
{
    FILE *fp;

    char str[MAXL], *tmp=NULL, **nwave=NULL, type[MAXL];
    int i, j;
    float a1, a2;
    
    if(!strcmp_case(method,"MIR") || !strcmp_case(method,"SIR"))return;
    

    if((fp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s (routine:CNS_phasing_wave)\n",
               inpfile);
        return;
    }
    
    i=0;
    nwave = new char*[50];
    while (fgets(str, sizeof str, fp) != NULL) {
        if(strstr_space(str, "From Friedel(F)->Fref LOCs     f'             f''    type")) {
            while (fgets(str, sizeof str, fp) != NULL) {
                if(is_space_line(str)==0 || strstr(str, "=========")) continue;
                if(pattern5(str,"wavelength ", "(",")",".",".")){
                    tmp = get_substr(str, ")", "\n");
                    if(sscanf(tmp, "%f %f %s", &a1, &a2, type)==3){
                        nwave[i] = new char [strlen(str)];
                        sprintf(nwave[i], "%.3f %.3f  %s",a1, a2, type);
                        i++;
                        if(i>=20){
                            printf("\nWarning! There are too wavelengthes!\n");
                            fclose(fp);  
                            return;
                        }
                    }
                    delete [] tmp;
                }else
                    break;
            }
        }
    }
    
    if(i>0){
        
        fprintf(fout,"\n\n#++++++ MAD wavelength+++++++++++++\n" );    
        fprintf(fout,"loop_\n");
        fprintf(fout,"_phasing_%s_set.expt_id\n",method);
        fprintf(fout,"_phasing_%s_set.clust_id\n",method);
        fprintf(fout,"_phasing_%s_set.set_id\n",method);
        fprintf(fout,"_phasing_%s_set.pdbx_f_prime_refined\n",method);
        fprintf(fout,"_phasing_%s_set.pdbx_f_double_prime_refined\n",method);
        fprintf(fout,"_phasing_%s_set.pdbx_atom_type\n",method);
        for(j=0;j<i; j++){
            fprintf(fout,"1 '%d wavelength' %d  %s\n", i,  j+1,   nwave[j]);
        }
        
    }
    
    free_memory_2d(nwave, i-1);
    
}

void CNS_heavy_cood(FILE *fout, char inpfile[], char method[])
/* extract the Heavy atom coordinates */
{
    FILE *fp;
    long i, n[10], yes=0;
    char str[MAXL], substr[MAXL], atomtype[80],  meth[80];
    
    
    if((fp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s (routine:CNS_heavy_cood)\n",inpfile);
        return;
    }
    
    fprintf(fout,"\n\n### %s heavy atom coordinates from CNS ####\n", method);

    while (fgets(str, sizeof str, fp) != NULL) {
        upperstr(str);
        if(pattern5(str, "{===>}", "SITE.TYPE_", ";", "", "") >0 ){
            yes=1;
            break;
        }
    }
    if(yes<=0) return;
    
    rewind(fp);
    fprintf(fout,"loop_\n");
    
    if(!strcmp_case(method,"MAD")){
        strcpy(meth,"_pdbx_phasing_MAD_set_site");
        fprintf(fout,"%s.id\n",meth);
    } else if(!strcmp_case(method,"MIR")){
        strcpy(meth,"_phasing_MIR_der_site");
        fprintf(fout,"%s.der_id\n",meth);
    }
    
    fprintf(fout,"%s.atom_type_symbol\n",meth);
    fprintf(fout,"%s.Cartn_x\n",meth);
    fprintf(fout,"%s.Cartn_y\n",meth);
    fprintf(fout,"%s.Cartn_z\n",meth);
    fprintf(fout,"%s.B_iso\n",meth);
    fprintf(fout,"%s.Occupancy\n",meth);

    for(i=0; i<10; i++) n[i]=0;
    
    while (fgets(str, sizeof str, fp) != NULL) {
        upperstr(str);
        if(pattern5(str, "{===>}", "site.","=",";", "")  ){

            if(strstr(str, "SITE.TYPE_")){
                n[0]++;
                
                get_data_set_id(str, "SITE.TYPE_", ";", substr);
                get_data_set_id(substr, "\"", "\"", atomtype);
                rid_of_space(atomtype);
                if(strlen(atomtype)<=0) strcpy(atomtype, ".");
                fprintf(fout,"%ld %s  ",n[0], atomtype);
            }
            
            if(strstr(str, "SITE.X_")){
                write_CNS_heavy_xyz(fout, str, "SITE.X_", ";");
                n[1]++;
            }
            if(strstr(str, "SITE.Y_")){
                write_CNS_heavy_xyz(fout, str, "SITE.Y_", ";");
                n[2]++;
            }
            if(strstr(str, "SITE.Z_")){
                write_CNS_heavy_xyz(fout, str, "SITE.Z_", ";");
                n[3]++;
            }
            if(strstr(str, "SITE.B_")){
                write_CNS_heavy_xyz(fout, str, "SITE.B_", ";");
                n[4]++;
            }
            if(strstr(str, "SITE.Q_")){
                write_CNS_heavy_xyz(fout, str, "SITE.Q_", ";");
                n[5]++;
            }
        }

        if(strstr(str, "{") && strstr(str, "}") && strstr(str, "CHOICE:")&&
           (strstr(str, "REFINE")||strstr(str, "FIX")||strstr(str, "IGNORE"))){
            if(n[0]==n[1] && n[0]==n[2] && n[0]==n[3] && n[0]==n[4] &&
               n[0]==n[5]){
                if(n[0]>0) fprintf(fout,"\n");
            }else{
                fprintf(fout,"? ? ? ? ? ? ? \n");
            }
        }
        
    }
    if(n[0]<=0 || n[1]<=0) fprintf(fout,"? ? ? ? ? ? ? \n");
    
}

void write_CNS_heavy_xyz(FILE *fout, char *str, const char *start, const char *end)
{
    char *substr=NULL, *tmp=NULL;

    substr = new char[strlen(str)+1];
    tmp = new char[strlen(str)+1];

    get_data_set_id(str, start, end, substr);
    
    if(strstr(substr, "=")){
        strcpy(tmp, strstr(substr,"=")+1);
        rid_of_space(tmp);
        fprintf(fout,"%s  ",tmp);
    }else
        fprintf(fout,".  ");
    
    delete [] substr;
    delete [] tmp;

}

void CNS_phasing_summary(FILE *fout, char *inpfile, char *method)
/* extract information from the phase refinement (phase.summary) */
{
    char **phase_set=NULL, str[512], tmp[512] ;
    float **r_cullis=NULL, **r_kraut=NULL, **loc=NULL, **ppower=NULL, **fom=NULL;
    int i, j=0, nset, nsh;
    FILE *fp;

    if((fp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s (routine:CNS_phasing)\n",inpfile);
        return;
    }

    fprintf(fout,"\n\n#++++++ %s CNS phasing summary+++++++++++++\n",method);

    CNS_fom(fout,  inpfile, method);
    
    phase_set = get_phase_set(fp, inpfile,  &nset);
    float (*items_all)[20]=new float [nset+1][20]; /*  overall property*/
    
    for(i=0; i<nset; i++){
        
        while(fgets(str, sizeof str, fp) != NULL){
            if(pattern5(str,"====","lack-of-closure","expression", ": ","")){
                strcpy(tmp, strstr(str,": ")+2);    
                rid_of_space(tmp);
                if(!strcmp_case(tmp, phase_set[i]))
                    break;
            }
        }
        if(!strcmp_case(tmp, phase_set[i])){
//            printf("expression= %d %s\n", i, phase_set[i]);
                
            while(fgets(str, sizeof str, fp) != NULL){
            
                if(strstr_space(str, " generalized Cullis R-value")){
                    r_cullis = cns_phase_item(fp, &nsh);
                }else if(strstr_space(str, " Kraut R-value")){
                    r_kraut = cns_phase_item(fp, &nsh);
                }else if(strstr_space(str, " std of lack-of-closure")){
                    loc = cns_phase_item(fp, &nsh);
                }else if(strstr_space(str, " phasing power")){
                    ppower = cns_phase_item(fp,  &nsh);
                }else if(strstr_space(str, " figure-of-merit")){
                    fom = cns_phase_item(fp,  &nsh);
                }else if(strstr_space(str, "== lack-of-closure expression:")||
                         strstr_space(str, "== lack-of-closure expression :")){
                    break;
                }
                if(strstr_space(str, "== lack-of-closure expression:")||
                   strstr_space(str, "== lack-of-closure expression :")||
                   strstr_space(str,"figure of merit after final cycle"))
                    break;
            }
            rewind(fp);
            if(i==0 && nsh>1 && r_cullis[0][1] >0)
                phase_set_head(fout, method, (char *) "set_shell");
            if( r_cullis[0][1]<=0 ) continue;
            
            for(j=0; j<nsh-1; j++){
                fprintf(fout,"%s %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f ",
                       phase_set[i], r_cullis[j][0], r_cullis[j][1],
                       r_cullis[j][2], r_cullis[j][3], r_cullis[j][4],
                       r_kraut[j][2], r_kraut[j][3], r_kraut[j][4]);
                
                fprintf(fout,"%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f\n", 
                       loc[j][2], loc[j][3], loc[j][4],
                       ppower[j][2], ppower[j][3], ppower[j][4],
                       fom[j][2], fom[j][3], fom[j][4]);
            }
        }
        j=nsh-1;
        if(r_cullis[j][0]>0 && r_cullis[j][1]>0){
            
            items_all[i][0]= r_cullis[j][0];
            items_all[i][1]= r_cullis[j][1];
            items_all[i][2]= r_cullis[j][2];
            items_all[i][3]= r_cullis[j][3];
            items_all[i][4]= r_cullis[j][4];
            items_all[i][5]= r_kraut[j][2];
            items_all[i][6]= r_kraut[j][3];
            items_all[i][7]= r_kraut[j][4];
            items_all[i][8]= loc[j][2];
            items_all[i][9]= loc[j][3];
            items_all[i][10]= loc[j][4];
            items_all[i][11]= ppower[j][2];
            items_all[i][12]= ppower[j][3];
            items_all[i][13]= ppower[j][4];
            items_all[i][14]= fom[j][2];
            items_all[i][15]= fom[j][3];
            items_all[i][16]= fom[j][4];
        }
        
    }
    if(r_cullis[j][0]<=0 ||  r_cullis[j][1]<=0) return;
        
    phase_set_head(fout, method,(char *) "set");
    for(i=0; i<nset; i++){
        if(!strcmp_case(method,"MIR")){
            fprintf(fout,"%s %d  1 %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f ",phase_set[i],i+1,
                    items_all[i][0],items_all[i][1],items_all[i][2],items_all[i][3],
                    items_all[i][4],items_all[i][5],items_all[i][6],items_all[i][7]);
            
        }else {
            fprintf(fout,"%s %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f ",phase_set[i],
                    items_all[i][0],items_all[i][1],items_all[i][2],items_all[i][3],
                    items_all[i][4],items_all[i][5],items_all[i][6],items_all[i][7]);
        }
        
        
        fprintf(fout,"%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f\n", 
                items_all[i][8],items_all[i][9],items_all[i][10],
                items_all[i][11],items_all[i][12],items_all[i][13],
                items_all[i][14],items_all[i][15],items_all[i][16]);
    }
    delete [] r_cullis; delete [] r_kraut; delete []loc;
    delete []ppower; delete []fom; delete []items_all;
    
}

float **cns_phase_item(FILE *fp,  int *nsh)
/* for each item (r_cullis ..), only phase set is extracted.
   Ending with --------
*/
{
    char str[512];
    int n, ns=0 ;
    float **shell=NULL, sh0,sh1,sh2, sh3,sh4;
    const char dash[]= "------------------------------------------";

    shell = new float *[NBIN];
    
    while(fgets(str, sizeof str, fp) != NULL){
        if(strstr_space(str,"phase set")){
            break;
        }
    }
        
    while(fgets(str, sizeof str, fp) != NULL){
        if(strstr(str, dash)){
            break;
        }
        
        n=sscanf(str, "%*d %f %f %*f %f %f %f", &sh0,&sh1,&sh2,&sh3,&sh4);
        if(n==5){
            shell[ns]=new float[6];
            shell[ns][0]=sh0;
            shell[ns][1]=sh1;
            shell[ns][2]=sh2;
            shell[ns][3]=sh3;
            shell[ns][4]=sh4;
            
            if(ns>=NBIN-1){
                printf("Statistics for resolution shell is not extracted!\n");
                *nsh =0;
                return (shell);
            }
            ns++;
        }
    }
    *nsh = ns;
    return (shell);
    
}

char **get_phase_set(FILE *fp, char *inpfile, int *nset)
/* get the phase set */
{
    char str[512],tmp[512], **strings=NULL;
    int  n=0, nphs=40; //max 40 derivatives
    
    strings = new char *[nphs];
    while(fgets(str, sizeof str, fp) != NULL){
        if(pattern5(str,"====", "lack-of-closure", "expression", ": ", "")){
            strcpy(tmp, strstr(str,": ")+2);
            strings[n] = new char[strlen(tmp)+1];
            rid_of_space(tmp);
            strcpy(strings[n], tmp);
            
            if(n>=nphs-1){
                printf("The LOG file (%s) may not be checked.\n", inpfile);
                *nset =0;
                return (strings);
            }
            n++;
        }
    }
    *nset = n;
    rewind(fp);
    return (strings);
}

void phase_set_head(FILE *fout, char *method, char *category)
/* print head */
{
    char meth[60];

    
    fprintf(fout,"\nloop_\n");
    
    if(!strcmp_case(method,"MAD")){
        if(!strcmp_case(category, "set_shell"))
            strcpy(meth,"_pdbx_phasing_MAD_set_shell");
        else 
            strcpy(meth,"_pdbx_phasing_MAD_set");

        fprintf(fout,"%s.id \n",meth);
        fprintf(fout,"%s.d_res_high \n",meth);
        fprintf(fout,"%s.d_res_low \n",meth);
        fprintf(fout,"%s.R_cullis_centric \n",meth);
        fprintf(fout,"%s.R_cullis_acentric \n",meth);
        fprintf(fout,"%s.R_cullis \n",meth);
        fprintf(fout,"%s.R_kraut_centric \n",meth);
        fprintf(fout,"%s.R_kraut_acentric \n",meth);
        fprintf(fout,"%s.R_kraut \n",meth);
        fprintf(fout,"%s.loc_centric \n",meth);
        fprintf(fout,"%s.loc_acentric \n",meth);
        fprintf(fout,"%s.loc \n",meth);
        fprintf(fout,"%s.power_centric \n",meth);
        fprintf(fout,"%s.power_acentric \n",meth);
        fprintf(fout,"%s.power \n",meth);
        fprintf(fout,"%s.fom_centric \n",meth);
        fprintf(fout,"%s.fom_acentric \n",meth);
        fprintf(fout,"%s.fom \n",meth);

    }else if(!strcmp_case(method,"MIR")){

        if(!strcmp_case(category, "set_shell")){
            strcpy(meth,"_phasing_MIR_der_shell");
            fprintf(fout,"%s.der_id \n",meth);
            
            fprintf(fout,"%s.d_res_high \n",meth);
            fprintf(fout,"%s.d_res_low \n",meth);
            fprintf(fout,"%s.pdbx_R_cullis_centric \n",meth);
            fprintf(fout,"%s.pdbx_R_cullis_acentric \n",meth);
            fprintf(fout,"%s.R_cullis \n",meth);
            fprintf(fout,"%s.pdbx_R_kraut_centric \n",meth);
            fprintf(fout,"%s.pdbx_R_kraut_acentric \n",meth);
            fprintf(fout,"%s.R_kraut \n",meth);
            fprintf(fout,"%s.pdbx_loc_centric \n",meth);
            fprintf(fout,"%s.pdbx_loc_acentric \n",meth);
            fprintf(fout,"%s.loc \n",meth);
            fprintf(fout,"%s.pdbx_power_centric \n",meth);
            fprintf(fout,"%s.pdbx_power_acentric \n",meth);
            fprintf(fout,"%s.power \n",meth);
            fprintf(fout,"%s.pdbx_fom_centric \n",meth);
            fprintf(fout,"%s.pdbx_fom_acentric \n",meth);
            fprintf(fout,"%s.fom \n",meth);
        

        }else{
            
            strcpy(meth,"_phasing_MIR_der");
            fprintf(fout,"%s.id \n",meth);
            fprintf(fout,"%s.der_set_id \n",meth);
            fprintf(fout,"%s.native_set_id \n",meth);
            
            fprintf(fout,"%s.d_res_high \n",meth);
            fprintf(fout,"%s.d_res_low \n",meth);
            fprintf(fout,"%s.R_cullis_centric \n",meth);
            fprintf(fout,"%s.R_cullis_acentric \n",meth);
            fprintf(fout,"%s.pdbx_R_cullis \n",meth);
            fprintf(fout,"%s.pdbx_R_kraut_centric \n",meth);
            fprintf(fout,"%s.pdbx_R_kraut_acentric \n",meth);
            fprintf(fout,"%s.pdbx_R_kraut \n",meth);
            fprintf(fout,"%s.pdbx_loc_centric \n",meth);
            fprintf(fout,"%s.pdbx_loc_acentric \n",meth);
            fprintf(fout,"%s.pdbx_loc \n",meth);
            fprintf(fout,"%s.power_centric \n",meth);
            fprintf(fout,"%s.power_acentric \n",meth);
            fprintf(fout,"%s.pdbx_power \n",meth);
            fprintf(fout,"%s.pdbx_fom_centric \n",meth);
            fprintf(fout,"%s.pdbx_fom_acentric \n",meth);
            fprintf(fout,"%s.pdbx_fom \n",meth);
        }
        
        
    }
}


void CNS_fom(FILE *fout, char inpfile[], char method[])
/* get the figure of merit for global phase set from CNS */
{
    char str[512],meth[80];
    int nf, n, nbin, k=0;
    float v1,v2,v4,v5,v6;
    FILE *fp;
    
    const char dash[]= "---------------------------------------";

        
    if((fp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s (routine:CNS_fom)\n",inpfile);
        return;
    }

    while (fgets(str, sizeof str, fp) != NULL) { /* for overall shell*/
        if (strstr_space(str,"Figure of merit for global phase set") >0 ){
           
            while (fgets(str, sizeof str, fp) != NULL){
                if(pattern5(str, "bin", "resolution range", "refl", "|" ,"") >0 )
                    break;
            }
            
            if(!strcmp_case(method,"MAD")){
                strcpy(meth,"_pdbx_phasing_MAD_shell");
                
                fprintf(fout,"\nloop_\n");
                fprintf(fout,"%s.d_res_high \n",meth);
                fprintf(fout,"%s.d_res_low  \n",meth);
                fprintf(fout,"%s.Reflns    \n",meth);
                fprintf(fout,"%s.fom_centric \n",meth);
                fprintf(fout,"%s.fom_acentric \n",meth);
                fprintf(fout,"%s.fom  \n",meth);
            }
            else if(!strcmp_case(method,"MIR")){
                strcpy(meth,"_phasing_MIR_shell");
            
                fprintf(fout,"\nloop_\n");
                fprintf(fout,"%s.d_res_high \n",meth);
                fprintf(fout,"%s.d_res_low \n",meth);
                fprintf(fout,"%s.Reflns  \n",meth);
                fprintf(fout,"%s.fom_centric \n",meth);
                fprintf(fout,"%s.fom_acentric \n",meth);
                fprintf(fout,"%s.fom  \n",meth);
            }
            
            k=0;
            while (fgets(str, sizeof str, fp) != NULL){
                if(pattern5(str, "bin", "resolution range", "ref", "" ,"") >0)
                    break;
                n=sscanf(str, "%d %f %f %d %f %f %f",&nbin,
                     &v1, &v2, &nf, &v4, &v5, &v6 );
                
                if(n==7) {
                    k++;
                    fprintf(fout,"%.2f %.2f %d %.4f %.4f %.4f\n",
                            v1, v2, nf, v4, v5, v6);
                }
            }
            if(k<=0) fprintf(fout," ? ? ? ? ? ?\n");
            break;
            
        }
    }
    
    while (fgets(str, sizeof str, fp) != NULL){  /*overall*/
        n=sscanf(str, "%d %f %f %d %f %f %f",&nbin,
                 &v1, &v2, &nf, &v4, &v5, &v6 );
        if(strstr(str, dash)) break;
        if(n==7){
            if(!strcmp_case(method,"MAD")){
                strcpy(meth,"_phasing_MAD");
            
                fprintf(fout,"\n\n");
                fprintf(fout,"%s.entry_id  '%s'\n",meth,DATA_BLOCK);
                fprintf(fout,"%s.pdbx_d_res_high  %.3f\n",meth,v1);
                fprintf(fout,"%s.pdbx_d_res_low  %.3f\n",meth,v2);
                fprintf(fout,"%s.pdbx_Reflns  %d\n",meth,nf);
                fprintf(fout,"%s.pdbx_fom_centric  %.3f\n",meth,v4);
                fprintf(fout,"%s.pdbx_fom_acentric  %.3f\n",meth,v5);
                fprintf(fout,"%s.pdbx_fom  %.3f\n",meth,v6);
            }
            else if(!strcmp_case(method,"MIR")){
                strcpy(meth,"_phasing_MIR");
                
                fprintf(fout,"\n\n");
                fprintf(fout,"%s.entry_id  '%s'\n",meth,DATA_BLOCK);
                fprintf(fout,"%s.d_res_high  %.3f\n",meth,v1);
                fprintf(fout,"%s.d_res_low  %.3f\n",meth,v2);
                fprintf(fout,"%s.Reflns  %d\n",meth,nf);
                fprintf(fout,"%s.fom_centric  %.3f\n",meth,v4);
                fprintf(fout,"%s.fom_acentric  %.3f\n",meth,v5);
                fprintf(fout,"%s.fom  %.3f\n",meth,v6);
            }
            break;
        }
    }
    fclose(fp);
}


void SHARP_phase(FILE *fout, char *inpfile, char *format, char *method)
/* extract the information from sharp.
   wavelegth, f', f", resolution, and statistics
*/
{
    
    FILE *fp;
    
    if((fp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s (routine:SHARP_phase)\n",inpfile);
        return;
    }
    
    if(!strcmp_case(format, "LOG")){        
//        SHARP_phasing_wave(fout, inpfile, method);
        SHARP_phasing_stat(fout, inpfile, method);
        SHARP_phasing_new_v2(fout, inpfile, method);
        
    }else if(!strcmp_case(format, "PDB") ){
        heavy_coord_pdb(fout, inpfile, method);
    }else if(!strcmp_case(format, "CIF")){
        write_cif(fout,  inpfile, method);
    }else{
        printf("The input format (%s) is wrong!\n", format);
        printf("Please give the correct format (-iLOG,-iCIF,-iPDB).\n");
        usage();
    }
    fclose(fp);
}




void SHARP_phasing_wave(FILE *fout, char *inpfile, char *method)
/* to extract wavelength density_modification program
   maxmum 30 wavelength
*/

{
    FILE *finp;

    long i=0;
    int n=0;
    char str[512], juck[LEN];
    char res_L[NW][LEN], res_H[NW][LEN], fp[NW][LEN], fdp[NW][LEN];
    char rfp[NW][LEN], rfdp[NW][LEN], a1[LEN],a2[LEN];

    if((finp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s(SHARP_phasing_wave)\n",inpfile);
        return;
    }

    for(n=0; n<10; n++){
        strcpy(res_L[n], "?");
        strcpy(res_H[n], "?");
        strcpy(fp[n], "?");
        strcpy(fdp[n], "?");
        strcpy(rfp[n], "?");
        strcpy(rfdp[n], "?");
    }

    
    while (fgets(str, sizeof str, finp) != NULL) {
        upperstr(str);
        if (strstr_space(str, "   WAVELENGTH-TEXT {")>0){/*experimental */
            
            while (fgets(str, sizeof str, finp) != NULL){
                upperstr(str);
                if(strstr_space(str, "   WAVELENGTH {")>0 ) break;
                if(strstr(str, "  RESOLUTION  ")){
                    if(sscanf(str, "%s  %s %s", juck, a1 ,a2)==3 ){
                        strcpy(res_L[i], a1);
                        strcpy(res_H[i], a2);
                       
                    }
                } else if(strstr(str, " ATOM_F' ") && strstr(str, " ATOM_F\" ")){
                    if(sscanf(str, "%*s %*s %s %*s %*s %s %*s",  a1 ,a2)==2 ){
                        strcpy(fp[i], a1);
                        strcpy(fdp[i], a2);
                        i++;

                    }
                    
                }
            }
        }
    }

    
    if(i>0){
        fprintf(fout,"\n\n#++++++ SHARP MAD wavelegth+++++++++++++\n" );
        fprintf(fout,"loop_\n");
        if(!strcmp_case(method,"MAD")){ 
            fprintf(fout,"_phasing_%s_set.expt_id\n",method);
            fprintf(fout,"_phasing_%s_set.clust_id\n",method);
            fprintf(fout,"_phasing_%s_set.set_id\n",method);
            fprintf(fout,"_phasing_%s_set.d_res_low\n",method);
            fprintf(fout,"_phasing_%s_set.d_res_high\n",method);
            fprintf(fout,"_phasing_%s_set.f_prime\n",method);
            fprintf(fout,"_phasing_%s_set.f_double_prime\n",method);
            for(n=0; n<i; n++){
                fprintf(fout, "1 '%ld wavelength' %4d  %s %s %s %s\n",
                       i, n+1, res_L[n],res_H[n], fp[n], fdp[n]);
            }
        }
    }
    fclose(finp);
    
  
}


void SHARP_phasing_stat(FILE *fout, char *inpfile, char *method)
/* extract phasing statistics like FOM, R_Cullis, R_Kraut, Phasing power*/
{
    int i=0, j=0;
    int nshell_rc=0,nshell_kr=0,nshell_pw=0,nitem_rc=0,nitem_kr=0,nitem_pw=0;

    char **head, deriv_id[20];
    float rc_c[NBIN][30],rc_a[NBIN][30], kr_a[NBIN][30], kr_c[NBIN][30];
    float pw_a[NBIN][30], pw_c[NBIN][30], rc_all_a[30],rc_all_c[30];
    float kr_all_a[30],kr_all_c[30], pw_all_a[30],pw_all_c[30];
    
    SHARP_stat_fom(fout, inpfile, method);
    
    head = cmatrix(0,80, 0, 30);
    SHARP_stat(inpfile,(char *) "CULLIS ",rc_c, rc_a, head, rc_all_a, rc_all_c,
               &nshell_rc, &nitem_rc);
    SHARP_stat(inpfile,(char *)"KRAUT ", kr_c, kr_a, head, kr_all_a, kr_all_c,
               &nshell_kr, &nitem_kr);
    SHARP_stat(inpfile,(char *)"POWER ", pw_c, pw_a, head, pw_all_a, pw_all_c,
               &nshell_pw, &nitem_pw);
/*       
    printf("%3d %3d  %3d %3d  %3d %3d\n", nshell_rc, nshell_kr, nshell_pw,
           nitem_rc, nitem_kr, nitem_pw ); 

    if(!(nshell_rc == nshell_kr && nshell_kr==nshell_pw && nshell_kr>=1)||
       !(nitem_rc==nitem_kr && nitem_pw==nitem_kr && nitem_kr>=1)){
        return;
    }
*/
    if(!(nshell_rc == nshell_kr && nshell_kr==nshell_pw && nshell_kr>=1)||
       !(nitem_rc>1 && nitem_kr>1 && nitem_pw>1 )){
        return;
    }
    
    fprintf(fout,"\n\n#++++++  SHARP %s phasing statistics +++++++++\n",method);
    
    fprintf(fout,"\nloop_\n");
    
    if(!strcmp_case(method,"MIR")){    
        fprintf(fout,"_phasing_%s_der.der_id  \n",method);
        fprintf(fout,"_phasing_%s_der.d_res_low \n",method);
        fprintf(fout,"_phasing_%s_der.d_res_high \n",method);
        fprintf(fout,"_phasing_%s_der.R_cullis_centric \n",method);
        fprintf(fout,"_phasing_%s_der.R_cullis_acentric \n",method);
        fprintf(fout,"_phasing_%s_der.pdbx_R_kraut_centric \n",method);
        fprintf(fout,"_phasing_%s_der.pdbx_R_kraut_acentric \n",method);
        fprintf(fout,"_phasing_%s_der.power_centric \n",method);
        fprintf(fout,"_phasing_%s_der.power_acentric \n",method);
    }
    else if(!strcmp_case(method,"MAD")){
        fprintf(fout,"_pdbx_phasing_%s_set.id  \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set.d_res_low \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set.d_res_high \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set.R_cullis_centric \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set.R_cullis_acentric \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set.R_kraut_centric \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set.R_kraut_acentric \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set.power_centric \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set.power_acentric \n",method);
    }
    
    
    for(i=0; i<nitem_pw-3; i++){ /*overall*/
        strcpy(deriv_id,head[i+3]+5);
        fprintf(fout,"%s  ? %7.2f %7.3f %7.3f  %7.3f %7.3f  %7.3f %7.3f\n",
                deriv_id, rc_c[nshell_pw-1][2], rc_all_c[i], 
                rc_all_a[i], kr_all_c[i], kr_all_a[i], pw_all_c[i], pw_all_a[i]);
        
    }
    
    fprintf(fout,"\n#\n");
    
    fprintf(fout,"\nloop_\n");
    
    if(!strcmp_case(method,"MIR")){    
        fprintf(fout,"_phasing_%s_der_shell.der_id  \n",method);
        fprintf(fout,"_phasing_%s_der_shell.d_res_low \n",method);
        fprintf(fout,"_phasing_%s_der_shell.d_res_high \n",method);
        fprintf(fout,"_phasing_%s_der_shell.pdbx_R_cullis_centric \n",method);
        fprintf(fout,"_phasing_%s_der_shell.pdbx_R_cullis_acentric \n",method);
        fprintf(fout,"_phasing_%s_der_shell.pdbx_R_kraut_centric \n",method);
        fprintf(fout,"_phasing_%s_der_shell.pdbx_R_kraut_acentric \n",method);
        fprintf(fout,"_phasing_%s_der_shell.pdbx_power_centric \n",method);
        fprintf(fout,"_phasing_%s_der_shell.pdbx_power_acentric \n",method);
    }else if(!strcmp_case(method,"MAD")){
    
        fprintf(fout,"_pdbx_phasing_%s_set_shell.id  \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set_shell.d_res_low \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set_shell.d_res_high \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set_shell.R_cullis_centric \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set_shell.R_cullis_acentric \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set_shell.R_kraut_centric \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set_shell.R_kraut_acentric \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set_shell.power_centric \n",method);
        fprintf(fout,"_pdbx_phasing_%s_set_shell.power_acentric \n",method);
    }
    
    
    for(i=3; i<nitem_pw; i++){ /*each shell*/
        strcpy(deriv_id,head[i]+5);
        
        for(j=0; j<nshell_pw; j++){
            if(j==0){
                fprintf(fout,"%s       ? %7.3f %7.3f %7.3f  %7.3f %7.3f  %7.3f %7.3f\n",
                       deriv_id, rc_c[j][2],    rc_c[j][i], rc_a[j][i],
                       kr_c[j][i], kr_a[j][i], pw_c[j][i], pw_a[j][i]);
            }else if(j>0){
                fprintf(fout,"%s %7.3f %7.3f %7.3f %7.3f  %7.3f %7.3f  %7.3f %7.3f\n",
                       deriv_id, rc_c[j-1][2],rc_c[j][2],   rc_c[j][i], rc_a[j][i],
                        kr_c[j][i], kr_a[j][i], pw_c[j][i], pw_a[j][i]);
            }
        }
        fprintf(fout,"\n");
        
    }

}



void SHARP_stat(char *inpfile, char *id,  float rc_c[80][30], float rc_a[80][30],
                char **head,float *all_a, float *all_c, int *nshell, int *nitem)
/* extract R_Cullis, R_Kraut, Phasing power*/
{
    FILE *fp;

    long  ia=0, na=0, ic=0, nc=0, j=0, k=0, m=0, nstr1=0, ns1=0, nh1=0;
    char  **line, str[512];

    *nshell = 0;
    *nitem =0;

    if((fp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s (routine:SHARP_stat)\n",inpfile);
        return;
    }
    
    line = cmatrix(0,300, 0, 100);
    
    
    ic=0;
    nc=0;
    while (fgets(str, sizeof str, fp) != NULL) {
        upperstr(str);
        if(strstr(str, id) && strstr(str, "STATISTICS")){
            while (fgets(str, sizeof str, fp) != NULL){
                upperstr(str);
                
                if(strstr(str, "STATISTICS"))break;
                if(strstr(str, "   CENTRICS") ){
                    
                    k=0;
                    while (fgets(str, sizeof str, fp) != NULL){
                        if(strlen(str)<30)continue;
                        if(strstr(str, "CENTRICS") || strstr(str, "-------"))break;
                        upperstr(str);
                        
                         if(strstr(str, "BINS") && strstr(str, "SPACING") ){
                            token_str(str+k, " ", &nh1, line);
                            for(j=0; j<nh1; j++)
                                strcpy(head[j+ic], line[j]);
                            m=0;
                            
                            while (fgets(str, sizeof str, fp) != NULL){
                                upperstr(str);
                                if(strlen(str)<10)continue;
                                
                                if(strstr(str, "OVERALL  ")){
                                    token_str(str+24, " ", &ns1, line);
                                    for(j=0; j<ns1; j++)
                                        all_c[j+nc] = atof(line[j]);
                                    nc=ns1+nc;
                                        /*
                                    printf("???? %d  %d  %d \n", ic, k, n);
                                        */
                                    break;
                                    
                                }
                                token_str(str+k, " ", &nstr1, line);
                                for(j=0; j<nstr1; j++){
                                    rc_c[m][j+ic]=atof(line[j]);
                                        /*
                                    strcpy(item[j+ic], line[j]);
                                    rc_c[m][j+i]=atof(item[j+i]);
                                    printf( "%3d %3d %s ", m, i+j, item[j+i]); 
                                        */
                                }
                                m++;
                                /*
                                printf( "\n");
                                */
                            }
                            ic=ic+nstr1;
                            k=24;
                            
                         }
                             
                    }
                }
            }
        }
    }
    fclose(fp);

    ia=0;
    na=0;    
    fp = fopen(inpfile, "r");
    while (fgets(str, sizeof str, fp) != NULL) {
        upperstr(str);
        if(strstr(str, id) && strstr(str, "STATISTICS")){
            while (fgets(str, sizeof str, fp) != NULL){
                upperstr(str);
                
                if(strstr(str, "STATISTICS"))break;

                else if(strstr(str, "   ACENTRICS") ){
                    
                    k=0;
                    while (fgets(str, sizeof str, fp) != NULL){
                        if(strlen(str)<30)continue;
                        if(strstr(str, "CENTRICS") || strstr(str, "-------"))break;
                        upperstr(str);
                        
                         if(strstr(str, "BINS") && strstr(str, "SPACING") ){
                            token_str(str+k, " ", &nh1, line);
                            for(j=0; j<nh1; j++)
                                strcpy(head[j+ia], line[j]);
                            m=0;
                            
                            while (fgets(str, sizeof str, fp) != NULL){
                                upperstr(str);
                                if(strlen(str)<10)continue;
                                
                                if(strstr(str, "OVERALL  ")){
                                    token_str(str+24, " ", &ns1, line);
                                    for(j=0; j<ns1; j++)
                                        all_a[j+na] = atof(line[j]);
                                    na=ns1+na;
                                        /*
                                    printf("???? %d  %d  %d \n", ia, k, na);
                                        */
                                    break;
                                    
                                }
                                token_str(str+k, " ", &nstr1, line);
                                for(j=0; j<nstr1; j++){
                                    rc_a[m][j+ia]=atof(line[j]);
                                    
                                     /*
                                    strcpy(item[j+i], line[j]);
                                    rc_a[m][j+i]=atof(item[j+i]);
                                    printf( "%3d %3d %s ", m, i+j, item[j+i]); 
                                     
                                     printf( "%3d %3d %8.3f ", m, ia+j, rc_a[m][j+ia]);
                                     */
                                }
                                m++;
                                /*
                                printf( "\n");
                                */
                            }
                            ia=ia+nstr1;
                            k=24;
                            
                         }
                             
                    }
                }
                
                
            }
        }
    }
    *nshell=m;
    *nitem=ia;
                         
    fclose(fp);
    
}

void SHARP_stat_fom(FILE *fout, char *inpfile, char *method)
/* extract FOM from SHARP output*/
{
    FILE *fp;
    int na[NBIN],nc[NBIN],n=0, na_a=0, nc_a=0, i=0;
    char str[512];
    float foma[NBIN], fomc[NBIN], resL[NBIN], resH[NBIN], foma_a, fomc_a;
    

    if((fp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s (:SHARP_stat_fom)\n",inpfile);
        return;
    }
       
    while (fgets(str, sizeof str, fp) != NULL) {
        if(strstr_space(str,"BINS  Dmin  Dmax  Nacen  FOMacen  Ncen  FOMcen")>0 ){
            
            while (fgets(str, sizeof str, fp) != NULL) {
                upperstr(str);
                if(strlen(str)<30) continue;
                
                if(strstr(str, " OVERALL ")){
                    if(strlen(str)>30)
                        sscanf(str,"%*s %d %f %d %f", &na_a, &foma_a, &nc_a, &fomc_a);
                    
                    break;
                }
                if(sscanf(str,"%*s %f %f %d %f %d %f",
                          &resL[n], &resH[n], &na[n], &foma[n], &nc[n], &fomc[n]) == 6){
                    n++;
                    if(n>299) printf("Too many bins, please increase the memory\n");
                }
            }
        }
    }
    fclose(fp);
    if(n>0){

        fprintf(fout, "###########SHARP FOM ########\n");
        
        fprintf(fout,"\nloop_\n");
        if(!strcmp_case(method,"MIR")){
            fprintf(fout,"_phasing_MIR_shell.d_res_low \n");
            fprintf(fout,"_phasing_MIR_shell.d_res_high \n");
            fprintf(fout,"_phasing_MIR_shell.reflns_acentric\n");
            fprintf(fout,"_phasing_MIR_shell.FOM_acentric \n");
            fprintf(fout,"_phasing_MIR_shell.reflns_centric\n");
            fprintf(fout,"_phasing_MIR_shell.FOM_centric \n");
        }else if (!strcmp_case(method,"MAD")){
            fprintf(fout,"_pdbx_phasing_MAD_shell.d_res_low \n");
            fprintf(fout,"_pdbx_phasing_MAD_shell.d_res_high \n");
            fprintf(fout,"_pdbx_phasing_MAD_shell.reflns_acentric\n");
            fprintf(fout,"_pdbx_phasing_MAD_shell.fom_acentric \n");
            fprintf(fout,"_pdbx_phasing_MAD_shell.reflns_centric\n");
            fprintf(fout,"_pdbx_phasing_MAD_shell.fom_centric \n");
        }
        for(i=0; i<n; i++){
            fprintf(fout,"%6.2f %6.2f %6d %6.3f %6d %6.3f\n",
                    resL[i], resH[i], na[i], foma[i], nc[i], fomc[i]);
        }
    }
    fprintf(fout,"\n\n");
//    printf("\nhello = %d  %d\n", na_a, nc_a);
    
    if((na_a>0 || nc_a>0) && n>0 ){        
        if(!strcmp_case(method,"MIR")){
            fprintf(fout, "_phasing_%s.entry_id   %s\n",method, DATA_BLOCK);
            fprintf(fout, "_phasing_%s.d_res_low   %8.3f\n",method,resL[0] );
            fprintf(fout, "_phasing_%s.d_res_high   %8.3f\n",method,  resH[n-1]);
            fprintf(fout, "_phasing_%s.reflns_acentric  %d\n",method, na_a);
            fprintf(fout, "_phasing_%s.FOM_acentric  %.3f\n",method,foma_a);
            fprintf(fout, "_phasing_%s.reflns_centric  %d\n",method, nc_a);
            fprintf(fout, "_phasing_%s.FOM_centric  %.3f\n",method, fomc_a);
        }else if (!strcmp_case(method,"MAD")){
            fprintf(fout, "_phasing_%s.entry_id   %s\n",method, DATA_BLOCK);
            fprintf(fout, "_phasing_%s.pdbx_d_res_low   %8.3f\n",method, resL[0]);
            fprintf(fout, "_phasing_%s.pdbx_d_res_high   %8.3f\n",method, resH[n-1]);
            fprintf(fout, "_phasing_%s.pdbx_reflns_acentric  %d\n",method, na_a);
            fprintf(fout, "_phasing_%s.pdbx_fom_acentric  %.3f\n",method,foma_a);
            fprintf(fout, "_phasing_%s.pdbx_reflns_centric  %d\n",method, nc_a);
            fprintf(fout, "_phasing_%s.pdbx_fom_centric  %.3f\n",method, fomc_a);
        }
    }
    
}


void SHARP_phasing_new_v2(FILE *fout, char *inpfile, char *method)
/* for version 2.0 */
{
    long i=0, j=0;
    long   nba_rc=0, nbin_rc=0,  nba_pp=0, nbin_pp=0, nba=100, nbin=200;
    double **res_H, **res_L;
    double **rc_ai, **rc_ci, **rc_aa, **rc_ca;
    long   **nrc_ai, **nrc_ci, **nrc_aa, **nrc_ca;
    double *rc_ai_all ,*rc_ci_all, *rc_aa_all,*rc_ca_all;
    long   *nrc_ai_all ,*nrc_ci_all, *nrc_aa_all, *nrc_ca_all;
    
    double **pp_ai, **pp_ci, **pp_aa, **pp_ca;
    long   **npp_ai, **npp_ci, **npp_aa, **npp_ca;
    double *pp_ai_all, *pp_ci_all, *pp_aa_all, *pp_ca_all;
    long   *npp_ai_all, *npp_ci_all, *npp_aa_all, *npp_ca_all;
    long  *nbach_rc, *nbach_pp;
    
    char *rc=(char *)"Dmin Dmax Nacen Ncen RC_acen RC_cen Nacen Ncen RC_acen RC_cen";
    char *pp=(char *)"Dmin Dmax Nacen Ncen PP_acen PP_cen Nacen Ncen PP_acen PP_cen";
    char str[512];
    FILE *fp;
    
           
    res_L=dmatrix(0, nba, 0, nbin);
    res_H=dmatrix(0, nba, 0, nbin);
    rc_ai=dmatrix(0, nba, 0, nbin);
    rc_ci=dmatrix(0, nba, 0, nbin);
    rc_aa=dmatrix(0, nba, 0, nbin);
    rc_ca=dmatrix(0, nba, 0, nbin);
    
    nrc_ai=lmatrix(0, nba, 0, nbin);
    nrc_ci=lmatrix(0, nba, 0, nbin);
    nrc_aa=lmatrix(0, nba, 0, nbin);
    nrc_ca=lmatrix(0, nba, 0, nbin);

    rc_ai_all=dvector(0, nba);
    rc_ci_all=dvector(0, nba);
    rc_aa_all=dvector(0, nba);
    rc_ca_all=dvector(0, nba);

    nrc_ai_all=lvector(0, nba);
    nrc_ci_all=lvector(0, nba);
    nrc_aa_all=lvector(0, nba);
    nrc_ca_all=lvector(0, nba);

    pp_ai=dmatrix(0, nba, 0, nbin);
    pp_ci=dmatrix(0, nba, 0, nbin);
    pp_aa=dmatrix(0, nba, 0, nbin);
    pp_ca=dmatrix(0, nba, 0, nbin);
    
    npp_ai=lmatrix(0, nba, 0, nbin);
    npp_ci=lmatrix(0, nba, 0, nbin);
    npp_aa=lmatrix(0, nba, 0, nbin);
    npp_ca=lmatrix(0, nba, 0, nbin);
    
    pp_ai_all=dvector(0, nba);
    pp_ci_all=dvector(0, nba);
    pp_aa_all=dvector(0, nba);
    pp_ca_all=dvector(0, nba);

    npp_ai_all=lvector(0, nba);
    npp_ci_all=lvector(0, nba);
    npp_aa_all=lvector(0, nba);
    npp_ca_all=lvector(0, nba);

    nbach_rc = lvector(0, nba);
    nbach_pp = lvector(0, nba);

        

    
    if((fp = fopen(inpfile, "r"))==NULL) {        
        printf("Can not open the file %s(SHARP_phasing_stat_v2)\n",inpfile);
        return;
    }
    
    while (fgets(str, sizeof str, fp) != NULL) { /* for R-Cullis */
        if(strstr_space(str,"RC_acen  : R-Cullis for acentric reflexions in this")>0 )
            break;
    }
    nba_rc=0;
    while (fgets(str, sizeof str, fp) != NULL) {
        if(strstr_space(str,"Phasing Power")>0  || 
           strstr_space(str,"Dmin : lowest resolution of the" )>0) break;
        if(pattern5(str,"BATCH", " number", ": " ,"", "")>0 ){
             nba_rc++;
             get_data_new_v2(fp,rc,res_L,res_H,
                             rc_ai, rc_ci, nrc_ai, nrc_ci,
                             rc_aa, rc_ca, nrc_aa, nrc_ca,
                             rc_ai_all, rc_ci_all, nrc_ai_all, nrc_ci_all,
                             rc_aa_all, rc_ca_all, nrc_aa_all, nrc_ca_all,
                             nba_rc, &nbin_rc);
             nbach_rc[nba_rc]=nbin_rc;
        }
    }

    rewind(fp);
    while (fgets(str, sizeof str, fp) != NULL) { /* for phasing power*/
        if(strstr_space(str,"PP_acen  : phasing power for acentric reflexions in this")>0 )
            break;
    }
    nba_pp=0;
    while (fgets(str, sizeof str, fp) != NULL) {
        if(strstr_space(str,"Cullis R-factor")>0 ||
           strstr_space(str, "BINS : resolution bin number")>0 )break;
        if(pattern5(str,"BATCH", " number", ": " ,"", "")>0 ){
            nba_pp++;
            get_data_new_v2(fp,pp,res_L,res_H,
                            pp_ai, pp_ci, npp_ai, npp_ci,
                            pp_aa, pp_ca, npp_aa, npp_ca,
                            pp_ai_all, pp_ci_all, npp_ai_all, npp_ci_all,
                            pp_aa_all, pp_ca_all, npp_aa_all, npp_ca_all,
                            nba_pp, &nbin_pp);
            nbach_pp[nba_pp]=nbin_pp;
        }
    }

    if(nba_rc<=0  && nba_pp<=0) return;
    

    fprintf(fout,"\nloop_\n");
    if(!strcmp_case(method,"MIR")){    
        fprintf(fout,"_phasing_MIR_der_shell.der_id  \n");
        fprintf(fout,"_phasing_MIR_der_shell.d_res_low \n");
        fprintf(fout,"_phasing_MIR_der_shell.d_res_high \n");
        fprintf(fout,"_phasing_MIR_der_shell.pdbx_reflns_acentric\n");
        fprintf(fout,"_phasing_MIR_der_shell.pdbx_reflns_centric \n");
        
        if(nba_rc>0 && nbin_rc>0){
            fprintf(fout,"_phasing_MIR_der_shell.pdbx_R_cullis_acentric\n");
            fprintf(fout,"_phasing_MIR_der_shell.pdbx_R_cullis_centric\n");
        }
        
        if(nba_pp>0 && nbin_pp>0){
            fprintf(fout,"_phasing_MIR_der_shell.pdbx_power_acentric \n");
            fprintf(fout,"_phasing_MIR_der_shell.pdbx_power_centric\n");
        }
        
    }
    else if(!strcmp_case(method,"MAD")){
        fprintf(fout,"_pdbx_phasing_MAD_set_shell.id  \n");
        fprintf(fout,"_pdbx_phasing_MAD_set_shell.d_res_low \n");
        fprintf(fout,"_pdbx_phasing_MAD_set_shell.d_res_high \n");
        fprintf(fout,"_pdbx_phasing_MAD_set_shell.reflns_acentric\n");
        fprintf(fout,"_pdbx_phasing_MAD_set_shell.reflns_centric\n");

        if(nba_rc>0 && nbin_rc>0){
            fprintf(fout,"_pdbx_phasing_MAD_set_shell.R_cullis_acentric \n");
            fprintf(fout,"_pdbx_phasing_MAD_set_shell.R_cullis_centric\n");
        }
        if(nba_pp>0 && nbin_pp>0){
            fprintf(fout,"_pdbx_phasing_MAD_set_shell.power_acentric\n");
            fprintf(fout,"_pdbx_phasing_MAD_set_shell.power_centric\n");
        }
        
    }
    
    if(nba_rc>0  && nba_pp<=0){ /* only Rc */    
        for(i=1; i<=nba_rc; i++){ /*each set (batch)*/
            for(j=1; j<=nbach_rc[i]; j++){ 
                fprintf(fout,"ISO_%ld %.2f %.2f   %ld  %ld  %.3f %.3f\n",
                        i,res_L[i][j], res_H[i][j], nrc_ai[i][j], nrc_ci[i][j],
                        rc_ai[i][j], rc_ci[i][j]);
            }
            
            if(!strcmp_case(method,"MAD")){ 
                for(j=1; j<=nbach_rc[i]; j++){ 
                    fprintf(fout,"ANO_%ld %.2f %.2f   %ld  %ld   %.3f %.3f\n",
                            i,res_L[i][j], res_H[i][j],  nrc_aa[i][j], nrc_ca[i][j],
                            rc_aa[i][j], rc_ca[i][j]);
                }
            }
            
        }
    }

    if(nba_pp>0  && nba_rc<=0){/* only PP */

         for(i=1; i<=nba_pp; i++){ /*each set (batch)*/
             
            for(j=1; j<=nbach_pp[i]; j++){ 
                fprintf(fout,"ISO_%ld %.2f %.2f   %ld  %ld  %.3f %.3f\n",
                        i,res_L[i][j], res_H[i][j], npp_ai[i][j], npp_ci[i][j],
                        pp_ai[i][j], pp_ci[i][j]);
            }
            if(!strcmp_case(method,"MAD")){ 

                for(j=1; j<=nbach_pp[i]; j++){ 
                    fprintf(fout,"ANO_%ld %.2f %.2f  %ld  %ld   %.3f %.3f\n",
                            i,res_L[i][j], res_H[i][j], npp_aa[i][j], npp_ca[i][j],
                            pp_aa[i][j], pp_ca[i][j]);
                } 
            }
         }
    }
     

    if(nba_rc==nba_pp &&  nba_pp>0 &&  nba_rc>0 ){ /*both Rc and PP */
        for(i=1; i<=nba_rc; i++){ 
            for(j=1; j<=nbach_rc[i]; j++){
                fprintf(fout,"ISO_%ld  %.2f %.2f   %ld  %ld    %.3f %.3f %.3f %.3f\n",
                       i, res_L[i][j], res_H[i][j], nrc_ai[i][j], nrc_ci[i][j], 
                       rc_ai[i][j], rc_ci[i][j], pp_ai[i][j], pp_ci[i][j]);
            }
            
            if(!strcmp_case(method,"MAD")){ 
                for(j=1; j<=nbach_rc[i]; j++){
                    fprintf(fout, "ANO_%ld  %.2f %.2f %ld  %ld  %.3f %.3f   %.3f %.3f\n",
                            i, res_L[i][j], res_H[i][j],nrc_aa[i][j], nrc_ca[i][j],
                            rc_aa[i][j], rc_ca[i][j], pp_aa[i][j], pp_ca[i][j]);
                }
            }
            
        }
    }


// for _phasing_MIR_der or _pdbx_phasing_MAD_set
    
    if(!strcmp_case(method,"MIR")){
        
        fprintf(fout,"\nloop_\n");

        fprintf(fout,"_phasing_MIR_der.id  \n");
        fprintf(fout,"_phasing_MIR_der.der_set_id  \n");
        fprintf(fout,"_phasing_MIR_der.native_set_id  \n");
        fprintf(fout,"_phasing_MIR_der.d_res_low \n");
        fprintf(fout,"_phasing_MIR_der.d_res_high \n");
        fprintf(fout,"_phasing_MIR_der.reflns_acentric\n");
        fprintf(fout,"_phasing_MIR_der.reflns_centric \n");

        if(nba_rc>0 && nbin_rc>0){
            fprintf(fout,"_phasing_MIR_der.R_cullis_acentric\n");
            fprintf(fout,"_phasing_MIR_der.R_cullis_centric \n");
        }
        
        if(nba_pp>0 && nbin_pp>0){
            fprintf(fout,"_phasing_MIR_der.power_acentric\n");
            fprintf(fout,"_phasing_MIR_der.power_centric \n");
        }
    
        if(nba_rc>0 && nbin_rc>0 && nba_pp<=0){ /*Rc only*/
            for(i=1; i<=nba_rc; i++){ 
                fprintf(fout,"ISO_%ld %ld 1 %.2f %.2f  %ld %ld  %.3f %.3f\n",
                        i,i,res_L[i][1], res_H[i][nbin_rc],  nrc_ai_all[i], nrc_ci_all[i],
                        rc_ai_all[i], rc_ci_all[i]);
            }
        }
    
        if(nba_pp>0 && nbin_pp>0 && nba_rc<=0){ /*PP only*/
            for(i=1; i<=nba_pp; i++){ 
                fprintf(fout,"ISO_%ld %ld 1 %.2f %.2f  %ld %ld  %.3f %.3f\n",
                        i,i,res_L[i][1], res_H[i][nbin_pp],  npp_ai_all[i], npp_ci_all[i],
                        pp_ai_all[i], pp_ci_all[i]);
            }
        }

    
        if(nba_rc==nba_pp && nbin_rc==nbin_pp && nba_pp>0 && nbin_pp>0 ){/*PP & Rc*/
            for(i=1; i<=nba_rc; i++){
                fprintf(fout, "ISO_%ld %ld 1 %.2f %.2f  %ld %ld  %.3f %.3f %.3f %.3f\n", i,i,
                        res_L[i][1], res_H[i][nbin_rc],  nrc_ai_all[i],nrc_ci_all[i],
                        rc_ai_all[i], rc_ci_all[i], pp_ai_all[i], pp_ci_all[i]);
            }
        
        }

        
    }else if(!strcmp_case(method,"MAD")){

        fprintf(fout,"\nloop_\n");
        fprintf(fout,"_pdbx_phasing_MAD_set.id  \n");
        fprintf(fout,"_pdbx_phasing_MAD_set.d_res_low \n");
        fprintf(fout,"_pdbx_phasing_MAD_set.d_res_high \n");
        fprintf(fout,"_pdbx_phasing_MAD_set.reflns_acentric \n");
        fprintf(fout,"_pdbx_phasing_MAD_set.reflns_centric \n");
       
        if(nba_rc>0 && nbin_rc>0){
            fprintf(fout,"_pdbx_phasing_MAD_set.R_cullis_acentric\n");
            fprintf(fout,"_pdbx_phasing_MAD_set.R_cullis_centric \n");
        }
        
        if(nba_pp>0 && nbin_pp>0){
            fprintf(fout,"_pdbx_phasing_MAD_set.power_acentric\n");
            fprintf(fout,"_pdbx_phasing_MAD_set.power_centric \n");
        }
   

    
        if(nba_rc>0 && nbin_rc>0 && nba_pp<=0){ /*Rc only*/
            for(i=1; i<=nba_rc; i++){ 
                fprintf(fout,"ISO_%ld %.2f %.2f  %ld %ld  %.3f %.3f\n",
                        i,res_L[i][1], res_H[i][nbin_rc],  nrc_ai_all[i], nrc_ci_all[i],
                        rc_ai_all[i], rc_ci_all[i]);
            }
            for(i=1; i<=nba_rc; i++){ 
                fprintf(fout,"ANO_%ld %.2f %.2f  %ld %ld  %.3f %.3f\n",
                        i,res_L[i][1], res_H[i][nbin_rc],  nrc_aa_all[i], nrc_ca_all[i],
                        rc_aa_all[i], rc_ca_all[i]);
            }
        }
    
        if(nba_pp>0 && nbin_pp>0 && nba_rc<=0){ /*PP only*/
            for(i=1; i<=nba_pp; i++){ 
                fprintf(fout,"ISO_%ld %.2f %.2f  %ld %ld  %.3f %.3f\n",
                        i,res_L[i][1], res_H[i][nbin_pp],  npp_ai_all[i], npp_ci_all[i],
                        pp_ai_all[i], pp_ci_all[i]);
            }
            for(i=1; i<=nba_pp; i++){ 
                fprintf(fout,"ANO_%ld %.2f %.2f  %ld %ld  %.3f %.3f\n",
                        i,res_L[i][1], res_H[i][nbin_pp],  npp_aa_all[i], npp_ca_all[i],
                        pp_aa_all[i], pp_ca_all[i]);
            }
        }

    
        if(nba_rc==nba_pp && nbin_rc==nbin_pp && nba_pp>0 && nbin_pp>0 ){/*PP & Rc*/
            for(i=1; i<=nba_rc; i++){
                fprintf(fout, "ISO_%ld %.2f %.2f  %ld %ld  %.3f %.3f %.3f %.3f\n", i,
                        res_L[i][1], res_H[i][nbin_rc],  nrc_ai_all[i],nrc_ci_all[i],
                        rc_ai_all[i], rc_ci_all[i], pp_ai_all[i], pp_ci_all[i]);
            }

            for(i=1; i<=nba_rc; i++){
                fprintf(fout,"ANO_%ld %.2f %.2f  %ld %ld  %.3f %.3f %.3f %.3f\n", i,
                        res_L[i][1], res_H[i][nbin_rc],  nrc_aa_all[i],nrc_ca_all[i],
                        rc_aa_all[i], rc_ca_all[i], pp_aa_all[i], pp_ca_all[i]);
            }
        
        }
    }
    
    
    free_dmatrix(res_L,0, nba, 0, nbin);
    free_dmatrix(res_H,0, nba, 0, nbin);
    free_dmatrix(rc_ai,0, nba, 0, nbin);
    free_dmatrix(rc_ci,0, nba, 0, nbin);
    free_dmatrix(rc_aa,0, nba, 0, nbin);
    free_dmatrix(rc_ca,0, nba, 0, nbin);

    free_dvector(rc_ai_all,0, nba);
    free_dvector(rc_ci_all,0, nba);
    free_dvector(rc_aa_all,0, nba);
    free_dvector(rc_ca_all,0, nba);

    free_dmatrix(pp_ai,0, nba, 0, nbin);
    free_dmatrix(pp_ci,0, nba, 0, nbin);
    free_dmatrix(pp_aa,0, nba, 0, nbin);
    free_dmatrix(pp_ca,0, nba, 0, nbin);
    
    free_dvector(pp_ai_all,0, nba);
    free_dvector(pp_ci_all,0, nba);
    free_dvector(pp_aa_all,0, nba);
    free_dvector(pp_ca_all,0, nba);
    
}

                        
                    
                    
 void get_data_new_v2(FILE *fp, char *rc, double **res_L, double **res_H, 
                     double **rc_ai, double **rc_ci, long **nrc_ai, long **nrc_ci, 
                     double **rc_aa, double **rc_ca, long **nrc_aa, long **nrc_ca, 
                     double *rc_ai_all, double *rc_ci_all, long *nrc_ai_all, long *nrc_ci_all,
                     double *rc_aa_all, double *rc_ca_all, long *nrc_aa_all, long *nrc_ca_all,
                     long nba, long *nb)
   
                    
{
    float t1, t2,  t5, t6,  t9, t10;
    long   t3=0, t4=0, t7=0, t8=0;
    char tmp[512], str[512];
    long n=0, nbin=0, nt=0;
    
//            printf("THE FILE IS:  %s\n", rc);
    
    while (fgets(str, sizeof str, fp) != NULL) {
        if(strstr_space(str,rc)>0){
            nbin=0;
            while (fgets(str, sizeof str, fp) != NULL) {
        
                n=sscanf(str, "%ld %f %f %ld %ld %f %f %ld %ld %f %f",&nt, &t1,&t2,
                         &t3,&t4,&t5,&t6,&t7,&t8,&t9,&t10);
                
                if(n==11){
                    nbin++;
                
                    res_L[nba][nbin]=t1;
                    res_H[nba][nbin]=t2;
                    
                    nrc_ai[nba][nbin]=t3;
                    nrc_ci[nba][nbin]=t4;
                    rc_ai[nba][nbin]=t5;
                    rc_ci[nba][nbin]=t6;
                            
                    nrc_aa[nba][nbin]=t7;
                    nrc_ca[nba][nbin]=t8;
                    rc_aa[nba][nbin]=t9;
                    rc_ca[nba][nbin]=t10;
                
                }
                
                if (strstr_space( str, "OVERALL")>0){
                    n=sscanf(str, "%s %ld %ld %f %f %ld %ld %f %f",tmp, &t3,&t4,
                             &t5,&t6,&t7,&t8, &t9,&t10);
                    if(n==9) {
                        nrc_ai_all[nba]=t3;
                        nrc_ci_all[nba]=t4;
                        rc_ai_all[nba]=t5;
                        rc_ci_all[nba]=t6;
                        
                        nrc_aa_all[nba]=t7;
                        nrc_ca_all[nba]=t8;
                        rc_aa_all[nba]=t9;
                        rc_ca_all[nba]=t10;
                    }
                    break;
                }
            }
            *nb=nbin;
            return;
        }
        
    }
    
}



void BNP_statistics(FILE *fout, char *inpfile, char *format, char *method);


/*************************/
void SNB_phase(FILE *fout, char *inpfile, char *format, char *method)
/* only the heavy atom coordinates  are extracted, since there is no
   phase information available
*/
{
    
    if(!strcmp_case(format, "PDB") ){
        fprintf(fout,"\n\n####### %s heavy atom coordinates ######\n", method);
        heavy_coord_pdb(fout, inpfile, method);
    }else if(!strcmp_case(format, "CIF")){
        write_cif(fout,  inpfile, method);
    }else if(!strcmp_case(format, "LOG")){
        printf("Warning! LOG file from SnB is not supported\n");
        return;
    }else{
        printf("The input format (%s) is wrong!\n", format);
        printf("Please give the correct format (-iCIF,-iPDB).\n");
        usage();
    }
    
}

void BNP_phase(FILE *fout, char *inpfile, char *format, char *method)
{

//    printf("input file=%s:: format=%s :: method=%s\n", inpfile,format,method);
    
    if(!strcmp_case(format, "PDB") ){
        fprintf(fout,"\n\n####### %s heavy atom coordinates ######\n", method);
        heavy_coord_pdb(fout, inpfile, method);
    }else if(!strcmp_case(format, "CIF")){
        write_cif(fout,  inpfile, method);
    }else if(!strcmp_case(format, "LOG")){
        BNP_statistics(fout,inpfile, format,method);
        
    }else{
        printf("The input format (%s) is wrong!\n", format);
        printf("Please give the correct format (-iCIF,-iPDB).\n");
        usage();
    }
         
}

void BNP_statistics(FILE *fout, char *inpfile, char *format, char *method)
{
    char **text=NULL,  atom[500][5], meth[80], pow_shell[800][50],setid[NW][10];
    char tmp[256],tmp1[256],tmp2[256],tmp3[256],tmp4[256];
    int   nrow=0, i, n, j,k,m, nwave=0, nxyz=0, num, nshell=0, deriv1[NW];
    int npow=0, npow_shell=0,  nfom=0, refn_shell[NW],nref_fom=0 ;
    int setnum, start=0,len;
    
    float fprime[NW],fdprime[NW], value, a1,a2,fom_shell[NW],res_shell[NW] ;
    float fom_all=0 ,pow_set[NW];
    
    float xf[500], yf[500], zf[500], occ[500], bt[500];
    float sxf[500], syf[500], szf[500], socc[500], sbt[500];
    
    float rc_c[NW],rk_a[NW], fom_c[NW], fom_a[NW], fom[NW], pow[NW];


    
    text = get_lines_from_file(inpfile, &nrow);
    
    if(nrow<5  ){
        printf("Error! The input file (%s) may be wrong or not supported.\n", inpfile);
        return;
    }
// get wavelength    
    for(i=0; i<nrow; i++){ //
        if(strstr_space(text[i],"ADDITIONAL SCATTERING FACTOR INFORMATION SUPPLIED BY USER")){
            n=0;
            for(k=i+1; k<nrow-1; k++){ //wave length
                if(pattern5(text[k], "set", "_iso", "", "", "") ||
                   strstr_space(text[k], "PHASE CALCULATIONS BASED")) {
                    i=k;
                    break;
                }
                   
                if(pattern5(text[k],"FOR ATOM TYPE ", "DELTA f PRIME =", ".", "DELTA f DOUBLE PRIME =", ".")){
                    strcpy(tmp, strchr(text[k], '=')+1);
                    if(sscanf(tmp, "%f", &value) == 1){
                        fprime[n]=value;
                        strcpy(tmp, strrchr(text[k], '=')+1);
                        if(sscanf(tmp, "%f", &value) == 1){
                            fdprime[n]=value;
                            n++;
                        }
                    }
                }
                
                if(n>=NW){
                    printf("Error! Too many wavelength!\n");
                    nwave=0;
                }
            }
            nwave=n;
        }
    }

 //write all over wavelength information
    if(nwave>0 && ( strstr_case(method, "MAD")||strstr_case(method, "SAD")) ){
        fprintf(fout,"\n\n#++++++ MAD wavelength+++++++++++++\n" );    
        fprintf(fout,"loop_\n");
        fprintf(fout,"_phasing_%s_set.expt_id\n",method);
        fprintf(fout,"_phasing_%s_set.clust_id\n",method);
        fprintf(fout,"_phasing_%s_set.set_id\n",method);
        fprintf(fout,"_phasing_%s_set.f_prime\n",method);
        fprintf(fout,"_phasing_%s_set.f_double_prime\n",method);
    
        for(i=0; i<nwave; i++){
            fprintf(fout, "1 '%d wavelength' %d   %.2f   %.2f\n",
                    nwave,i+1, fprime[i],  fdprime[i]);
        }
    }

    
    n=0;
    for(i=nrow-1; i>=0; i--){ //
        n++;
        if(strstr(text[i], "PARAMETERS AFTER LEAST SQUARES REFINEMENT")){
            start=i-1;
            break;
        }
    }
//    printf( "start=%d  %d %d\n", start, nrow, n);


    
    
    if(n<20) start=0;
    
///////////////////////////    
    for(i=start; i<nrow; i++){ //
    
        if(strstr_space(text[i],"PARAMETERS AFTER LEAST SQUARES REFINEMENT")){
            n=0;
            for(k=i+1; k<nrow-6; k++){ //fractional xyz
                if(strstr_space(text[k], "POST REFINEMENT SUMMARY AFTER")) {
                    i=k-1;
                    break;
                }
                
                if(strstr(text[k],"********") &&
                   pattern5(text[k+1],"* Z ", " *", "", "", "") &&
                     strstr(text[k+2],"********") &&
                   pattern5(text[k+3]," X ", ".", ".", ".", ".") &&
                   pattern5(text[k+4]," Y ", ".", ".", ".", ".") &&
                   pattern5(text[k+5]," Z ", ".", ".", ".", ".") &&
                   pattern5(text[k+6]," OCCUPANCY", ".", ".", ".", ".") &&
                   pattern5(text[k+7]," BETA ", ".", ".", ".", ".") ){
                    
                    j=sscanf(text[k+1], "%s%s%s%s%s", tmp1,tmp2,tmp3,tmp4,tmp);
                    if(j==4 && strstr(tmp1, "*") && strstr(tmp4, "*")){
                        len=strlen(tmp3);
                        for(m=0 ; m<len; m++){
                            if(isdigit(tmp3[m]))
                                continue;
                            else
                                break;
                        }
                        
                        strcpy(atom[n], tmp3+m);
                    }
                    j=sscanf(text[k+3], "%s%s%s%s%s", tmp1,tmp2,tmp3,tmp4,tmp);
                    if(j==5 && strstr(tmp1, "X")){
                        xf[n]=atof(tmp4);
                        sxf[n]=atof(tmp);
                    }
                    
                    j=sscanf(text[k+4], "%s%s%s%s%s", tmp1,tmp2,tmp3,tmp4,tmp);
                    if(j==5 && strstr(tmp1, "Y")){
                        yf[n]=atof(tmp4);
                        syf[n]=atof(tmp);
                    }
                    
                    j=sscanf(text[k+5], "%s%s%s%s%s", tmp1,tmp2,tmp3,tmp4,tmp);
                    if(j==5 && strstr(tmp1, "Z")){
                        zf[n]=atof(tmp4);
                        szf[n]=atof(tmp);
                    }
                    
                    j=sscanf(text[k+6], "%s%s%s%s%s", tmp1,tmp2,tmp3,tmp4,tmp);
                    if(j==5 && strstr(tmp1, "OCCUPANCY")){
                        occ[n]=atof(tmp4);
                        socc[n]=atof(tmp);
                    }
                    
                    j=sscanf(text[k+7], "%s%s%s%s%s", tmp1,tmp2,tmp3,tmp4,tmp);
                    if(j==5 && strstr(tmp1, "BETA")){
                        bt[n]=atof(tmp4);
                        sbt[n]=atof(tmp);
                    }
                    k=k+7;
                    n++;
                    if(n>=500){
                        printf("Error! Too many scattering atoms!\n");
                        nxyz=0;
                    } 
                }
            }
            nxyz=n;
        }
    }

// write sub-structure coordinate XYZ (fraction)

    strcpy(meth, "_pdbx_phasing_site");
    
    if(nxyz>0 ){
        fprintf(fout,"\n\n#++++++++++++++\n" );    
        fprintf(fout,"loop_\n");
        fprintf(fout,"%s.id\n", meth);
        fprintf(fout,"%s.atom_type_symbol\n", meth);
        fprintf(fout,"%s.fract_x\n", meth);
        fprintf(fout,"%s.fract_x_esd\n", meth);
        fprintf(fout,"%s.fract_y\n", meth);
        fprintf(fout,"%s.fract_y_esd\n", meth);
        fprintf(fout,"%s.fract_z\n", meth);
        fprintf(fout,"%s.fract_z_esd\n", meth);
        fprintf(fout,"%s.occupancy\n", meth);
        fprintf(fout,"%s.occupancy_esd\n", meth);
        fprintf(fout,"%s.B_iso\n", meth);
        fprintf(fout,"%s.B_iso_esd\n", meth);
        for(i=0; i<nxyz; i++){ 
             fprintf(fout,"%d '%s'  %.4f %.4f  %.4f %.4f  %.4f %.4f  %.2f %.4f  %.2f %.4f \n",
                    i+1,atom[i], xf[i],sxf[i],  yf[i],syf[i],  zf[i],szf[i],  occ[i],socc[i],  bt[i],sbt[i]);
        }
        
    }

///////////////////////////    
    
    
    for(i=start; i<nrow; i++){ 
//over all   fom         
         if(pattern5(text[i],"MEAN FIGURE OF MERIT =",  "FOR", "PHASED REFLECTIONS","","")){
            fom_all=get_float(text[i],"=","FOR");
            nref_fom = get_int(text[i],"FOR", "PHASED");
            
//over all   fom shell            
        }else if(strstr(text[i],"MEAN FIGURES OF MERIT AS FUNCTION OF RESOLUTION")&&
                 pattern5(text[i+1],"FOM", "d", "No. Refl", "", "")){
            n=0;  
            for(j=i+2; j<nrow-2; j++){
                if(sscanf(text[j], "%f%f%d", &a1, &a2, &num)==3){
                    fom_shell[n]=a1;
                    res_shell[n]=a2;
                    refn_shell[n]=num;
                    n++;
                    if(n>=NW){
                        printf("Warning! Too many data sets for fom!\n");
                        nshell=0;
                    } 
                }else{
                    break;
                }
            }
            nshell=n;
        }
    }

//write all over FOM
    fprintf(fout,"\n\n"); 
    fprintf(fout,"_phasing_%s.entry_id  '%s'\n",method, DATA_BLOCK);
    fprintf(fout,"_phasing_%s.pdbx_d_res_high  %.2f\n",method,res_shell[0]);
    fprintf(fout,"_phasing_%s.pdbx_d_res_low   %.2f\n", method,res_shell[nshell-1]);
    fprintf(fout,"_phasing_%s.pdbx_fom  %.3f\n",method,fom_all);
    fprintf(fout,"_phasing_%s.pdbx_Reflns  %d\n",method,nref_fom);


// write overall FOM with average shell (new category!!)    
    if(nshell>0){
        fprintf(fout,"\n\n#++++++++++++++\n" );    
        fprintf(fout,"loop_\n");
        fprintf(fout,"_pdbx_phasing_fom.bin \n");
        fprintf(fout,"_pdbx_phasing_fom.avg_resolution \n");
        fprintf(fout,"_pdbx_phasing_fom.fom  \n");
        fprintf(fout,"_pdbx_phasing_fom.Reflns    \n");
        for(i=0; i<nshell; i++){
            fprintf(fout,"%d  %6.2f %6.2f %4d\n",
                    i+1, res_shell[i], fom_shell[i],refn_shell[i]);
        }
    }


///////////////////////////    

    for(i=start; i<nrow; i++){ //R_cullis, R_kraut,
    
        if(strstr_space(text[i],"------  POST REFINEMENT SUMMARY AFTER PASS ")){
            n=0;
            for(j=i+1; j<nrow-2; j++){
                
                if(strstr_case(text[j],"START OF NEXT PHASIN")||
                   strstr_case(text[j],"INDIVIDUAL DATA SET RES")) {
                    i=j;
                    break;
                }
                
                
                if(pattern5(text[j], "STATISTICS FOR SET", "AFTER REFINEMENT", "","","")){
                    if(strstr(text[j+1], " = ") && strstr(text[j+1], "REFLECTIONS") &&
                       strstr(text[j+2], " = ") && strstr(text[j+2], "REFLECTIONS")){
                        
                        for(k=j+1; k<j+3; k++){ 
                            if(pattern5(text[k],"R CULLIS", " = ", "FOR", " CENTRIC","")){
                                rc_c[n] = get_float(text[k]," = ","FOR");
                            }else if(pattern5(text[k],"R KRAUT "," = ", "FOR", " ACENTRIC","")){
                                rk_a[n] = get_float(text[k],"=","FOR");
                            } 
                        }
                        n++;
                        j=j+2;
                        
                    }else if(strstr(text[j+1], " = ") && strstr(text[j+1], "REFLECTIONS") &&
                             !strstr(text[j+2], " = ") && !strstr(text[j+2], "REFLECTIONS")){
                        
                        if(pattern5(text[j+1],"R KRAUT "," = ", "FOR", " ACENTRIC","")){
                                rk_a[n] = get_float(text[j+1],"=","FOR");
                        } 
                        n++;
                        j=j+1;
                        
                    }
                }
                if(n>=NW){
                    printf("Warning! Too many data sets for rc_c, rk_a!\n");
                   
                } 
            }
           
        }
    }
    
    
    for(i=start; i<nrow; i++){ //
    
        if(strstr_case(text[i],"INDIVIDUAL DATA SET RESULTS BASED ON")){
            n=0;
            for(j=i+1; j<nrow-2; j++){
                
                if(strstr_case(text[j],"** RESULTS FROM COMBINED PROBABILITY DISTRIBUTIONS **")||
                   strstr_case(text[j],"DATA SETS CONTRIBUTED")) {
                    i=j-1;
                    break;
                }
                
                if(pattern5(text[j],   "SET ", "_", "o", "", "") &&
                   pattern5(text[j+1], "MEAN FIGURE OF MERIT", "=", "FOR", "", "REFLECTIONS")){
                    
                    if(pattern5(text[j+1],"MEAN FIGURE OF MERIT =", "FOR", "REFLECTIONS","","")&&
                       !strstr_case(text[j+1],"CENTRIC")){
                        fom[n] = get_float(text[j+1],"=","FOR");
                    }
                    if(pattern5(text[j+2],"MEAN FIGURE OF MERIT =", "FOR", " CENTRIC REFLECTIONS","","")){
                        fom_c[n] = get_float(text[j+2],"=","FOR");
                    }
                    
                    if(pattern5(text[j+3],"MEAN FIGURE OF MERIT =", "FOR", " ACENTRIC REFLECTIONS","","")){
                        fom_a[n] = get_float(text[j+3],"=","FOR");
                    }
                    
                    sscanf(text[j],"%s%d%s", tmp1, &num, tmp2);
                    if(strstr(tmp1,"SET") && strstr(tmp2,"o")){
                        deriv1[n] = num;
                        strcpy(setid[n], tmp2);
                        n++;
                        if(n>=NW){
                            printf("Warning! Too many data sets for fom!\n");
                            nfom=0;
                        } 
                    }
                }
            }
            nfom=n;
        }
    }
   
    for(i=start; i<nrow; i++){ //
    
        if (strstr(text[i],"PHASING POWER BREAKDOWN BASED ON CURRENT PROTEIN PHASES")){
            
            n=0;
            npow=0;
            
            for(j=i+1; j<nrow-2; j++){
                
                if(strstr_case(text[j]," DIFFERENCE FOURIER FILE FOR SET")) {
                    i=j;
                    break;
                }

                if(pattern5(text[j],"SET ", "_", "", "", "") &&
                    pattern5(text[j+1],"< d >", "Phasing Power", "Bias", "No. Refl", "")){
                    sscanf(text[j], "%*s%d%s",  &setnum, tmp1);
                    
                        
                    for(k=j+2; k<nrow-2; k++){
                        
                        if(sscanf(text[k], "%f%f%*s%d", &a1, &a2, &num)==3){
                            if(setnum>0) sprintf(tmp2,"%d %s %.2f %.2f %4d",
                                                 setnum, tmp1, a1, a2, num);
                            strcpy(pow_shell[n], tmp2);
                            n++;
                            if(n>=20*NW){
                                printf("Warning! Too many data sets for power!\n");
                                npow_shell=0;
                            } 
                        }else if(pattern5(text[k], "OVERALL MEAN d =", "PHASING POWER =", ".", "","")){
                            pow_set[npow]=get_float_after_id(text[k],"POWER =");
                            npow++;
                            break;
                            
                        }
                        
                    }
                }
            }
            npow_shell=n;
        }
        
    }
//    printf( "start=%d  %d  %d  %d %d %d %d %d\n",start, nrow, nxyz, nrc, nfom,npow,nshell,npow_shell );

    
// write FOM ,R_cullis,R_kraut,POWER for each set
    

    fprintf(fout,"\nloop_\n");
    if(!strcmp_case(method,"MAD")){
        strcpy(meth,"_pdbx_phasing_MAD_set");
        
        fprintf(fout,"%s.id \n",meth);
        fprintf(fout,"%s.d_res_low \n",meth);
        fprintf(fout,"%s.d_res_high \n",meth);
        fprintf(fout,"%s.fom_centric \n",meth);
        fprintf(fout,"%s.fom_acentric \n",meth);
        fprintf(fout,"%s.fom \n",meth);
        fprintf(fout,"%s.R_cullis_centric \n",meth);
        fprintf(fout,"%s.R_kraut_acentric \n",meth);
        fprintf(fout,"%s.power \n",meth);
        
        for(i=0; i<nfom; i++){
            fprintf(fout, " '%s'", setid[i]);
            (res_shell[0]>0) ? fprintf(fout, " %.2f ", res_shell[0]) : fprintf(fout, " ? ");
            (res_shell[nshell-1]>0) ? fprintf(fout, " %.2f ", res_shell[nshell-1]) : fprintf(fout, " ? ");
            (fom_c[i]>0) ? fprintf(fout, " %.3f ", fom_c[i]) : fprintf(fout, " ? ");
            (fom_a[i]>0) ? fprintf(fout, " %.3f ", fom_a[i]) : fprintf(fout, " ? ");
            (fom[i]>0) ? fprintf(fout, " %.3f ", fom[i]) : fprintf(fout, " ? ");
            (rc_c[i]>0) ? fprintf(fout, " %.3f ", rc_c[i]) : fprintf(fout, " ? ");
            (rk_a[i]>0) ? fprintf(fout, " %.3f ",rk_a[i] ) : fprintf(fout, " ? ");
            (pow_set[i]>0) ? fprintf(fout, " %.3f ",pow_set[i] ) : fprintf(fout, " ? ");
            fprintf(fout, "\n");
            
        }
        
    }else if(!strcmp_case(method,"MIR")){
        strcpy(meth,"_phasing_MIR_der");
            
        fprintf(fout,"%s.id \n",meth);
        fprintf(fout,"%s.der_set_id \n",meth);
        fprintf(fout,"%s.native_set_id \n",meth);
        fprintf(fout,"%s.d_res_low \n",meth);
        fprintf(fout,"%s.d_res_high \n",meth);
        fprintf(fout,"%s.pdbx_fom_centric \n",meth);
        fprintf(fout,"%s.pdbx_fom_acentric \n",meth);
        fprintf(fout,"%s.pdbx_fom \n",meth);
        fprintf(fout,"%s.R_cullis_centric \n",meth);
        fprintf(fout,"%s.pdbx_R_kraut_acentric \n",meth);
        fprintf(fout,"%s.pdbx_power \n",meth);

        for(i=0; i<nfom; i++){
            fprintf(fout, " '%s' %d  1", setid[i], deriv1[i]);
            (res_shell[0]>0) ? fprintf(fout, " %.2f ", res_shell[0]) : fprintf(fout, " ? ");
            (res_shell[nshell-1]>0) ? fprintf(fout, " %.2f ", res_shell[nshell-1]) : fprintf(fout, " ? ");
            (fom_c[i]>0) ? fprintf(fout, " %.3f ", fom_c[i]) : fprintf(fout, " ? ");
            (fom_a[i]>0) ? fprintf(fout, " %.3f ", fom_a[i]) : fprintf(fout, " ? ");
            (fom[i]>0) ? fprintf(fout, " %.3f ", fom[i]) : fprintf(fout, " ? ");
            (rc_c[i]>0) ? fprintf(fout, " %.3f ", rc_c[i]) : fprintf(fout, " ? ");
            (rk_a[i]>0) ? fprintf(fout, " %.3f ",rk_a[i] ) : fprintf(fout, " ? ");
            (pow[i]>0) ? fprintf(fout, " %.3f ",pow[i] ) : fprintf(fout, " ? ");
            
        }
    }


 // write POWER for each set with shell (new category!!)

    if(npow_shell>0){
        strcpy(meth,"_pdbx_phasing_power_shell");
        fprintf(fout,"\nloop_\n");
        
        fprintf(fout,"%s.id \n",meth);
        fprintf(fout,"%s.set_name \n",meth);
        fprintf(fout,"%s.avg_resolution \n",meth);
        fprintf(fout,"%s.power \n",meth);
        fprintf(fout,"%s.reflns\n",meth);
        for(i=0; i<npow_shell; i++){
            fprintf(fout,"%s\n",pow_shell[i]);
//            printf("%s\n",pow_shell[i]);
        }
    }

    if(nrow) free_memory_2d(text,nrow);
    
//    printf("8  input file=%s:: format=%s :: method=%s\n", inpfile,format,method);
    
}
    
/* ------------------------------------------------*/
void SHELX_phase(FILE *fout, char *inpfile, char *format, char *method)
/* only the heavy atom coordinates  are extracted, since there is no
   phase information available
*/
{
    
    FILE *fp;

    if((fp = fopen(inpfile, "r"))==NULL) {
        open_file_error(inpfile,"SHELX_phase");
    }

    if(!strcmp_case(format, "PDB") ){
        fprintf(fout,"\n\n####### %s heavy atom coordinates ######\n", method);
        heavy_coord_pdb(fout, inpfile, method);
    }else if(!strcmp_case(format, "CIF")){
        write_cif(fout,  inpfile, method);
    }else if(!strcmp_case(format, "LOG")){
        printf("Warning! LOG file from SHELX is not supported\n");
        return;
    }else{
        printf("The input format (%s) is wrong!\n", format);
        printf("Please give the correct format (-iCIF,-iPDB).\n");
        usage();
    }
    fclose(fp);
}


