/*
FILE:     text_parser.C
*/
/*
VERSION:  3.004
*/
/*
DATE:     9/11/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
/* the input coordinate file either is PDF format or CIF */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "rcsb_math.h"
#include "structure.h"
extern EXPTL exptl;

#include "util.h"

void values_in_bracket(FILE *inp, char *first_item, char *value, int *key);
void write_values(FILE *fout, FILE *finp, int nitem, char *id1, char *id2,
                  char *id3, char *id4, char *id5, char *id6, char *id7,
                  char *id8, char *id9);
void initialize_id(char *id1,char *id2,char *id3,char *id4,char *id5,
                   char *id6, char *id7,char *id8, char *id9);

void test_sequence_correction(char *value);
void parse_infor_common(FILE *fout, int nitem, char **items, char **values);
void parse_infor_xray(FILE *fout, int nitem, char **items, char **values);
void parse_infor_nmr(FILE *fout, char *inpfile);
void write_text_parser_cif_head(FILE *fout, char *id);
int check_template(int k,int nitem,char **items, char **values,
                  const char *id1,const char *id2);
void  token_value(FILE *fout,int k,int nitem,char **items,
                  char **values,char token[80][80], int ntoken,int key);


//----------------------------------------------------------------//

extern CELL cell;
//extern SYMMETRY symmetry;

void parse_adit_text_to_mmcif(FILE *fout, char *inpfile, int exp)
// parse information from text (in lib) to mmcif
{
    char  **values, **items,  item[80], *value;
    int i,nline=0,nitem=0,n=0,key,len;
    FILE *finp=NULL;
    
    nline = num_of_line(inpfile);

    if(nline<100){
        printf("Error! No items extracted from the template file.\n");
        return;
    }
    
    items = new char *[nline];
    values = new char *[nline];
    value = new char [10000];
    
    finp=fopen(inpfile, "r") ;
    
    while(1) { 
        values_in_bracket(finp,  item, value,  &key);
        if(key>0)break;
        len=strlen(item);
        if(len>0){
            items[n]=new char[len+1];
            strcpy(items[n], item);
            
            len=strlen(value);
            values[n]=new char[len+1];
            strcpy(values[n],value);
            
            n++;
            
        }
        strcpy(item, "");
    }
    nitem=n;
    for(i=0;i<nitem; i++){
//        printf("i=%3d  nitem=%2d, items=%s, values=%s\n",i, nitem, items[i], values[i]);
    }
    
        
    parse_infor_common(fout, nitem, items, values);
    if(exp==0){
        parse_infor_nmr(fout,inpfile );
    }else{
        parse_infor_xray(fout, nitem, items, values);
    }
    
    delete [] value;
    for(i=0; i<nitem; i++){
        delete [] items[i];
        delete [] values[i];
    }
    fclose(finp);
}



void  token_value(FILE *fout,int k,int nitem,char **items,
                  char **values,char token[80][80], int ntoken,int key)
//write the values parsed in the template file.
{
    char **val;
    int  n=0,j, i=0, len=0, m=0,nloop=0,nonv=0;

    val=new char *[ntoken+1];
    for(j=0; j<ntoken; j++)    val[j]=(char *)"?";
    
    n=0;
    for(i=k; i<nitem; i++){
        for(j=0; j<ntoken; j++){
            
            if(!strcmp_case(items[i], token[j])){

                if(!strcmp_case(items[i],"molecule_one_letter_sequence") )
                    test_sequence_correction(values[i]);
                
                len=strlen(values[i])+1;
                val[j]=new char [len];
                strcpy(val[j], values[i]);
                n++;
            }
        }
        
        if(n==ntoken){
            if(n==1 && strlen(val[0])<=0) continue;
            nonv=0;
            for(m=0; m<ntoken; m++){
                if(strlen(val[m])==0)nonv++;
            }
            if(nonv == ntoken) continue; //all values are empty
            
            nloop++;
            for(m=0; m<ntoken; m++){
                len=strlen(val[m]);
                if(key==0)
                    (len>0) ? fprintf(fout,"\n;%s\n; ",val[m]):fprintf(fout,"\n;?\n;");
                else
                    (len>0) ? fprintf(fout,"%d \n;%s\n; ",nloop,val[m]):fprintf(fout,"\n;?\n;");
            }
            n=0;
            fprintf(fout,"\n");
            
        }
    }
}


void parse_infor_common(FILE *fout, int nitem, char **items, char **values)
// extract information common for Xray and NMR
{
    char   id[80][80], author_id[80];
    int    n, n1=0,n2, ntoken=0, i, len=0;



//CATEGORY 2:   Molecular Entity Sequence
    ntoken = 5;
    strcpy(id[0],"molecule_entity_id");
    strcpy(id[1],"molecule_entity_type");
    strcpy(id[2],"molecule_one_letter_sequence");
    strcpy(id[3],"molecule_chain_id");
    strcpy(id[4],"target_DB_id");
    n=check_template( 0, nitem, items, values,id[0], id[2]);
    if(n>0){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_entity_poly.entity_id\n");
        fprintf(fout,"_entity_poly.type\n");
        fprintf(fout,"_entity_poly.pdbx_seq_one_letter_code\n");
        fprintf(fout,"_entity_poly.pdbx_strand_id\n");
        fprintf(fout,"_entity_poly.pdbx_target_identifier\n");

        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    
//CATEGORY 3, (Contact Authors for PI & others)
    ntoken = 15;
    strcpy(id[0], "contact_author_PI_id");
    strcpy(id[1], "contact_author_PI_salutation");
    strcpy(id[2], "contact_author_PI_first_name");
    strcpy(id[3], "contact_author_PI_last_name");
    strcpy(id[4], "contact_author_PI_middle_name");    
    strcpy(id[5], "contact_author_PI_role");
    strcpy(id[6], "contact_author_PI_organization_type");
    strcpy(id[7], "contact_author_PI_email");      
    strcpy(id[8], "contact_author_PI_address");
    strcpy(id[9], "contact_author_PI_city");
    strcpy(id[10], "contact_author_PI_State_or_Province");
    strcpy(id[11], "contact_author_PI_Zip_Code");
    strcpy(id[12], "contact_author_PI_Country");
    strcpy(id[13], "contact_author_PI_fax_number");
    strcpy(id[14], "contact_author_PI_phone_numer");

    n1=check_template( 0, nitem, items, values,id[0], id[2]);
    if(n1){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_pdbx_contact_author.id\n");
        fprintf(fout,"_pdbx_contact_author.name_salutation\n");
        fprintf(fout,"_pdbx_contact_author.name_first\n");
        fprintf(fout,"_pdbx_contact_author.name_last\n");
        fprintf(fout,"_pdbx_contact_author.name_mi\n");
        fprintf(fout,"_pdbx_contact_author.role\n");
        fprintf(fout,"_pdbx_contact_author.organization_type\n");
        fprintf(fout,"_pdbx_contact_author.email\n");
        fprintf(fout,"_pdbx_contact_author.address_1\n");
        fprintf(fout,"_pdbx_contact_author.city\n");
        fprintf(fout,"_pdbx_contact_author.state_province\n");
        fprintf(fout,"_pdbx_contact_author.postal_code\n");
        fprintf(fout,"_pdbx_contact_author.country\n");
        fprintf(fout,"_pdbx_contact_author.fax\n");
        fprintf(fout,"_pdbx_contact_author.phone\n");
        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    
    ntoken = 15;
    strcpy(id[0], "contact_author_id");
    strcpy(id[1], "contact_author_salutation");
    strcpy(id[2], "contact_author_first_name");
    strcpy(id[3], "contact_author_last_name");
    strcpy(id[4], "contact_author_middle_name");    
    strcpy(id[5], "contact_author_role");
    strcpy(id[6], "contact_author_organization_type");
    strcpy(id[7], "contact_author_email");      
    strcpy(id[8], "contact_author_address");
    strcpy(id[9], "contact_author_city");
    strcpy(id[10], "contact_author_State_or_Province");
    strcpy(id[11], "contact_author_Zip_Code");
    strcpy(id[12], "contact_author_Country");
    strcpy(id[13], "contact_author_fax_number");
    strcpy(id[14], "contact_author_phone_numer");

    n2=check_template( 0, nitem, items, values,id[0], id[2]);
    if(n2){
        if(n1<=0){
            printf("You should enter the entry for PI first\n");
        } else
            token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    
//CATEGORY  (Structure Genomics)

    ntoken= 3;
    strcpy(id[0],"SG_project_id");
    strcpy(id[1],"SG_project_name");
    strcpy(id[2],"full_name_of_SG_center");

    n=check_template( 0, nitem, items, values,id[0], "");
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout," _pdbx_SG_project.id\n");
        fprintf(fout," _pdbx_SG_project.project_name\n");
        fprintf(fout," _pdbx_SG_project.full_name_of_center\n");
    
        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    

//CATEGORY 4,  (Release Status)
    n1=0;
    n2=0;
    for(i=0; i<nitem; i++){
        len = strlen(values[i]);
        if(len<0) continue;
        
        if(!strcmp_case(items[i], "Release_status_for_coordinates")){
            fprintf(fout,"_pdbx_database_status.dep_release_code_coordinates ");
            fprintf(fout,"\n;%s\n; \n",values[i]);
        }else if(strstr_case(items[i], "Release_status_for_NMR_constraints")){
            fprintf(fout,"_pdbx_database_status.dep_release_code_nmr_constraints ");
            fprintf(fout," \n;%s\n;\n",values[i]);
        }else if(strstr_case(items[i], "Release_status_for_structure_factor")){
            fprintf(fout,"_pdbx_database_status.dep_release_code_struct_fact ");
            fprintf(fout," \n;%s\n;\n",values[i]);
        }else if(strstr_case(items[i], "Release_status_for_sequence")){
            fprintf(fout,"_pdbx_database_status.dep_release_code_sequence ");
            fprintf(fout," \n;%s\n;\n",values[i]);
       
//CATEGORY 5 (Structure Title)        
        }else if(!strcmp_case(items[i], "structure_title")){  
            fprintf(fout,"\n_struct.entry_id UNNAMED \n");
            fprintf(fout,"_struct.title \n;%s\n;\n",values[i]);
        }else if(!strcmp_case(items[i], "structure_details")){  
            fprintf(fout,"\n_struct.ndb_details \n;%s\n;\n",values[i]);


//CATEGORY 12 (Keywords), 

        }else if(!strcmp_case(items[i], "structure_keywords")){
            fprintf(fout,"\n_struct_keywords.entry_id  UNNAMED \n");
            fprintf(fout,"_struct_keywords.text \n;%s\n;\n\n",values[i]);
        }
    }


//CATEGORY  (Author)
    ntoken = 1;
    strcpy(id[0],"structure_author_name");
    n=check_template( 0, nitem, items, values,id[0], "");
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_audit_author.id\n");
        fprintf(fout,"_audit_author.name\n");
        token_value(fout, 0, nitem, items, values, id, ntoken, 1);
    }
    
            
//CATEGORY 8:   Molecule Names
    ntoken = 1;
    strcpy(id[0],"molecule_name");
    n=check_template( 0, nitem, items, values,id[0], "");
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_entity.id\n");
        fprintf(fout,"_entity.pdbx_description\n");
        token_value(fout, 0, nitem, items, values, id, ntoken, 1);
    }
    
//CATEGORY (Citation Authors)

    ntoken = 1;
    strcpy(id[0],"primary_citation_author_name");
    n=check_template( 0, nitem, items, values,id[0], "");
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_citation_author.citation_id\n");
        fprintf(fout,"_citation_author.name\n");
    
        n1=0;
        n2=0;
        for(i=0; i<nitem; i++){
            len = strlen(values[i]);
            if(len<=0) continue;
            if(!strcmp_case(items[i], "primary_citation_author_name")){
                n1++;
                fprintf(fout,"primary   \n;%s\n;\n",  values[i]);
            }
    
            if(!strcmp_case(items[i], "citation_author_id")){
                strcpy(author_id,values[i]);
            }
        
            if(!strcmp_case(items[i], "citation_author_name") && strlen(author_id)>0){
                n2++;
                fprintf(fout,"'%s'   \n;%s\n;\n",author_id, values[i]);
            }
        }
    }
    
    

//CATEGORY 7, (Citation Article)
    ntoken = 7;
    strcpy(id[0],"primary_citation_id");
    strcpy(id[1],"primary_citation_journal_abbrev");
    strcpy(id[2],"primary_citation_title");
    strcpy(id[3],"primary_citation_year");
    strcpy(id[4],"primary_citation_journal_volume");
    strcpy(id[5],"primary_citation_page_first");
    strcpy(id[6],"primary_citation_page_last");

    n=check_template( 0, nitem, items, values,id[0],id[1]);
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_citation.id \n");
        fprintf(fout,"_citation.journal_abbrev \n");
        fprintf(fout,"_citation.title \n");
        fprintf(fout,"_citation.year \n");
        fprintf(fout,"_citation.journal_volume \n");
        fprintf(fout,"_citation.page_first \n");
        fprintf(fout,"_citation.page_last\n");
    
        token_value(fout, 0, nitem, items, values, id, ntoken, 0);

        
        ntoken = 7;
        strcpy(id[0],"citation_id");
        strcpy(id[1],"citation_journal_abbrev");
        strcpy(id[2],"citation_title");
        strcpy(id[3],"citation_year");
        strcpy(id[4],"citation_journal_volume");
        strcpy(id[5],"citation_page_first");
        strcpy(id[6],"citation_page_last");
        n1=check_template( 0, nitem, items, values,id[0],id[1]);
        if(n1){
            token_value(fout, 0, nitem, items, values, id, ntoken, 0);
        }
    }
    

    
//CATEGORY 9:   Molecule Details
    
    ntoken = 4;
    strcpy(id[0],"molecular_entity_id");
    strcpy(id[1],"Fragment_name");
    strcpy(id[2],"Specific_mutation");
    strcpy(id[3],"Enzyme_Comission_number");

    n=check_template( 0, nitem, items, values,id[0],id[1]);
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_entity_keywords.entity_id\n");
        fprintf(fout,"_entity_keywords.ndb_fragment\n");
        fprintf(fout,"_entity_keywords.ndb_mutation\n");
        fprintf(fout,"_entity_keywords.ndb_ec\n");
        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    
//CATEGORY 10 (Genetically Manipulated Source)    

    ntoken = 9;
    strcpy(id[0],"Manipulated_entity_id");
    strcpy(id[1],"Source_organism_scientific_name");
    strcpy(id[2],"Source_organism_gene");
    strcpy(id[3],"Source_organism_strain");
    strcpy(id[4],"Expression_system_scientific_name");
    strcpy(id[5],"Expression_system_strain");
    strcpy(id[6],"Expression_system_vector_type");
    strcpy(id[7],"Expression_system_plasmid_name");
    strcpy(id[8],"Manipulated_source_details");
    
    n=check_template( 0, nitem, items, values,id[0],"");
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_entity_src_gen.entity_id\n");
        fprintf(fout,"_entity_src_gen.pdbx_gene_src_scientific_name\n");
        fprintf(fout,"_entity_src_gen.pdbx_gene_src_gene\n");
        fprintf(fout,"_entity_src_gen.gene_src_strain\n");
        fprintf(fout,"_entity_src_gen.pdbx_host_org_scientific_name\n");
        fprintf(fout,"_entity_src_gen.pdbx_host_org_strain\n");
        fprintf(fout,"_entity_src_gen.pdbx_host_org_vector_type\n");
        fprintf(fout,"_entity_src_gen.plasmid_name\n");
        fprintf(fout,"_entity_src_gen.gene_src_details\n");
        
        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }

 //CATEGORY 11 (Category Natural Source)

    ntoken = 4;
    strcpy(id[0],"natural_source_entity_id");
    strcpy(id[1],"natural_source_scientific_name");
    strcpy(id[2],"natural_source_organism_strain");
    strcpy(id[3],"natural_source_details");

    n=check_template( 0, nitem, items, values,id[0],"");
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_entity_src_nat.entity_id \n");
        fprintf(fout,"_entity_src_nat.pdbx_organism_scientific\n");
        fprintf(fout,"_entity_src_nat.strain\n");
        fprintf(fout,"_entity_src_nat.details\n");
        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    
 //CATEGORY  (Category Synthetic Source)

    
    ntoken = 2;
    strcpy(id[0],"synthetic_source_entity_id");
    strcpy(id[1],"synthetic_source_description");
    
    n=check_template( 0, nitem, items, values,id[0],id[1]);
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_rcsb_entity_src_syn.entity_id \n");
        fprintf(fout,"_rcsb_entity_src_syn.details\n");
        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    
    
}

void parse_infor_xray(FILE *fout, int nitem, char **items, char **values)
// extracting the information related only to Xray  
{
    char  sg[20], id[80][80];;
    int  n, ntoken=0, i,  ncryst=0, len=0;

//CATEGORY 1  (Cell Parameters)
    for(i=0; i<nitem; i++){
        len=strlen(values[i]);
        if(len<=0)continue;
        
        if(!strcmp_case(items[i], "space_group") ){
            space_group_convertor(values[i], sg);
//            strcpy(symmetry.space_group_name_H_M,  sg);
            fprintf(fout,"_symmetry.entry_id      UNNAMED ");
            fprintf(fout,"_symmetry.space_group_name_H-M  '%s'\n",sg);
                    
        }else if(!strcmp_case(items[i],"space_group_number") ){
//            symmetry.Int_Tables_number = atoi(values[i]);
            fprintf(fout,"_symmetry.Int_Tables_number  '%s'\n",values[i]);
            
        }else if(!strcmp_case(items[i], "unit_cell_a") ){
            cell.length_a = atof(values[i]);
//            fprintf(fout,"_cell.length_a   \"%s\"\n",values[i]);
            
        }else if(!strcmp_case(items[i], "unit_cell_b")){
            cell.length_b = atof(values[i]);
//            fprintf(fout,"_cell.length_b   \"%s\"\n",values[i]);
            
        }else if(!strcmp_case(items[i], "unit_cell_c")){
            cell.length_c = atof(values[i]);
//            fprintf(fout,"_cell.length_c   \"%s\"\n",values[i]);
            
        }else if(!strcmp_case(items[i], "unit_cell_alpha")){
            cell.angle_alpha = atof(values[i]);
//            fprintf(fout,"_cell.angle_alpha      \"%s\"\n",values[i]);
            
        }else if(!strcmp_case(items[i], "unit_cell_beta")){
            cell.angle_beta = atof(values[i]);
//            fprintf(fout,"_cell.angle_beta       \"%s\"\n",values[i]);
            
        }else if(!strcmp_case(items[i], "unit_cell_gamma")){
            cell.angle_gamma = atof(values[i]);
//            fprintf(fout,"_cell.angle_gamma      \"%s\"\n",values[i]);

        }else if(!strcmp_case(items[i], "crystal_number")&& len>0 ){
            ncryst=atoi(values[i]);
        }
        
    }
    if(ncryst>0){
        fprintf(fout,"#\n");
        fprintf(fout,"\n_exptl.crystals_number %d\n",ncryst);
        fprintf(fout,"_exptl.entry_id   UNNAMED\n");
        fprintf(fout,"_exptl.method   'X-RAY DIFFRACTION'\n");
    }    
    
//CATEGORY   (Biological Assembly)
            
    fprintf(fout,"#\nloop_\n");
    fprintf(fout,"_struct_biol.id\n");
    fprintf(fout,"_struct_biol.details\n");

    ntoken = 1;
    strcpy(id[0],"biological_assembly");
    n=check_template( 0, nitem, items, values,id[0],"");
    if(n){
        token_value(fout, 0, nitem, items, values, id, ntoken, 1);
    }else
        fprintf(fout,"1 'biological unit is the same as asym.'\n");

//CATEGORY  (Crystal Property)
    

    ntoken = 4;
    strcpy(id[0],"crystals_number");
    strcpy(id[1],"crystals_solvent_content");
    strcpy(id[2],"crystals_matthews_coefficient");
    strcpy(id[3],"crystals_mosaicity");

    n=check_template( 0, nitem, items, values,id[0],"");
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_exptl_crystal.id \n");
        fprintf(fout,"_exptl_crystal.density_percent_sol\n");
        fprintf(fout,"_exptl_crystal.density_Matthews\n");
        fprintf(fout,"_exptl_crystal.pdbx_mosaicity\n");

        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    

    
//CATEGORY 15 (Methods and Conditions)
    ntoken = 5;
    strcpy(id[0],"crystal_number");
    strcpy(id[1],"crystallization_method");
    strcpy(id[2],"crystallization_pH");
    strcpy(id[3],"crystallization_temperature");
    strcpy(id[4],"crystallization_details");
    
    n=check_template( 0, nitem, items, values,id[0],"");
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_exptl_crystal_grow.crystal_id \n");
        fprintf(fout,"_exptl_crystal_grow.method\n");
        fprintf(fout,"_exptl_crystal_grow.pH \n");
        fprintf(fout,"_exptl_crystal_grow.temp\n");
        fprintf(fout,"_exptl_crystal_grow.pdbx_details\n");

        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    
//CATEGORY  (Radiation Source)
    
    ntoken = 4;
    strcpy(id[0],"radiation_experiment");
    strcpy(id[1],"radiation_source");
    strcpy(id[2],"radiation_source_type");
    strcpy(id[3],"radiation_wavelengths");

    n=check_template( 0, nitem, items, values,id[0],"");
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_diffrn_source.diffrn_id\n");
        fprintf(fout,"_diffrn_source.source\n");
        fprintf(fout,"_diffrn_source.type\n");
        fprintf(fout,"_diffrn_source.rcsb_wavelength_list\n");

        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    
//---    

    
    ntoken = 5;
    strcpy(id[0],"radiation_experiment");
    strcpy(id[1],"radiation_detector");
    strcpy(id[2],"radiation_detector_type");
    strcpy(id[3],"data_collection_date");
    strcpy(id[4],"radiation_detector_details");
    n=check_template( 0, nitem, items, values,id[0],"");
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_diffrn_detector.diffrn_id \n");
        fprintf(fout,"_diffrn_detector.detector \n");
        fprintf(fout,"_diffrn_detector.type \n");
        fprintf(fout,"_diffrn_detector.ndb_collection_date \n");
        fprintf(fout,"_diffrn_detector.details \n");
        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    
//---    
    
    ntoken = 3;
    strcpy(id[0],"radiation_experiment");
    strcpy(id[1],"data_collection_protocol");
    strcpy(id[2],"data_collection_monochromator");

    n=check_template( 0, nitem, items, values,id[0],"");
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_diffrn_radiation.diffrn_id \n");
        fprintf(fout,"_diffrn_radiation.rcsb_diffrn_protocol \n");
        fprintf(fout,"_diffrn_radiation.monochromator \n");

        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    
//---    

    
    ntoken= 2;
    strcpy(id[0],"radiation_experiment");
    strcpy(id[1],"radiation_wavelengths");
    n=check_template( 0, nitem, items, values,id[0],id[1]);
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_diffrn_radiation_wavelength.id \n");
        fprintf(fout,"_diffrn_radiation_wavelength.wavelength \n");
        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    

    
//CATEGORY   (Collection Temperature)
    
    ntoken= 2;
    strcpy(id[0],"radiation_experiment");
    strcpy(id[1],"data_collection_temperature");
    n=check_template( 0, nitem, items, values,id[0],id[1]);
    if(n){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_diffrn.id\n");
        fprintf(fout,"_diffrn.ambient_temp\n");
        token_value(fout, 0, nitem, items, values, id, ntoken, 0);
    }
    
}

int check_template(int k,int nitem,char **items, char **values,
                  const char *id1, const char *id2)
{
    int i, key1=0, key2=0, check=0;


    if(strlen(id1)>0 && strlen(id2)>0){
        for(i=k; i<nitem; i++){
            if(!strcmp_case(items[i], id1) && strlen(values[i])>0){
                key1=1;
            }else if(!strcmp_case(items[i], id2) && strlen(values[i])>0){
                key2=1;
            }
            if(key1==1 && key2==1) {
                check=1;
                break;
            }
        }
    }else if(strlen(id1)>0 && strlen(id2)<=0){
        for(i=k; i<nitem; i++){
            if(!strcmp_case(items[i], id1) && strlen(values[i])>0){
                check=1;
                break;
            }
        }
    }else if(strlen(id2)>0 && strlen(id1)<=0){
        for(i=k; i<nitem; i++){
            if(!strcmp_case(items[i], id2) && strlen(values[i])>0){
                check=1;
                break;
            }
        }
    }
    
    return check;
}





/////////////////////////////////////////////////////////////////////

void parse_infor_nmr(FILE *fout, char *inpfile)
// extracting the information related only to NMR    
{
    char *value, item[80];
    char id1[60],id2[60],id3[60],id4[60],id5[60],id6[60],id7[60],id8[60],id9[60];
    int key,  nitem=0;
    FILE *finp=NULL;
    
    if((finp=fopen(inpfile, "r")) ==NULL)
        open_file_error(inpfile, "extract_infor_for_nmr");

    value = new char[10000];
    
//CATEGORY 12:   Ensemble

    rewind(finp);
    while(1) { 
        values_in_bracket(finp,  item, value,  &key);
        if(key>0)break;
        if( strlen(value)<=0 ) continue;
        if(!strcmp_case(item, "conformers_calculated_total_number")){
            fprintf(fout,"\n _pdbx_nmr_ensemble.conformers_calculated_total_number\n;%s\n;\n\n",value);
            
        }else if(!strcmp_case(item, "conformers_submitted_total_number")){
            fprintf(fout,"\n _pdbx_nmr_ensemble.conformers_submitted_total_number\n;%s\n;\n\n",value);
            fprintf(fout,"\n _pdbx_nmr_ensemble.entry_id  UNNAMED\n");
            
        }else if(!strcmp_case(item, "conformers_selection_criteria")){
            fprintf(fout,"\n_pdbx_nmr_ensemble.conformer_selection_criteria \n;%s\n;\n\n",value);
        }
        
    }

//CATEGORY 13:   Representative Conformers
    rewind(finp);
    while(1) { 
        values_in_bracket(finp,  item, value,  &key);
        if(key>0)break;
        if( strlen(value)<=0 ) continue;
        if(!strcmp_case(item, "conformer_id")){
            fprintf(fout,"\n_pdbx_nmr_representative.conformer_id \n;%s\n;\n\n",value);
            
        }else if(!strcmp_case(item, "conformer_selection_criteria")){
            fprintf(fout,"\n_pdbx_nmr_representative.selection_criteria \n;%s\n;\n\n",value);
        }
    }

//CATEGORY 14:   Sample Details
    fprintf(fout,"#\nloop_\n");
    fprintf(fout,"_pdbx_nmr_sample_details.solution_id\n");
    fprintf(fout,"_pdbx_nmr_sample_details.contents\n");
    fprintf(fout,"_pdbx_nmr_sample_details.solvent_system\n");
    
    nitem = 3;
    initialize_id(id1,id2,id3,id4,id5,id6,id7,id8,id9);
    strcpy(id1,"solution_id");
    strcpy(id2,"solution_content");
    strcpy(id3,"solvent_system");

    write_values(fout, finp,  nitem, id1,id2,id3,id4,id5,id6,id7 ,id8,id9);

//CATEGORY 15:   Sample Conditions

     fprintf(fout,"#\nloop_\n");
     fprintf(fout,"_pdbx_nmr_exptl_sample_conditions.conditions_id\n");
     fprintf(fout,"_pdbx_nmr_exptl_sample_conditions.temperature\n");
     fprintf(fout,"_pdbx_nmr_exptl_sample_conditions.pressure\n");
     fprintf(fout,"_pdbx_nmr_exptl_sample_conditions.pH\n");
     fprintf(fout,"_pdbx_nmr_exptl_sample_conditions.ionic_strength\n");
   
    
    nitem = 5;
    initialize_id(id1,id2,id3,id4,id5,id6,id7,id8,id9);
    strcpy(id1,"Conditions_id");
    strcpy(id2,"Temperature");
    strcpy(id3,"Pressure");
    strcpy(id4,"pH_value");
    strcpy(id5,"Ionic_strength");

    write_values(fout, finp,  nitem, id1,id2,id3,id4,id5,id6,id7,id8,id9 );

// CATEGORY 16:   Spectrometer
    
    fprintf(fout,"#\nloop_\n");
    fprintf(fout,"_pdbx_nmr_spectrometer.spectrometer_id\n");
    fprintf(fout,"_pdbx_nmr_spectrometer.manufacturer\n");
    fprintf(fout,"_pdbx_nmr_spectrometer.model\n");
    fprintf(fout,"_pdbx_nmr_spectrometer.field_strength\n");

    nitem = 4;
    initialize_id(id1,id2,id3,id4,id5,id6,id7,id8,id9);
    strcpy(id1,"spectrometer_id");
    strcpy(id2,"spectrometer_manufacturer");
    strcpy(id3,"spectrometer_model");
    strcpy(id4,"spectrometer_field_strength");
    write_values(fout, finp,  nitem, id1,id2,id3,id4,id5,id6,id7,id8,id9 );

// CATEGORY 17:    Experiment Type
    
    fprintf(fout,"#\nloop_\n");
    fprintf(fout,"_pdbx_nmr_exptl.experiment_id\n");
    fprintf(fout,"_pdbx_nmr_exptl.solution_id\n");
    fprintf(fout,"_pdbx_nmr_exptl.conditions_id\n");
    fprintf(fout,"_pdbx_nmr_exptl.type\n");

    nitem = 4;
    initialize_id(id1,id2,id3,id4,id5,id6,id7,id8,id9);
    strcpy(id1,"experiment_type_id");
    strcpy(id2,"solution_type_id");
    strcpy(id3,"conditions_type_id");
    strcpy(id4,"Experiment_type");
    write_values(fout, finp,  nitem, id1,id2,id3,id4,id5,id6,id7,id8,id9 );

//CATEGORY 18:   Method and Details

    rewind(finp);
    while(1) { 
        values_in_bracket(finp,  item, value,  &key);
        if(key>0)break;
        if( strlen(value)<=0 ) continue;
        if(!strcmp_case(item, "NMR_method")){
            fprintf(fout,"\n_pdbx_nmr_refine.entry_id 	UNNAMED \n");
            fprintf(fout,"\n_pdbx_nmr_refine.method \n;%s\n;\n\n",value);
            
        }else if(!strcmp_case(item, "NMR_details")){
            fprintf(fout,"\n_pdbx_nmr_refine.details \n;%s\n;\n\n",value);
        }
    }
    fclose(finp);
    if(value) delete [] value;
}

void initialize_id(char *id1,char *id2,char *id3,char *id4,char *id5,
                   char *id6, char *id7,char *id8, char *id9)
{
    strcpy(id1," ?? ");
    strcpy(id2," ?? ");
    strcpy(id3," ?? ");
    strcpy(id4," ?? ");
    strcpy(id5," ?? ");
    strcpy(id6," ?? ");
    strcpy(id7," ?? ");
    strcpy(id8," ?? ");
    strcpy(id9," ?? ");
}

void test_sequence_correction(char *value)
{
    int i=0, n;
    
    n = strlen(value);
    for(i=0; i<n-1; i++){
        if(value[i] == '?'){
            printf("\nSequence error occurred!\n");
            printf("There are \'????\' in your chemical sequences in the file(data_template.text).\n");
            printf("Please do the following corrections, before proceeding.\n");
            printf("If \'????\' are caused by missing residues, substitute them by the missing residue names\n");
            printf("Otherwise, please delete the \'????\'.\n");
            
            exit(0);
        }
    }
}

void write_values(FILE *fout, FILE *finp, int nitem, char *id1,char *id2,
                  char *id3,char *id4,char *id5, char *id6, char *id7,
                  char *id8, char *id9)
/*A general route for writting values */
{
    char item[80], *value=NULL;
    char *str1=NULL, *str2=NULL, *str3=NULL, *str4=NULL;
    char *str5=NULL, *str6=NULL, *str7=NULL, *str8=NULL, *str9=NULL;
    int  n1=0, n2=0, n3=0, n4=0, n5=0, n6=0, n7=0, n8=0, n9=0;
    int  nn1=0, key=0, len=0;
   
    rewind(finp);
    
    while(1) {
        value= new char[10000]; //max values
        values_in_bracket(finp,  item, value,  &key);

        rid_of_front_end_space(value);
        
        len=strlen(value)+2;
        
        if(key>0)break;

            
        if(strstr_case(item, id1)){
            
            str1=new char[len];
            strcpy(str1, value);
            n1++;
        }else if(strstr_case(item, id2)){
            str2=new char[len+2];
            strcpy(str2, value);
            n2++;
        }else if(strstr_case(item, id3)){
            str3=new char[len+2];
            strcpy(str3, value);
            n3++;
        }else if(strstr_case(item, id4)){
            
            str4=new char[len+2];
            strcpy(str4, value);
            n4++;
            
        }else if(strstr_case(item, id5)){
            str5=new char[len+2];
            strcpy(str5, value);
            n5++; 
        }else if(strstr_case(item, id6)){
            str6=new char[len+2];
            strcpy(str6, value);
            n6++; 
        }else if(strstr_case(item, id7)){
            str7=new char[len+2];
            strcpy(str7, value);
            n7++; 
        }else if(strstr_case(item, id8)){
            str8=new char[len+2];
            strcpy(str8, value);
            n8++; 
        }else if(strstr_case(item, id9)){
            str9=new char[len+2];
            strcpy(str9, value);
            n9++; 
        }
        delete [] value;
        
        if(nitem==1 &&  n1>0){
            if(strlen(str1)>0){
                nn1++;
                fprintf(fout,"\n;%s\n; ", str1);
                n1=0; 
                str1=NULL;
            }else{
                n1=0; 
                str1=NULL;
            }
            fprintf(fout,"\n");
            
        }else if(nitem==2 &&  n1>0 && n2>0){
            if(strlen(str1)>0){
                nn1++;
                fprintf(fout,"\n;%s\n; ", str1);
                (strlen(str2)>0) ? fprintf(fout,"\n;%s\n; ",str2):fprintf(fout,"\n;?\n;");
                n1=0; n2=0;
                str1=NULL; str2=NULL;
            }else{
                n1=0; n2=0;
                str1=NULL; str2=NULL;
            }
            fprintf(fout,"\n");
            
        }else if (nitem==3 &&  n1>0 && n2>0 && n3>0){
            if(strlen(str1)>0){
                nn1++;
                fprintf(fout,"\n;%s\n; ", str1);
                (strlen(str2)>0) ? fprintf(fout,"\n;%s\n; ",str2):fprintf(fout,"\n;?\n;");
                (strlen(str3)>0) ? fprintf(fout,"\n;%s\n; ",str3):fprintf(fout,"\n;?\n;");
                n1=0; n2=0; n3=0;
                str1=NULL; str2=NULL; str3=NULL;
            }else{
                n1=0; n2=0; n3=0;
                str1=NULL; str2=NULL; str3=NULL;
            }
            fprintf(fout,"\n");
            
        }else if (nitem==4 && n1>0 && n2>0 && n3>0 && n4>0){
            
            if(strstr_case(id1,"entity_id") && strstr_case(id2,"entity_type")&&
               strstr_case(id3,"one_letter") && strstr_case(id4, "chain_id"))
                test_sequence_correction(str3);
         
            
                                     
            if(strlen(str1)>0){
                
                nn1++;
                
                fprintf(fout,"\n;%s\n; ",str1);
                (strlen(str2)>0) ? fprintf(fout,"\n;%s\n; ",str2):fprintf(fout,"\n;?\n;");
                (strlen(str3)>0) ? fprintf(fout,"\n;%s\n; ",str3):fprintf(fout,"\n;?\n;");
                (strlen(str4)>0) ? fprintf(fout,"\n;%s\n; ",str4):fprintf(fout,"\n;?\n;");
                
                n1=0; n2=0; n3=0; n4=0;
                delete [] str1;delete [] str2;delete [] str3;delete [] str4;
                 
//                str1=NULL; str2=NULL; str3=NULL; str4=NULL;
            }else{
//                str1=NULL; str2=NULL; str3=NULL; str4=NULL;
                delete [] str1;delete [] str2;delete [] str3;delete [] str4;
                n1=0; n2=0; n3=0; n4=0;  
            }
            fprintf(fout,"\n");
            
            
            
        }else if (nitem==5 &&  n1>0 && n2>0 && n3>0 && n4>0 && n5>0 ){
            if(strlen(str1)>0){
                nn1++;
                fprintf(fout,"\n;%s\n; ", str1);
                (strlen(str2)>0) ? fprintf(fout,"\n;%s\n; ",str2):fprintf(fout,"\n;?\n;");
                (strlen(str3)>0) ? fprintf(fout,"\n;%s\n; ",str3):fprintf(fout,"\n;?\n;");
                (strlen(str4)>0) ? fprintf(fout,"\n;%s\n; ",str4):fprintf(fout,"\n;?\n;");
                (strlen(str5)>0) ? fprintf(fout,"\n;%s\n; ",str5):fprintf(fout,"\n;?\n;");
                n1=0; n2=0; n3=0; n4=0;n5=0;
                str1=NULL; str2=NULL; str3=NULL; str4=NULL; str5=NULL;
            }else{
                n1=0; n2=0; n3=0; n4=0;n5=0;
                str1=NULL; str2=NULL; str3=NULL; str4=NULL; str5=NULL;
            }
            fprintf(fout,"\n");
            
        }else if (nitem==6 && n1>0 && n2>0 && n3>0 && n4>0 && n5>0 && n6>0){
            if(strlen(str1)>0){
                nn1++;
                fprintf(fout,"\n;%s\n; ", str1);
                (strlen(str2)>0) ? fprintf(fout,"\n;%s\n; ",str2):fprintf(fout,"\n;?\n;");
                (strlen(str3)>0) ? fprintf(fout,"\n;%s\n; ",str3):fprintf(fout,"\n;?\n;");
                (strlen(str4)>0) ? fprintf(fout,"\n;%s\n; ",str4):fprintf(fout,"\n;?\n;");
                (strlen(str5)>0) ? fprintf(fout,"\n;%s\n; ",str5):fprintf(fout,"\n;?\n;");
                (strlen(str6)>0) ? fprintf(fout,"\n;%s\n; ",str6):fprintf(fout,"\n;?\n;");
                n1=0; n2=0; n3=0; n4=0;n5=0;n6=0;
                str1=NULL; str2=NULL; str3=NULL; str4=NULL; str5=NULL;str6=NULL;
            }else{
                n1=0; n2=0; n3=0; n4=0;n5=0;n6=0;
                str1=NULL; str2=NULL; str3=NULL; str4=NULL; str5=NULL;str6=NULL;
            }
            fprintf(fout,"\n");
            
        }else if (nitem==7 && n1>0 && n2>0 && n3>0 && n4>0 && n5>0 && n6>0 && n7>0){
            if(strlen(str1)>0){
                nn1++;
                fprintf(fout,"\n;%s\n; ", str1);
                (strlen(str2)>0) ? fprintf(fout,"\n;%s\n; ",str2):fprintf(fout,"\n;?\n;");
                (strlen(str3)>0) ? fprintf(fout,"\n;%s\n; ",str3):fprintf(fout,"\n;?\n;");
                (strlen(str4)>0) ? fprintf(fout,"\n;%s\n; ",str4):fprintf(fout,"\n;?\n;");
                (strlen(str5)>0) ? fprintf(fout,"\n;%s\n; ",str5):fprintf(fout,"\n;?\n;");
                (strlen(str6)>0) ? fprintf(fout,"\n;%s\n; ",str6):fprintf(fout,"\n;?\n;");
                (strlen(str7)>0) ? fprintf(fout,"\n;%s\n; ",str7):fprintf(fout,"\n;?\n;");
                n1=0; n2=0; n3=0; n4=0;n5=0;n6=0;n7=0;
                str1=NULL; str2=NULL; str3=NULL; str4=NULL; str5=NULL;str6=NULL;str7=NULL;
            }else{
                n1=0; n2=0; n3=0; n4=0;n5=0;n6=0;n7=0;
                str1=NULL; str2=NULL; str3=NULL; str4=NULL; str5=NULL;str6=NULL;str7=NULL;
            }
            fprintf(fout,"\n");
            
        }else if (nitem==8 && n1>0 && n2>0 && n3>0 && n4>0 && n5>0 && n6>0 && n7>0 && n8>0){
            if(strlen(str1)>0){
                nn1++;
                fprintf(fout,"\n;%s\n; ", str1);
                (strlen(str2)>0) ? fprintf(fout,"\n;%s\n; ",str2):fprintf(fout,"\n;?\n;");
                (strlen(str3)>0) ? fprintf(fout,"\n;%s\n; ",str3):fprintf(fout,"\n;?\n;");
                (strlen(str4)>0) ? fprintf(fout,"\n;%s\n; ",str4):fprintf(fout,"\n;?\n;");
                (strlen(str5)>0) ? fprintf(fout,"\n;%s\n; ",str5):fprintf(fout,"\n;?\n;");
                (strlen(str6)>0) ? fprintf(fout,"\n;%s\n; ",str6):fprintf(fout,"\n;?\n;");
                (strlen(str7)>0) ? fprintf(fout,"\n;%s\n; ",str7):fprintf(fout,"\n;?\n;");
                (strlen(str8)>0) ? fprintf(fout,"\n;%s\n; ",str8):fprintf(fout,"\n;?\n;");
                n1=0; n2=0; n3=0; n4=0;n5=0;n6=0;n7=0;n8=0;
                str1=NULL; str2=NULL; str3=NULL; str4=NULL; str5=NULL;str6=NULL;str7=NULL;str8=NULL;
            }else{
                n1=0; n2=0; n3=0; n4=0;n5=0;n6=0;n7=0;n8=0;
                str1=NULL; str2=NULL; str3=NULL; str4=NULL; str5=NULL;str6=NULL;str7=NULL;str8=NULL;
            }
            fprintf(fout,"\n");
            
        }else if (nitem==9 && n1>0 && n2>0 && n3>0 && n4>0 && n5>0 && n6>0 && n7>0 && n8>0 && n9>0){
            if(strlen(str1)>0){
                nn1++;
                fprintf(fout,"\n;%s\n; ", str1);
                (strlen(str2)>0) ? fprintf(fout,"\n;%s\n; ",str2):fprintf(fout,"\n;?\n;");
                (strlen(str3)>0) ? fprintf(fout,"\n;%s\n; ",str3):fprintf(fout,"\n;?\n;");
                (strlen(str4)>0) ? fprintf(fout,"\n;%s\n; ",str4):fprintf(fout,"\n;?\n;");
                (strlen(str5)>0) ? fprintf(fout,"\n;%s\n; ",str5):fprintf(fout,"\n;?\n;");
                (strlen(str6)>0) ? fprintf(fout,"\n;%s\n; ",str6):fprintf(fout,"\n;?\n;");
                (strlen(str7)>0) ? fprintf(fout,"\n;%s\n; ",str7):fprintf(fout,"\n;?\n;");
                (strlen(str8)>0) ? fprintf(fout,"\n;%s\n; ",str8):fprintf(fout,"\n;?\n;");
                (strlen(str9)>0) ? fprintf(fout,"\n;%s\n; ",str9):fprintf(fout,"\n;?\n;");
                n1=0; n2=0; n3=0; n4=0;n5=0;n6=0;n7=0;n8=0;n9=0;
                str1=NULL; str2=NULL; str3=NULL; str4=NULL; str5=NULL;str6=NULL;str7=NULL;str8=NULL;str9=NULL;
            }else{
                n1=0; n2=0; n3=0; n4=0;n5=0;n6=0;n7=0;n8=0;n9=0;
                str1=NULL; str2=NULL; str3=NULL; str4=NULL; str5=NULL;str6=NULL;str7=NULL;str8=NULL;str9=NULL;
            }
            fprintf(fout,"\n"); 
        }
        
    }
    
    if(nitem== 1 && nn1==0 ){
        fprintf(fout,"? \n"); 
    }else if(nitem==2  && nn1==0 ){
        fprintf(fout,"? ? \n"); 
    }else if(nitem==3  && nn1==0 ){
        fprintf(fout,"? ? ? \n"); 
    }else if(nitem==4  && nn1==0 ){
        fprintf(fout,"? ? ? ? \n"); 
    }else if(nitem==5  && nn1==0 ){
        fprintf(fout,"? ? ? ? ?\n"); 
    }else if(nitem==6  && nn1==0 ){
        fprintf(fout,"? ? ? ? ? ? \n"); 
    }else if(nitem==7  && nn1==0 ){
        fprintf(fout,"? ? ? ? ? ? ?\n"); 
    }else if(nitem==8  && nn1==0 ){
        fprintf(fout,"? ? ? ? ? ? ? ?\n"); 
    }else if(nitem==9  && nn1==0 ){
        fprintf(fout,"? ? ? ? ? ? ? ? ?\n"); 
    }    
}

void values_in_bracket(FILE *inp, char *first_item, char *value, int *key)
/* first_item: the identifier.  
   value: the total characters in " " <>.
   key: =0 not the end, =1 end of file
*/
{
    int i=0,j=0,n=1, m=0,k=0;
    int c;
    char *tmp=NULL, lett[10000];
    
    *key=0;
/* skip every letter until < is met */
    for(j=0; (c=getc(inp)) != '<' ; j++){
        if(c==EOF)  {
            *key=1;
            break;
        }
    }
/*take off the space between < and the first letter */    
    for(j=0; (c=getc(inp)) == ' ' ; j++){
        if(c==EOF)  {
            *key=1;
            break;
        }
    }
    lett[0]=c; 
    for(j=0; (c=getc(inp)) != '>' ; j++){
        if(c==EOF)  {
            *key=1;
            break;
        }
//        if(c!='\n'){   /* omit \n, put all the chars in one line.*/                
            lett[n]=c;
            n=n+1;
//        }
    }
    lett[n] = '\0';
    if(pattern5(lett, "=", "\"", "\"", "", "")<=0){
        strcpy(first_item, "");
        strcpy(value, "");
        return;
    }
    
    tmp = new char[n+1];

    i=strcspn(lett,"="); /* get the first item */
    m=0;
    for (k=0;k<i; ++k){
        if(isspace(lett[k])) continue;
        
        first_item[m]=lett[k];
        m++;
    }
    first_item[m]='\0';
        

    strcpy(tmp, strchr(lett,'=')+1);  /*get string after the first word */
    j=0;
    n=strlen(tmp);
    for (i=0;i<n; ++i){
        if((tmp[i]!='\"')){
            value[j]=tmp[i];
            j++;
        }
    }    
    value[j]='\0';
    
    rid_of_front_end_space(value);
    delete [] tmp;
    
}



void write_text_parser_cif_head(FILE *fout, char *id)
// 
{
    
    
//CATEGORY 2:   Molecular Entity Sequence
    if(!strcmp_case(id, "entity_poly")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_entity_poly.entity_id\n");
        fprintf(fout,"_entity_poly.type\n");
        fprintf(fout,"_entity_poly.pdbx_seq_one_letter_code\n");
        fprintf(fout,"_entity_poly.pdbx_strand_id\n");
        fprintf(fout,"_entity_poly.pdbx_target_identifier\n");
    
//CATEGORY 3, (Contact Authors for PI & others)
    }else if(!strcmp_case(id, "pdbx_contact_author")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_pdbx_contact_author.id\n");
        fprintf(fout,"_pdbx_contact_author.name_salutation\n");
        fprintf(fout,"_pdbx_contact_author.name_first\n");
        fprintf(fout,"_pdbx_contact_author.name_last\n");
        fprintf(fout,"_pdbx_contact_author.name_mi\n");
        fprintf(fout,"_pdbx_contact_author.role\n");
        fprintf(fout,"_pdbx_contact_author.organization_type\n");
        fprintf(fout,"_pdbx_contact_author.email\n");
        fprintf(fout,"_pdbx_contact_author.address_1\n");
        fprintf(fout,"_pdbx_contact_author.city\n");
        fprintf(fout,"_pdbx_contact_author.state_province\n");
        fprintf(fout,"_pdbx_contact_author.postal_code\n");
        fprintf(fout,"_pdbx_contact_author.country\n");
        fprintf(fout,"_pdbx_contact_author.fax\n");
        fprintf(fout,"_pdbx_contact_author.phone\n");
        
//CATEGORY  (Structure Genomics)
    }else if(!strcmp_case(id, "pdbx_SG_project")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout," _pdbx_SG_project.id\n");
        fprintf(fout," _pdbx_SG_project.project_name\n");
        fprintf(fout," _pdbx_SG_project.full_name_of_center\n");
        
//CATEGORY  (structure Author)        
    }else if(!strcmp_case(id, "audit_author")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_audit_author.id\n");
        fprintf(fout,"_audit_author.name\n");
        
//CATEGORY 8:   Molecule Names
    }else if(!strcmp_case(id, "entity")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_entity.id\n");
        fprintf(fout,"_entity.pdbx_description\n");
        
//CATEGORY (Citation Authors)
    }else if(!strcmp_case(id, "citation_author")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_citation_author.citation_id\n");
        fprintf(fout,"_citation_author.name\n");
        
//CATEGORY 9, (Citation Article)
    }else if(!strcmp_case(id, "citation")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_citation.id \n");
        fprintf(fout,"_citation.journal_abbrev \n");
        fprintf(fout,"_citation.title \n");
        fprintf(fout,"_citation.year \n");
        fprintf(fout,"_citation.journal_volume \n");
        fprintf(fout,"_citation.page_first \n");
        fprintf(fout,"_citation.page_last\n");
        
//CATEGORY 9:   Molecule Details
    }else if(!strcmp_case(id, "entity_keywords")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_entity_keywords.entity_id\n");
        fprintf(fout,"_entity_keywords.ndb_fragment\n");
        fprintf(fout,"_entity_keywords.ndb_mutation\n");
        fprintf(fout,"_entity_keywords.ndb_ec\n");

        
//CATEGORY 10 (Genetically Manipulated Source)    
    }else if(!strcmp_case(id, "entity_src_gen")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_entity_src_gen.entity_id\n");
        fprintf(fout,"_entity_src_gen.pdbx_gene_src_scientific_name\n");
        fprintf(fout,"_entity_src_gen.pdbx_gene_src_gene\n");
        fprintf(fout,"_entity_src_gen.gene_src_strain\n");
        fprintf(fout,"_entity_src_gen.pdbx_host_org_scientific_name\n");
        fprintf(fout,"_entity_src_gen.pdbx_host_org_strain\n");
        fprintf(fout,"_entity_src_gen.pdbx_host_org_vector_type\n");
        fprintf(fout,"_entity_src_gen.plasmid_name\n");
        fprintf(fout,"_entity_src_gen.gene_src_details\n");
        
 //CATEGORY 11 (Category Natural Source)
    }else if(!strcmp_case(id, "entity_src_nat")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_entity_src_nat.entity_id \n");
        fprintf(fout,"_entity_src_nat.pdbx_organism_scientific\n");
        fprintf(fout,"_entity_src_nat.strain\n");
        fprintf(fout,"_entity_src_nat.details\n");
        
 //CATEGORY  (Category Synthetic Source)
    }else if(!strcmp_case(id, "rcsb_entity_src_syn")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_rcsb_entity_src_syn.entity_id \n");
        fprintf(fout,"_rcsb_entity_src_syn.details\n");
        



// X-ray part ================================
 //CATEGORY   (Biological Assembly)
    }else if(!strcmp_case(id, "struct_biol")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_struct_biol.id\n");
        fprintf(fout,"_struct_biol.details\n");
        
//CATEGORY  (Crystal Property)
    }else if(!strcmp_case(id, "exptl_crystal")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_exptl_crystal.id \n");
        fprintf(fout,"_exptl_crystal.density_percent_sol\n");
        fprintf(fout,"_exptl_crystal.density_Matthews\n");
        fprintf(fout,"_exptl_crystal.pdbx_mosaicity\n");
        
//CATEGORY 15 (Methods and Conditions)
    }else if(!strcmp_case(id, "exptl_crystal_grow")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_exptl_crystal_grow.crystal_id \n");
        fprintf(fout,"_exptl_crystal_grow.method\n");
        fprintf(fout,"_exptl_crystal_grow.pH \n");
        fprintf(fout,"_exptl_crystal_grow.temp\n");
        fprintf(fout,"_exptl_crystal_grow.pdbx_details\n");
        
//CATEGORY  (Radiation Source)
    }else if(!strcmp_case(id, "diffrn_source")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_diffrn_source.diffrn_id\n");
        fprintf(fout,"_diffrn_source.source\n");
        fprintf(fout,"_diffrn_source.type\n");
        fprintf(fout,"_diffrn_source.rcsb_wavelength_list\n");
        
//CATEGORY  detectors
    }else if(!strcmp_case(id, "diffrn_detector")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_diffrn_detector.diffrn_id \n");
        fprintf(fout,"_diffrn_detector.detector \n");
        fprintf(fout,"_diffrn_detector.type \n");
        fprintf(fout,"_diffrn_detector.ndb_collection_date \n");
        fprintf(fout,"_diffrn_detector.details \n");
        
        
//CATEGORY   radiation protocol
    }else if(!strcmp_case(id, "diffrn_radiation")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_diffrn_radiation.diffrn_id \n");
        fprintf(fout,"_diffrn_radiation.rcsb_diffrn_protocol \n");
        fprintf(fout,"_diffrn_radiation.monochromator \n");
        
//CATEGORY   wavelength
    }else if(!strcmp_case(id, "diffrn_radiation_wavelength")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_diffrn_radiation_wavelength.id \n");
        fprintf(fout,"_diffrn_radiation_wavelength.wavelength \n");
        
//CATEGORY   (Collection Temperature)
    }else if(!strcmp_case(id, "diffrn")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_diffrn.id\n");
        fprintf(fout,"_diffrn.ambient_temp\n");
        



// NMR================================
//CATEGORY 14:   NMR Sample Details

    }else if(!strcmp_case(id, "pdbx_nmr_sample_details")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_pdbx_nmr_sample_details.solution_id\n");
        fprintf(fout,"_pdbx_nmr_sample_details.contents\n");
        fprintf(fout,"_pdbx_nmr_sample_details.solvent_system\n");
        
//CATEGORY 15:  NMR  Sample Conditions
    }else if(!strcmp_case(id, "pdbx_nmr_exptl_sample_conditions")){
            fprintf(fout,"#\nloop_\n");
         fprintf(fout,"_pdbx_nmr_exptl_sample_conditions.conditions_id\n");
         fprintf(fout,"_pdbx_nmr_exptl_sample_conditions.temperature\n");
         fprintf(fout,"_pdbx_nmr_exptl_sample_conditions.pressure\n");
         fprintf(fout,"_pdbx_nmr_exptl_sample_conditions.pH\n");
         fprintf(fout,"_pdbx_nmr_exptl_sample_conditions.ionic_strength\n");
        
// CATEGORY 16: NMR  Spectrometer
    }else if(!strcmp_case(id, "pdbx_nmr_spectrometer")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_pdbx_nmr_spectrometer.spectrometer_id\n");
        fprintf(fout,"_pdbx_nmr_spectrometer.manufacturer\n");
        fprintf(fout,"_pdbx_nmr_spectrometer.model\n");
        fprintf(fout,"_pdbx_nmr_spectrometer.field_strength\n");
        
// CATEGORY 17:  NMR   Experiment Type
    }else if(!strcmp_case(id, "pdbx_nmr_exptl")){
        fprintf(fout,"#\nloop_\n");
        fprintf(fout,"_pdbx_nmr_exptl.experiment_id\n");
        fprintf(fout,"_pdbx_nmr_exptl.solution_id\n");
        fprintf(fout,"_pdbx_nmr_exptl.conditions_id\n");
        fprintf(fout,"_pdbx_nmr_exptl.type\n");
        
    }else if(!strcmp_case(id, "")){
        
        
    }
}


